program Solve(input,output,CRYPTO,CCOUNT);

{---------------------------------------------------------------------}
{  This program is an interactive tool for solving cryptograms that   }
{  are encoded using a simple substitution cipher.  The program       }
{  displays an encrypted message on the terminal screen, along        }
{  with an indication of which letters occur most frequently in the   }
{  ciphertext.  For ease of reference, the most common letters in     }
{  typical English text are also displayed.  The program then prompts }
{  the user to enter commands in an effort to decode the message.     }
{                                                                     }
{  The possible commands are                                          }
{       S x=y  -  substitute a plaintext letter y for the ciphertext  }
{                 letter x wherever it occurs                         }
{       H x    -  requests help on ciphertext letter x; the correct   }
{                 plaintext letter is substituted for x wherever      }
{                 it occurs                                           }
{       B x    -  replace the ciphertext letter x by a blank wherever }
{                 it occurs (to erase the effect of a previous        }
{                 incorrect S command)                                }
{       Q      -  quit, displaying the correctly deciphered message   }
{                                                                     }
{  The program shows the effect of each command by displaying a trial }
{  plaintext decipherment above the original ciphertext.  The         }
{  ciphertext message is displayed using lower case characters, and   }
{  the plaintext decipherment is displayed using upper case letters.  }
{  Commands may be entered using any combination of upper case and    }
{  lower case characters.                                             }
{                                                                     }
{  Cryptograms for solution are read from the file CRYPTO.  The first }
{  line of each cryptogram in CRYPTO gives the deciphering key needed }
{  to decode the message.  (Deciphering keys have the same format     }
{  used in program Cipher.)  Following the deciphering key are one    }
{  or more lines containing the enciphered message.  The end of the   }
{  enciphered message is marked by a line containing only the         }
{  character '#'.  The enciphered message may not contain the         }
{  characters '#' or '&'.                                             }
{                                                                     }
{  The file CCOUNT controls the sequence in which cryptograms from    }
{  CRYPTO are used by the program.  CCOUNT consists of one line that  }
{  contains two integers.  The first integer gives the number of the  }
{  last cryptogram from CRYPTO that was used, and the second gives    }
{  the total number of cryptograms stored in CRYPTO.  When the        }
{  program is executed, it reads these numbers and selects from       }
{  CRYPTO the next cryptogram after the one previously used.  If all  }
{  cryptograms have been used, the program returns to the first one   }
{  in CRYPTO.  CCOUNT is rewritten to reflect the use of the          }
{  cryptogram that is selected.                                       }
{                                                                     }
{  Input:  A cryptogram is read from file CRYPTO, as determined by    }
{          the information in CCOUNT.  Commands are read from the     }
{          terminal.                                                  }
{                                                                     }
{  Output: Updated information is rewritten to file CCOUNT.  The      }
{          result of executing each command is displayed on the       }
{          terminal.                                                  }
{                                                                     }
{  This program contains non-standard Pascal statements to support    }
{  Turbo Pascal and Berkeley Pascal versions.  Non-standard Pascal    }
{  statements appear in the following locations:                      }
{    1. in the 'uses' clause immediately following this comment block }
{    2. in the screen control procedures ClearScreen, ClearLine,      }
{       and MoveCursor                                                }
{    3. in procedures Initialize and CleanUp                          }
{  These non-standard statements are preceded by a comment that       }
{  begins with '## N'.                                                }
{---------------------------------------------------------------------}

{ ## Non-standard statements -- Turbo Pascal version }

uses Crt;

{ End Turbo Pascal version }

const
  MaximumLength = 400;        { maximum length for an enciphered message }

type
  Message = array[1..MaximumLength] of char;
  Key = array['a'..'z'] of char;

  LetterFrequencies = record               { table of letter frequencies }
                        Letter : char;       { letter of the alphabet }
                        Frequency : real     { frequency in ciphertext }
                      end;

  CryptoPuzzle = record                    { cryptogram information }
         CipherKey : Key;                    { deciphering key }
         Ciphertext : Message;               { enciphered message }
         CipherLength : integer;             { length of Ciphertext }
         CommonLetters : array[1..8]
                 of LetterFrequencies;       { most common ciphertext letters }
         Plaintext : Message;                { solution being constructed }
         Solved : boolean                    { solution is complete }
       end;  { CryptoPuzzle }

  CommandInfo = record                     { command information }
                  CommandCode : char;        { command code }
                  Parameter1 : char;         { first parameter }
                  Parameter2 : char          { second parameter }
                end;  { CommandInfo }

var
  CRYPTO : text;                      { file containing cryptograms }
  CCOUNT : text;                      { record of cryptograms used }
  Cryptogram : CryptoPuzzle;          { current cryptogram }
  Command : CommandInfo;              { current command }

{*********************************************************************}

{ ## Non-standard statements -- Berkeley Pascal version }
{
procedure ClrScr; external;
}
{ End Berkeley Pascal version }

procedure ClearScreen;

  {-------------------------------------------------------------------}
  {  ClearScreen clears the terminal screen and positions the         }
  {  cursor at the upper left corner of the screen.  It calls a       }
  {  non-standard Pascal library routine to perform this operation.   }
  {-------------------------------------------------------------------}

  begin

    ClrScr   { call library routine ClrScr to clear screen }

  end;  { ClearScreen }

{*********************************************************************}

{ ## Non-standard statements -- Berkeley Pascal version }
{
procedure GotoXY(X, Y : integer); external;
}
{ End Berkeley Pascal version }

procedure MoveCursor(Row, Col : integer);

  {-------------------------------------------------------------------}
  {  MoveCursor moves the cursor on the terminal screen to the        }
  {  row and column given by its input parameters.  The next write    }
  {  or writeln operation that sends output to the terminal will      }
  {  begin writing at the row and column location designated.  The    }
  {  procedure calls a non-standard Pascal library routine to perform }
  {  this operation.                                                  }
  {                                                                   }
  {  Input parameters:                                                }
  {    Row,Col - the row (line) and column (horizontal position) on   }
  {          the screen to which the cursor is to be moved.  The      }
  {          upper left-hand corner of the screen is row 1, column 1. }
  {-------------------------------------------------------------------}

  begin

    { call library routine GotoXY to move the cursor -- note that
      GotoXY expects the column value as its first parameter }

    GotoXY(Col,Row)

  end;  { MoveCursor }

{*********************************************************************}

{ ## Non-standard statements -- Berkeley Pascal version }
{
procedure ClrEol; external;
}
{ End Berkeley Pascal version }

procedure ClearLine;

  {-------------------------------------------------------------------}
  {  ClearLine clears the line on the terminal screen where the       }
  {  cursor is currently positioned, from the cursor position to the  }
  {  end of the line.  It calls a non-standard Pascal library routine }
  {  to perform this function.                                        }
  {-------------------------------------------------------------------}

  begin

    ClrEol     { call library routine ClrEol to clear line }

  end;  { ClearLine }

{*********************************************************************}

function Lowercase(Letter : char) : char;

  {-------------------------------------------------------------------}
  {  Lowercase is used to convert alphabetic characters to lower      }
  {  case.  If the input parameter is an upper case letter            }
  {  ('A' through 'Z'), the value of the function is the corresponding}
  {  lower case letter ('a' through 'z').  If the input parameter is  }
  {  any other character, the value of the function is equal to the   }
  {  input parameter.                                                 }
  {                                                                   }
  {  Input parameters:                                                }
  {    Letter - character to be converted to lower case               }
  {                                                                   }
  {  Output parameters : None                                         }
  {                                                                   }
  {  Function value : the value of the input parameter Letter,        }
  {                   converted to lower case if it was originally    }
  {                   in the range 'A' through 'Z'                    }
  {-------------------------------------------------------------------}

  var
    ConvertedLetter : char;       { the function value being computed }
    DistancePastA : integer;      { the distance in the alphabet between
                                    the letter being converted and 'A' }
  begin  { Lowercase }

    { if the input parameter is an upper case letter, convert it to
      the corresponding lower case letter.  otherwise, the converted
      letter is set equal to the input parameter. }

    if Letter in ['A'..'Z'] then
      begin
        DistancePastA := ord(Letter) - ord('A');
        ConvertedLetter := chr(ord('a') + DistancePastA)
      end  { if }
    else
      ConvertedLetter := Letter;

    { set the function value }

    Lowercase := ConvertedLetter

  end;  { Lowercase }

{*********************************************************************}

function Uppercase(Letter : char) : char;

  {-------------------------------------------------------------------}
  {  Uppercase is used to convert alphabetic characters to upper      }
  {  case.  If the input parameter is a lower case letter             }
  {  ('a' through 'z'), the value of the function is the corresponding}
  {  upper case letter ('A' through 'Z').  If the input parameter is  }
  {  any other character, the value of the function is equal to the   }
  {  input parameter.                                                 }
  {                                                                   }
  {  Input parameters:                                                }
  {    Letter - character to be converted to upper case               }
  {                                                                   }
  {  Output parameters : None                                         }
  {                                                                   }
  {  Function value : the value of the input parameter Letter,        }
  {                   converted to upper case if it was originally    }
  {                   in the range 'a' through 'z'                    }
  {-------------------------------------------------------------------}

  var
    ConvertedLetter : char;       { the function value being computed }
    DistancePastA : integer;      { the distance in the alphabet between
                                    the letter being converted and 'a' }
  begin  { Uppercase }

    { if the input parameter is a lower case letter, convert it to
      the corresponding upper case letter.  otherwise, the converted
      letter is set equal to the input parameter. }

    if Letter in ['a'..'z'] then
      begin
        DistancePastA := ord(Letter) - ord('a');
        ConvertedLetter := chr(ord('A') + DistancePastA)
      end  { if }
    else
      ConvertedLetter := Letter;

    { set the function value }

    Uppercase := ConvertedLetter

  end;  { Uppercase }

{*********************************************************************}

procedure Initialize;

  {-------------------------------------------------------------------}
  {  Initialize is called once at the beginning of the program.       }
  {  It assigns external names for the files CRYPTO and CCOUNT, if    }
  {  this is required in the version of Pascal being used.  It also   }
  {  clears the terminal screen.                                      }
  {-------------------------------------------------------------------}

  begin  { Initialize }

    { ## Non-standard statements -- Turbo Pascal version }

    assign(CRYPTO,'CRYPTO');
    assign(CCOUNT,'CCOUNT');

    { End Turbo Pascal version }

    ClearScreen

  end;  { Initialize }

{*********************************************************************}

procedure GetCipher(var Cryptogram : CryptoPuzzle);

  {-------------------------------------------------------------------}
  {  GetCipher is called to read a cryptogram to be solved from       }
  {  the file CRYPTO.  It first calls an internal procedure           }
  {  FindNextCipher to locate the next cryptogram to be processed.    }
  {  Then it reads the cryptogram, stores the ciphertext and the      }
  {  deciphering key in the record Cryptogram, and initializes the    }
  {  remaining fields in Cryptogram.                                  }
  {                                                                   }
  {  If the ciphertext consists of more than one line, end-of-line    }
  {  is indicated in Cryptogram.Ciphertext with the character '&'.    }
  {                                                                   }
  {  Input parameters: None                                           }
  {                                                                   }
  {  Output parameters:                                               }
  {    Cryptogram - the cryptogram to be solved                       }
  {-------------------------------------------------------------------}

  var
    Letter : char;              { used as subscript for CipherKey }
    I : integer;                { used as subscript for Ciphertext }
    CharacterCount : integer;   { count of characters in Ciphertext }
    Character : char;           { current ciphertext character }

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

procedure FindNextCipher;

  {-------------------------------------------------------------------}
  {  FindNextCipher is called at the beginning of GetCipher to locate }
  {  the next cryptogram in CRYPTO to be processed.  It reads the     }
  {  file CCOUNT, obtaining the number of the last cryptogram that    }
  {  was processed and the total number of cryptograms in CRYPTO.     }
  {  The cryptogram to be processed is the one following the one      }
  {  last processed, or the first one in CRYPTO if the last one was   }
  {  just processed.                                                  }
  {                                                                   }
  {  After determining the number of the cryptogram to be processed,  }
  {  FindNextCipher updates the file CCOUNT.  It then reads the file  }
  {  CRYPTO, skipping cryptograms until the beginning of the one to   }
  {  be processed is reached.                                         }
  {                                                                   }
  {  Input parameters: None                                           }
  {                                                                   }
  {  Output parameters: None                                          }
  {                                                                   }
  {  Side effect:  The file CRYPTO is positioned to the beginning     }
  {                of the cryptogram to be processed.                 }
  {-------------------------------------------------------------------}

  var
    PreviousCipherNumber : integer;    { cryptogram previously processed }
    LastCipherNumber : integer;        { number of cryptograms in CRYPTO }
    CurrentCipherNumber : integer;     { cryptogram to process this time }
    CipherCount : integer;             { count of cryptograms skipped }
    Flag : char;                       { character read from CRYPTO }

  begin  { FindNextCipher }

    { read CCOUNT and determine next cryptogram to process }

    reset(CCOUNT);
    readln(CCOUNT,PreviousCipherNumber,LastCipherNumber);
    if PreviousCipherNumber < LastCipherNumber then
      CurrentCipherNumber := PreviousCipherNumber + 1
    else
      CurrentCipherNumber := 1;

    { update CCOUNT to reflect cryptogram being processed }

    rewrite(CCOUNT);
    writeln(CCOUNT,CurrentCipherNumber:6,LastCipherNumber:6);

    { read CRYPTO, skipping cryptograms until the beginning of the one
      to be processed.  the end of each cryptogram in CRYPTO is marked
      by a line containing the character '#'. }

    reset(CRYPTO);
    CipherCount := 1;
    while CipherCount < CurrentCipherNumber do
      begin
        readln(CRYPTO,Flag);
        if Flag = '#' then CipherCount := CipherCount + 1
      end  { while }

  end;  { FindNextCipher }

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

  begin  { GetCipher }

    { find the cryptogram to be processed }

    FindNextCipher;

    { read the deciphering key }

    for Letter := 'a' to 'z' do
      read(CRYPTO,Cryptogram.CipherKey[Letter]);
    readln(CRYPTO);

    { read the ciphertext from CRYPTO, counting the characters and storing
      them in Cryptogram.Ciphertext }

    read(CRYPTO,Character);
    CharacterCount := 0;
    while Character <> '#' do
      begin
        CharacterCount := CharacterCount + 1;
        Cryptogram.Ciphertext[CharacterCount] := Character;
        if eoln(CRYPTO) then

          { represent end-of-line in Ciphertext with the character '&' }

          begin
            CharacterCount := CharacterCount + 1;
            Cryptogram.Ciphertext[CharacterCount] := '&';
            readln(CRYPTO)
          end;  { if eoln }

        read(CRYPTO,Character)
      end;  { while }

    { store length of ciphertext and set "cryptogram solved" flag to false }

    Cryptogram.CipherLength := CharacterCount;
    Cryptogram.Solved := false;

    { copy non-alphabetic characters in ciphertext to the working plaintext;
      initialize all other plaintext characters to blank }

    for I := 1 to Cryptogram.CipherLength do
      if Cryptogram.Ciphertext[I] in ['a'..'z'] then
        Cryptogram.Plaintext[I] := ' '
      else
        Cryptogram.Plaintext[I] := Cryptogram.Ciphertext[I]

  end;  { GetCipher }

{*********************************************************************}

procedure CountFrequencies(var Cryptogram : CryptoPuzzle);

  {-------------------------------------------------------------------}
  {  CountFrequencies is called to determine the most frequently      }
  {  occurring letters in a cryptogram and to enter these letters     }
  {  with their frequencies in Cryptogram.CommonLetters.  It uses an  }
  {  internal procedure CountCharacters to count the number of        }
  {  occurrences of each letter in the ciphertext, and another        }
  {  internal procedure BuildSortedList to sort the letters by        }
  {  number of occurrences.  It then calculates the relative          }
  {  frequency of the eight most common letters, and enters this      }
  {  information in Cryptogram.CommonLetters.                         }
  {                                                                   }
  {  Input parameters:                                                }
  {    Cryptogram - the cryptogram being processed by the program     }
  {                                                                   }
  {  Output parameters:                                               }
  {    Cryptogram - the input cryptogram, with common letters and     }
  {                 letter frequencies                                }
  {-------------------------------------------------------------------}

  type
    CountOfLetters = array['a'..'z'] of integer;
    LetterStats = record            { entry in table used for sorting }
                   Letter : char;      { letter of the alphabet }
                   Count : integer     { count of occurrences of that letter }
                 end;
    ListOfCounts = array [1..26] of LetterStats;

  var
    LetterCounts : CountOfLetters;    { count of occurrences for each letter }
    SortedList : ListOfCounts;        { table used for sorting by count }
    TotalLetters : integer;           { total number of letters counted }
    I : integer;                      { used as subscript for CommonLetters }

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

procedure CountCharacters(Cryptogram : CryptoPuzzle;
             var LetterCounts : CountOfLetters; var TotalLetters : integer);

  {-------------------------------------------------------------------}
  {  CountCharacters scans the characters in Cryptogram.Ciphertext,   }
  {  one character at a time.  It counts the alphabetic characters    }
  {  found, using the array LetterCounts.  The total number of        }
  {  letters counted is stored in TotalLetters.  All letters in the   }
  {  ciphertext are assumed to be in lower case.                      }
  {                                                                   }
  {  The array LetterCounts and the variable TotalLetters are         }
  {  initialized to zero before the counting begins.                  }
  {                                                                   }
  {  The counting of letters in this procedure is essentially similar }
  {  to that performed by the procedure CountCharacters in program    }
  {  Letters1.  The main difference is that this version takes the    }
  {  letters to be counted from Cryptogram.Ciphertext, whereas the    }
  {  version in Letters1 reads the characters from a data file.       }
  {                                                                   }
  {  Input parameters:                                                }
  {    Cryptogram - the cryptogram whose characters are to be counted }
  {                                                                   }
  {  Output parameters:                                               }
  {    LetterCounts - counts of each alphabetic character in          }
  {                   Cryptogram.Ciphertext                           }
  {    TotalLetters - the total number of letters counted             }
  {-------------------------------------------------------------------}

  var
    Character : char;      { character from ciphertext }
    I : integer;           { used as subscript for Cryptogram.Ciphertext }
    Letter : char;         { used as subscript for LetterCounts }

  begin  { CountCharacters }

    { initialize letter counts and total letters to zero }

    for Character := 'a' to 'z' do
      LetterCounts[Character] := 0;
    TotalLetters := 0;

    { scan the characters in Cryptogram.Ciphertext }

    for I := 1 to Cryptogram.CipherLength do
      begin
        Character := Cryptogram.Ciphertext[I];

        { if the character is 'a' through 'z', add one to the
          corresponding count in LetterCounts }

        if Character in ['a'..'z'] then
          LetterCounts[Character] := LetterCounts[Character] + 1
      end;  { for }

    { calculate the total number of letters counted }

    for Letter := 'a' to 'z' do
      TotalLetters := TotalLetters + LetterCounts[Letter]

  end;  { CountCharacters }

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

procedure BuildSortedList (LetterCounts : CountOfLetters;
                           var LetterList : ListOfCounts);

  {-------------------------------------------------------------------}
  {  BuildSortedList inserts each letter and the number of occurrences}
  {  of that letter into the array LetterList.  The entries in        }
  {  LetterList are then sorted by number of occurrences, using a     }
  {  simple exchange sort.                                            }
  {                                                                   }
  {  Input parameters:                                                }
  {    LetterCounts - number of occurrences of each letter            }
  {                                                                   }
  {  Output parameters:                                               }
  {    LetterList - list of letters and number of occurrences, sorted }
  {                 by number of occurrences                          }
  {-------------------------------------------------------------------}

  var
    Character : char;         { used as subscript for LetterCounts }
    I : integer;              { used as subscript for LetterList }
    Start : integer;          { working variables used in exchange sort }
    Compare : integer;
    Hold : LetterStats;

  begin  { BuildSortedList }

    { enter each letter and its number of occurrences into LetterList.
      Character is used as a subscript for LetterCounts, I as a subscript
      for LetterList }

    Character := 'a';
    for I := 1 to 26 do
      begin
        LetterList[I].Letter := Character;
        LetterList[I].Count := LetterCounts[Character];
        Character := succ(Character)
      end;

    { sort the entries in LetterList by number of occurrences }

    for Start := 1 to 25 do
      for Compare := Start+1 to 26 do

        { Start and Compare are subscripts indicating entries in
          LetterList, with Start preceding Compare in the array. }

        if LetterList[Start].Count < LetterList[Compare].Count then
          begin

            { if the number of occurrences in entry Start is less then
              the number of occurrences in entry Compare, exchange these
              two entries. }

            Hold := LetterList[Start];
            LetterList[Start] := LetterList[Compare];
            LetterList[Compare] := Hold
          end

  end;  { BuildSortedList }

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

  begin  { CountFrequencies }

    { count characters in ciphertext and build list sorted by number
      of occurrences }

    CountCharacters(Cryptogram,LetterCounts,TotalLetters);
    BuildSortedList(LetterCounts,SortedList);

    { insert the eight most common letters, with their frequencies,
      into Cryptogram.CommonLetters }

    for I := 1 to 8 do
      begin
        Cryptogram.CommonLetters[I].Letter := SortedList[I].Letter;
        Cryptogram.CommonLetters[I].Frequency :=
                    SortedList[I].Count / TotalLetters
      end  { for }

  end;  { CountFrequencies }

{*********************************************************************}

procedure DisplayCiphertext(Cryptogram : CryptoPuzzle);

  {-------------------------------------------------------------------}
  {  DisplayCiphertext is called once, after the cryptogram has been  }
  {  read and its letter frequencies calculated.  It displays the     }
  {  ciphertext on the terminal, beginning on line 2 of the screen.   }
  {  If the ciphertext consists of more than one line, the succeeding }
  {  lines are displayed on lines 5, 8, 11 and 14 of the screen.      }
  {  (Space is left between the lines to allow for later display of   }
  {  plaintext decipherments with each line of ciphertext.)  The      }
  {  eight most common ciphertext letters, with their frequencies,    }
  {  are displayed on line 17 of the screen; the eight most common    }
  {  English letters are displayed for reference on line 18.          }
  {                                                                   }
  {  Input parameters:                                                }
  {    Cryptogram - the cryptogram whose ciphertext and letter        }
  {                 frequencies are to be displayed                   }
  {                                                                   }
  {  Output parameters: None                                          }
  {-------------------------------------------------------------------}

  var
    I : integer;              { used as subscript for Cryptogram.Ciphertext }
    LineNumber : integer;     { current line number on the terminal screen }
    Percentage : integer;     { letter frequency, expressed as a percentage }

  begin  { DisplayCiphertext }

    { move cursor to the beginning of line 2 on the screen }

    LineNumber := 2;
    MoveCursor(LineNumber,1);

    { scan the ciphertext, writing each character to the screen.  if the
      character '&' (indicating end-of-line) is found, move to the start
      of the third line below the current cursor position }

    for I := 1 to Cryptogram.CipherLength do
      if Cryptogram.Ciphertext[I] = '&' then
        begin
          LineNumber := LineNumber + 3;
          MoveCursor(LineNumber,1)
        end  { if & }
      else
        write(Cryptogram.Ciphertext[I]);
    writeln;

    { move to line 17 and write list of common letters in ciphertext }

    MoveCursor(17,1);
    write('In ciphertext:');
    for I := 1 to 8 do
      begin

        { express each frequency as a percentage, to the nearest whole percent }

        Percentage := round(100 * Cryptogram.CommonLetters[I].Frequency);
        write('   ',Cryptogram.CommonLetters[I].Letter,Percentage:3,'%')
      end;
    writeln;

    { write list of common letters in English }

    write('In English:      E 13%   T  9%   A  8%   O  8%   N  7%   ');
    writeln('I  7%   R  7%   S  7%')

  end;  { DisplayCiphertext }

{*********************************************************************}

procedure DisplayPlaintext(Cryptogram : CryptoPuzzle);

  {-------------------------------------------------------------------}
  {  DisplayPlaintext is called after each command is processed.      }
  {  It displays the working decipherment from Cryptogram.Plaintext,  }
  {  positioning each character above the corresponding ciphertext    }
  {  character.  In order to do this, it begins the display on line 1 }
  {  of the terminal screen.  If the ciphertext consists of more than }
  {  one line, the succeeding lines of plaintext are displayed on     }
  {  lines 4, 7, 10 and 13 of the screen.                             }
  {                                                                   }
  {  Input parameters:                                                }
  {    Cryptogram - the cryptogram whose working plaintext is to be   }
  {                 displayed                                         }
  {                                                                   }
  {  Output parameters: None                                          }
  {-------------------------------------------------------------------}

  var
    I : integer;               { used as subscript for Cryptogram.Plaintext }
    LineNumber : integer;      { current line number on the terminal screen }

  begin  { DisplayPlaintext }
    LineNumber := 1;
    MoveCursor(LineNumber,1);

    { scan the plaintext, writing each character to the screen.  if the
      character '&' (indicating end-of-line) is found, move to the start
      of the third line below the current cursor position }

    for I := 1 to Cryptogram.CipherLength do
      if Cryptogram.Plaintext[I] = '&' then
        begin
          LineNumber := LineNumber + 3;
          MoveCursor(LineNumber,1)
        end  { if & }
      else
        write(Cryptogram.Plaintext[I]);
    writeln

  end;  { DisplayPlaintext }

{*********************************************************************}

procedure GetCommand(var Command : CommandInfo);

  {-------------------------------------------------------------------}
  {  GetCommand is called each time a new command is to be read from  }
  {  the terminal.  It uses an internal procedure ReadCommandLine     }
  {  to prompt the user for a command and to read the command line    }
  {  entered by the user.  It uses another internal procedure         }
  {  CheckCommand to test whether or not the command is valid.        }
  {  If the command is not valid, the terminal bell is sounded as     }
  {  an error indication.  This prompting and testing is repeated     }
  {  until a valid command is obtained.                               }
  {                                                                   }
  {  Input parameters: None                                           }
  {                                                                   }
  {  Output parameters:                                               }
  {    Command - the valid command and parameters obtained            }
  {-------------------------------------------------------------------}

  var
    CommandIsCorrect : boolean;      { command is valid }
    Bell : char;                     { character to sound terminal bell }

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

  procedure ReadCommandLine(var Command : CommandInfo);

  {-------------------------------------------------------------------}
  {  ReadCommandLine is called to prompt the user for a command and   }
  {  to read the command line entered in response to the  prompt.     }
  {  A maximum of five characters are read from the terminal (the     }
  {  longest valid command consists of five characters).  The first   }
  {  character entered is taken to be the command code.  The third    }
  {  character is taken to be the first parameter for the command     }
  {  (the parameter 'x' in commands  H x , B x , and  S x=y ).  The   }
  {  fifth character is taken to be the second parameter for the      }
  {  command (the parameter 'y' in command  S x=y ).  If fewer than   }
  {  five characters are entered on the command line, the unentered   }
  {  characters are considered as blanks.                             }
  {                                                                   }
  {  The command code and second parameter are converted to upper     }
  {  case, and the first parameter is converted to lower case.  The   }
  {  converted command code and parameters are returned in the        }
  {  procedure parameter Command.                                     }
  {                                                                   }
  {  Input parameters: None                                           }
  {                                                                   }
  {  Output parameters:                                               }
  {    Command - the command code and parameters read                 }
  {-------------------------------------------------------------------}

    var
       CommandLine : array[1..5] of char;  { the input from the terminal }
       I : integer;                        { used as subscript for CommandLine }

    begin  { ReadCommandLine }

      { move to line 21 on the terminal screen and prompt for a command }

      MoveCursor(21,1);
      ClearLine;
      write('Enter command: ');

      { read up to five characters.  if end-of-line is found before five
        characters are read, fill the unread characters with blanks }

      for I := 1 to 5 do
        if eoln(input) then
          CommandLine[I] := ' '
        else
          read(CommandLine[I]);
      readln;

      { convert cases and store command and parameters }

      Command.CommandCode := Uppercase(CommandLine[1]);
      Command.Parameter1 := Lowercase(CommandLine[3]);
      Command.Parameter2 := Uppercase(CommandLine[5])

    end;  { ReadCommandLine }

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

  procedure CheckCommand(Command : CommandInfo; var CommandIsCorrect : boolean);

  {-------------------------------------------------------------------}
  {  CheckCommand is called after a command line has been read.       }
  {  It checks whether or not the command code and command parameters }
  {  are valid, and sets the Boolean variable CommandIsCorrect        }
  {  accordingly.                                                     }
  {                                                                   }
  {  Input parameters:                                                }
  {    Command - the command code and parameters to be checked        }
  {                                                                   }
  {  Output parameters:                                               }
  {    CommandIsCorrect - the result of the check                     }
  {-------------------------------------------------------------------}

    begin  { CheckCommand }

      { command is assumed invalid until it is found to be correct }

      CommandIsCorrect := false;

      { if the command code is valid, check the parameters for the command }

      if Command.CommandCode in ['B','H','S','Q'] then
        case Command.CommandCode of
          'Q'     : if (Command.Parameter1 = ' ')
                    and (Command.Parameter2 = ' ')
                      then CommandIsCorrect := true;
          'B','H' : if (Command.Parameter1 in ['a'..'z'])
                    and (Command.Parameter2 = ' ')
                      then CommandIsCorrect := true;
          'S'     : if (Command.Parameter1 in ['a'..'z'])
                    and (Command.Parameter2 in ['A'..'Z'])
                      then CommandIsCorrect := true
          end  { case }

    end;  { CheckCommand }

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

  begin  { GetCommand }

    { if ASCII codes are being used, the character with ordinal value 7
      causes the terminal bell to sound }

    Bell := chr(7);

    { move to line 20 and display a summary of the possible commands }

    MoveCursor(20,1);
    writeln('Commands are  S x=y , H x , B x , Q ');

    { read and check command lines until a valid command is obtained.
      sound the terminal bell if an invalid command is detected. }

    repeat
      ReadCommandLine(Command);
      CheckCommand(Command,CommandIsCorrect);
      if not CommandIsCorrect then
        write(Bell)
    until CommandIsCorrect

  end;  { GetCommand }

{*********************************************************************}

procedure ProcessCommand(Command : CommandInfo; var Cryptogram : CryptoPuzzle);

  {-------------------------------------------------------------------}
  {  ProcessCommand is called to control the processing of each       }
  {  command entered by the user.  An internal procedure Substitute   }
  {  is used to insert a given plaintext character in positions       }
  {  that correspond to a given ciphertext character.  (The selection }
  {  of the plaintext and ciphertext characters depends upon the      }
  {  command being processed.)  After substitutions are performed,    }
  {  another internal procedure CheckSolved is used to test whether   }
  {  or not the cryptogram has been completely solved.                }
  {                                                                   }
  {  Input parameters:                                                }
  {     Command - the command to be processed                         }
  {     Cryptogram - the cryptogram being solved                      }
  {                                                                   }
  {  Output parameters:                                               }
  {     Cryptogram - with Plaintext updated to reflect the effect of  }
  {                  the command and Solved set to indicate whether   }
  {                  the solution is complete                         }
  {-------------------------------------------------------------------}

  var
    Letter : char;              { used as subscript for CipherKey }
    DecipheredLetter : char;    { the correct decipherment for a letter }

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

  procedure Substitute(var Cryptogram : CryptoPuzzle; CipherCharacter : char;
                       PlainCharacter : char);

  {-------------------------------------------------------------------}
  {  Substitute inserts PlainCharacter into the plaintext in all      }
  {  positions where CipherCharacter occurs in the ciphertext.  It is }
  {  used (in various ways) during the processing of all commands.    }
  {                                                                   }
  {  Input parameters:                                                }
  {    Cryptogram - the cryptogram being solved                       }
  {    PlainCharacter - the plaintext character to be substituted     }
  {    CipherCharacter - the ciphertext character being substituted   }
  {                      for                                          }
  {                                                                   }
  {  Output parameters:                                               }
  {    Cryptogram - with Plaintext modified by the substitution       }
  {-------------------------------------------------------------------}

    var
      I : integer;          { used as subscript for Ciphertext and Plaintext }

    begin  { Substitute }

      for I := 1 to Cryptogram.CipherLength do
        if Cryptogram.Ciphertext[I] = CipherCharacter then
          Cryptogram.Plaintext[I] := PlainCharacter

    end;  { Substitute }

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

  procedure CheckSolved(var Cryptogram : CryptoPuzzle);

  {-------------------------------------------------------------------}
  {  CheckSolved is called after each substitution to check whether   }
  {  or not the cryptogram is completely solved.  It scans the        }
  {  plaintext and ciphertext, checking whether each plaintext        }
  {  character is the correct decipherment of the corresponding       }
  {  ciphertext character.  (Cryptogram.CipherKey gives the correct   }
  {  decipherment for each ciphertext character.)  If any incorrect   }
  {  decipherment is found, the scan is stopped and Cryptogram.Solved }
  {  is set to false; if no errors are found, Cryptogram.Solved is    }
  {  set to true.                                                     }
  {                                                                   }
  {  Input parameters:                                                }
  {    Cryptogram - the cryptogram to be tested for solution          }
  {                                                                   }
  {  Output parameters:                                               }
  {    Cryptogram - with Solved set to indicate whether the solution  }
  {                 is correct                                        }
  {-------------------------------------------------------------------}

    var
      I : integer;                { subscript for Ciphertext and Plaintext }
      NoErrorsFound : boolean;    { ciphertext and plaintext match so far }
      CipherCharacter : char;     { character from ciphertext }
      Decipherment : char;        { correct decipherment of CipherCharacter }

    begin  { CheckSolved }

      { scan ciphertext and plaintext until an error is found or the end
        of the ciphertext is reached }

      I := 1;
      NoErrorsFound := true;
      while (NoErrorsFound) and (I < Cryptogram.CipherLength) do
        begin
          CipherCharacter := Cryptogram.Ciphertext[I];

          { if the current ciphertext character is alphabetic, get the
            correct decipherment of it from CipherKey.  check whether the
            corresponding plaintext character is correct. }

          if CipherCharacter in ['a'..'z'] then
            begin
              Decipherment := Cryptogram.CipherKey[CipherCharacter];
              if Cryptogram.Plaintext[I] <> Decipherment then
                NoErrorsFound := false
             end;
           I := I + 1
         end;  { while }

      { if no errors were found, the cryptogram is correctly solved }

      Cryptogram.Solved := NoErrorsFound

    end;  { CheckSolved }

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

  begin  { ProcessCommand }

    case Command.CommandCode of

      'B' : { command is  B x -- substitute blanks for 'x' }

            Substitute(Cryptogram,Command.Parameter1,' ');

      'S' : { command is  S x=y -- substitute 'y' for 'x' }

            begin  { 'S' }
              Substitute(Cryptogram,Command.Parameter1,Command.Parameter2);
              CheckSolved(Cryptogram)
            end;  { 'S' }

      'H' : { command is H x -- obtain the correct decipherment for 'x'
              and substitute it for 'x' }

            begin  { 'H' }
              DecipheredLetter := Cryptogram.CipherKey[Command.Parameter1];
              Substitute(Cryptogram,Command.Parameter1,DecipheredLetter);
              CheckSolved(Cryptogram)
            end;  { 'H' }

      'Q' : { command is Q -- substitute the correct decipherment for all
              ciphertext characters }

             begin  { 'Q' }
              for Letter := 'a' to 'z' do
                begin
                  DecipheredLetter := Cryptogram.CipherKey[Letter];
                  Substitute(Cryptogram,Letter,DecipheredLetter)
                end;  { for }
              Cryptogram.Solved := true
            end  { 'Q' }

      end  { case }

  end;  { ProcessCommand }

{*********************************************************************}

procedure CleanUp;

  {-------------------------------------------------------------------}
  {  CleanUp is called once at the end of the program.  It closes     }
  {  the files CRYPTO and CCOUNT, if this is required in the Pascal   }
  {  version being used.  It also moves the cursor to line 22 on      }
  {  the screen, and waits for the user to enter a carriage return    }
  {  before exiting the program.                                      }
  {-------------------------------------------------------------------}

  begin  { CleanUp }

    { ## Non-standard statements -- Turbo Pascal version }

    close(CRYPTO);
    close(CCOUNT);

    { End Turbo Pascal version }

    MoveCursor(22,1);
    ClearLine;
    write('Cryptogram solved -- press return to exit program');
    readln

  end;  { CleanUp }

{*********************************************************************}

begin  { main program }

  Initialize;

  { read the cryptogram to be solved, count the frequencies of ciphertext
    characters, and display the ciphertext on the screen }

  GetCipher(Cryptogram);
  CountFrequencies(Cryptogram);
  DisplayCiphertext(Cryptogram);

  { read and process commands until the cryptogram is solved.  after each
    command is processed, display the partial plaintext solution }

  while not Cryptogram.Solved do
    begin
      GetCommand(Command);
      ProcessCommand(Command,Cryptogram);
      DisplayPlaintext(Cryptogram)
    end;  { while }

  CleanUp

end.
