program Wator(input,output);

{---------------------------------------------------------------------}
{  This program simulates the ecological relationship of the two      }
{  dominant species on the unusual planet Wa-Tor.  This planet is     }
{  shaped like a torus (a doughnut), and is entirely covered by       }
{  water -- hence its name.  The two dominant species on Wa-Tor are   }
{  called sharks and fish, after the earthly creatures they most      }
{  closely resemble.  The sharks eat the fish, and the fish live on   }
{  microscopic organisms uniformly distributed throughout the water.  }
{                                                                     }
{  This program simulates the ocean of Wa-Tor on a rectangular grid   }
{  that has 18 rows and 79 columns.  Each cell in the grid can hold   }
{  one fish or one shark (or may be empty).  The left and right edges }
{  of the grid are considered adjacent --  for example, a shark at    }
{  right-hand side of the grid that swims to the right will reappear  }
{  at the left-hand side.  Similarly, the top and bottom edges of the }
{  grid are considered adjacent.  These adjacencies make the grid     }
{  equivalent to the torus shape of the planet.                       }
{                                                                     }
{  Time passes in discrete jumps called chronons.  During each        }
{  chronon, each fish swims to an adjacent empty cell (selected at    }
{  random).  If all adjacent cells are occupied, the fish does not    }
{  move.  After the fish move, each shark swims to an adjacent cell   }
{  occupied by a fish and eats the fish.  If there are no adjacent    }
{  cells containing fish, the shark moves at random in the same way   }
{  a fish does.  When each shark or fish reaches a certain age, it    }
{  produces a single offspring.  (Both species reproduce by cloning   }
{  themselves.)  If a shark does not find food within a certain       }
{  length of time, it starves.                                        }
{                                                                     }
{  At the beginning of the simulation, the program distributes sharks }
{  and fish in the ocean in a random pattern.  It then simulates the  }
{  swimming, breeding, and hunting behaviors described above for a    }
{  specified number of chronons.  After each move, the distribution   }
{  of sharks and fish is displayed on the terminal.  In this display, }
{  sharks are indicated by "o" and fish by ".".                       }
{                                                                     }
{  The procedure Display, which is contained in the file "display.i', }
{  is used to display the contents of the ocean.                      }
{                                                                     }
{  The program keeps track of the sharks and fish by maintaining a    }
{  list of each type of creature.  This list records the position of  }
{  the shark or fish, its age, and (for sharks) the length of time    }
{  since it has eaten.  An abstract data type ListOfRecords is used   }
{  to store these lists.  The procedures and function that implement  }
{  ListOfRecords are contained in the file "reclist.i".               }
{                                                                     }
{  The simulation is governed by the following parameters, which may  }
{  be set by the user at the beginning of the run:                    }
{                                                                     }
{    FishInitial  - the initial number of fish in the ocean           }
{    FishBreed    - the age (in chronons) at which a fish reproduces  }
{    FishLimit    - the maximum number of fish (limit imposed by the  }
{                   amount of fish food available)                    }
{    SharkInitial - the initial number of sharks in the ocean         }
{    SharkBreed   - the age (in chronons) at which a shark reproduces }
{    SharkStarve  - the length of time (in chronons) a shark can      }
{                   survive without eating                            }
{                                                                     }
{  Default values are used for any of these parameters that the user  }
{  does not specify.                                                  }
{                                                                     }
{  Further details and discussions about Wa-Tor can be found in a     }
{  "Computer Recreations" column by A. K. Dewdney (Scientific         }
{  American, December 1984.)  The simulation performed by this        }
{  program is similar to the one discussed there.                     }
{                                                                     }
{  Input:  The parameters for the simulation are entered by the user  }
{          from the terminal.                                         }
{                                                                     }
{  Output: The distribution of sharks and fish, and the total number  }
{          of each, are displayed on the terminal after each move.    }
{                                                                     }
{  This program contains non-standard Pascal statements to support    }
{  Turbo Pascal and Berkeley Pascal versions.  Non-standard Pascal    }
{  statements appear in the following locations:                      }
{    1. in the 'uses' clause immediately following this comment block }
{    2. in the function RandomInteger                                 }
{    3. in the procedure Display
{  These non-standard statements are preceded by a comment that       }
{  begins with '## N'.                                                }
{                                                                     }
{  This program is identical to the program Wator discussed in        }
{  Chapter 14.  However, the files screen.i, random.i, display.i, and }
{  reclist.i are placed directly in the program instead of being      }
{  inserted with "include" commands.  This is done to simplify the    }
{  exercises for Chapter 1.                                           }
{---------------------------------------------------------------------}

  { ## Non-standard statements -- Turbo Pascal version }

  uses Crt;

  { End Turbo Pascal version }

const

  { default values for the simulation parameters }

  FishInitialDefault = 200;
  FishBreedDefault = 3;
  FishLimitDefault = 400;
  SharkInitialDefault = 20;
  SharkBreedDefault = 3;
  SharkStarveDefault = 3;

  { end default values }

  CellRows = 18;               { number of rows in the ocean array }
  CellCols = 79;               { number of columns in the ocean array }

  FishSymbol = '.';            { character used to display a fish }
  SharkSymbol = 'o';           { character used to display a shark }
  PauseAfter = 50;             { number of chronons to run before pausing }
  MaximumListSize = 1422;      { maximum size of a list of sharks or fish --
                                 should be equal to CellRows * CellCols }

type

  CellArray = array[1..CellRows,1..CellCols] of char;
  RowSubscript = 1..CellRows;
  ColSubscript = 1..CellCols;

  ListEntry = record                   { entry in list of sharks or fish }
                Row : RowSubscript;      { row and column position }
                Col : ColSubscript;
                Age : integer;           { age of the shark or fish }
                TimeSinceFed : integer   { time since shark has eaten }
              end;  { ListEntry }

  ListOfRecords = record
                    MaximumSize : 1..MaximumListSize;
                    CurrentSize : integer;
                    Entries : array[1..MaximumListSize] of ListEntry;
                    CurrentEntry : integer;
                    NextEntry : integer;
                  end;  { ListOfRecords }

var

  {-----------------------------------------------------------------}
  {  The following variables represent parameter values used in the }
  {  simulation.  These variables are assigned values at the start  }
  {  of the simulation by the procedure GetParameters.  Thereafter, }
  {  they remain constant during the execution of the program.      }
  {                                                                 }
  {  These variables are used freely by the other procedures of the }
  {  program, as though they were true constants.  They are not     }
  {  passed as parameters, except to the procedure GetParameters.   }
  {-----------------------------------------------------------------}

  FishInitial : integer;       { initial number of fish in the ocean }
  FishBreed : integer;         { age at which a fish reproduces }
  FishLimit : integer;         { maximum number of fish }
  SharkInitial : integer;      { initial number of sharks in the ocean }
  SharkBreed : integer;        { age at which a shark reproduces }
  SharkStarve : integer;       { length of time a shark can go without eating }

  {-----------------------------------------------------------------}
  { End of simulation parameters                                    }
  {-----------------------------------------------------------------}

  Ocean : CellArray;           { current contents of ocean }
  OldOcean : CellArray;        { previous contents of ocean }
  FishList : ListOfRecords;    { list of fish at end of last breeding cycle }
  SharkList : ListOfRecords;   { list of sharks at end of last breeding cycle }
  CycleNumber : integer;       { number of chronons simulated so far }
  WantsToContinue : boolean;   { user wants to continue the simulation }

{*********************************************************************}

{ ## Non-standard statements -- Berkeley Pascal version }
{
procedure ClrScr; external;
}
{ End Berkeley Pascal version }

procedure ClearScreen;

  {-------------------------------------------------------------------}
  {  ClearScreen clears the terminal screen and positions the         }
  {  cursor at the upper left corner of the screen.  It calls a       }
  {  non-standard Pascal library routine to perform this operation.   }
  {-------------------------------------------------------------------}

  begin

    ClrScr   { call library routine ClrScr to clear screen }

  end;  { ClearScreen }

{*********************************************************************}

{ ## Non-standard statements -- Berkeley Pascal version }
{
procedure GotoXY(X, Y : integer); external;
}
{ End Berkeley Pascal version }

procedure MoveCursor(Row, Col : integer);

  {-------------------------------------------------------------------}
  {  MoveCursor moves the cursor on the terminal screen to the        }
  {  row and column given by its input parameters.  The next write    }
  {  or writeln operation that sends output to the terminal will      }
  {  begin writing at the row and column location designated.  The    }
  {  procedure calls a non-standard Pascal library routine to perform }
  {  this operation.                                                  }
  {                                                                   }
  {  Input parameters:                                                }
  {    Row,Col - the row (line) and column (horizontal position) on   }
  {          the screen to which the cursor is to be moved.  The      }
  {          upper left-hand corner of the screen is row 1, column 1. }
  {-------------------------------------------------------------------}

  begin

    { call library routine GotoXY to move the cursor -- note that
      GotoXY expects the column value as its first parameter }

    GotoXY(Col,Row)

  end;  { MoveCursor }

{*********************************************************************}

{ ## Non-standard statements -- Berkeley Pascal version }
{
procedure ClrEol; external;
}
{ End Berkeley Pascal version }

procedure ClearLine;

  {-------------------------------------------------------------------}
  {  ClearLine clears the line on the terminal screen where the       }
  {  cursor is currently positioned, from the cursor position to the  }
  {  end of the line.  It calls a non-standard Pascal library routine }
  {  to perform this function.                                        }
  {-------------------------------------------------------------------}

  begin

    ClrEol     { call library routine ClrEol to clear line }

  end;  { ClearLine }


{*********************************************************************}

procedure Display(var Cells : CellArray; var OldCells : CellArray;
                  StartLine : integer);

  {-------------------------------------------------------------------}
  {  Display displays the contents of the array Cells on the          }
  {  terminal, one row per line of the screen.  The first row of      }
  {  cells begins on the line specified by the parameter StartLine.   }
  {                                                                   }
  {  After displaying the contents of Cells, this version of the      }
  {  procedure delays further execution of the program by a specified }
  {  number of milliseconds.  This delay is introduced to adjust the  }
  {  pace of the screen display on fast machines.  The number of      }
  {  milliseconds of delay is specified by the constant DelayCount,   }
  {  which can be changed to provide satisfactory operation on your   }
  {  machine.  The delay is accomplished by calling the nonstandard   }
  {  Turbo Pascal procedure Delay.                                    }
  {                                                                   }
  {  It is assumed that the contents of the array OldCells have       }
  {  already been displayed, and that this display remains on the     }
  {  terminal screen.  Display attempts to avoid re-displaying        }
  {  cells that have not changed from the previous version.  It       }
  {  does this by scanning each row of Cells, comparing the cell      }
  {  contents to those stored in OldCells.                            }
  {                                                                   }
  {  If the contents of a cell have changed, an internal procedure    }
  {  DisplaySegment is called to display a segment of the row that    }
  {  begins with the changed cell.  The scan then continues, skipping }
  {  over the segment that was displayed.                             }
  {                                                                   }
  {  Before Display is called for the first time, the terminal screen }
  {  should be cleared to blanks, and OldCells should be initialized  }
  {  to blanks.  This insures that the first version of the array     }
  {  will be correctly displayed.                                     }
  {                                                                   }
  {  Input parameters:                                                }
  {    Cells - the array whose contents are to be displayed           }
  {    OldCells - the previous version of the array, or blanks if     }
  {               this is the first call to Display                   }
  {      Note: Cells and OldCells are var parameters for efficiency   }
  {            in parameter passing.                                  }
  {    StartLine - the line on the screen where the display of Cells  }
  {               is to begin                                         }
  {                                                                   }
  {  Output parameters: None                                          }
  {                                                                   }
  {  In order to use this procedure, the main program must contain    }
  {  the following constant and type definitions.  Items enclosed     }
  {  in <<...>> are to be supplied by the programmer.                 }
  {                                                                   }
  {    const                                                          }
  {      CellRows = <<integer between 1 and 20>>;                     }
  {      CellCols = <<integer between 1 and 79>>;                     }
  {                                                                   }
  {    type                                                           }
  {      CellArray = array[1..CellRows,1..CellCols] of char;          }
  {      RowSubscript = 1..CellRows;                                  }
  {      ColSubscript = 1..CellCols;                                  }
  {                                                                   }
  {  In addition, the main program must include the terminal output   }
  {  routines from the file "screen.i".                               }
  {-------------------------------------------------------------------}

  const
    DelayCount = 250;             { number of milliseconds to delay }

  var
    Row : RowSubscript;           { Row and Col are used as subscripts }
    Col : ColSubscript;           {    for Cells and OldCells          }
    Length : integer;             { the length of the segment displayed }
    EndOfRow : boolean;           { the end of a row has been reached }

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

  procedure DisplaySegment(var Cells: CellArray; var OldCells : CellArray;
                           Row : RowSubscript; Col : ColSubscript;
                           var Length : integer; StartLine : integer);

  {-------------------------------------------------------------------}
  {  DisplaySegment is called each time Display finds a cell that     }
  {  has changed in value from the previous version.  It scans the    }
  {  row containing the changed cell to find the segment to be        }
  {  displayed.  This scan begins with the changed cell and           }
  {  continues until the end of the row is reached, or until six      }
  {  consecutive unchanged cells are found in the row.                }
  {                                                                   }
  {  (This rule for constructing the segment to display takes into    }
  {  account a typical amount of overhead involved in moving the      }
  {  the cursor to a new location.  On most terminals, it is more     }
  {  efficient to display a small number of unchanged cells than to   }
  {  reposition the cursor to skip over them.)                        }
  {                                                                   }
  {  The segment that is found is displayed on the terminal, after    }
  {  moving the cursor to the proper starting location.  The length   }
  {  of the segment displayed is returned to the calling procedure.   }
  {                                                                   }
  {  Input parameters:                                                }
  {    Cells - the array whose contents are being displayed           }
  {    OldCells - the previous version of the array, or blanks if     }
  {               this is the first call to DisplaySegment            }
  {      Note: Cells and OldCells are var parameters for efficiency   }
  {            in parameter passing.                                  }
  {    Row,Col - the row and column in which a changed cell occurs    }
  {    StartLine - the line on the screen where the display of Cells  }
  {               begins                                              }
  {                                                                   }
  {  Output parameters:                                               }
  {    Length - the length of the segment displayed                   }
  {-------------------------------------------------------------------}

    var
      CurrentCol : ColSubscript;     { used as subscript for scanning the row }
      LastChangedCol : ColSubscript; { last changed cell found in the row }
      UnchangedCells : integer;      { number of consecutive unchanged cells }
      WriteCol : ColSubscript;       { used as subscript to display segment }
      EndOfSegment : boolean;        { the end of the segment to display
                                       has been found }

    begin  { DisplaySegment }

      { start scanning with the changed column passed as parameter }

      LastChangedCol := Col;
      CurrentCol := Col;
      UnchangedCells := 0;
      EndOfSegment := false;

      { scan cells until the end of the row is reached, or until six
        consecutive unchanged cells are found }

      repeat  { until EndOfSegment }
        if CurrentCol = CellCols then
          EndOfSegment := true
        else
          begin

            { not at end of row -- check whether next cell is changed }

            CurrentCol := CurrentCol + 1;
            if Cells[Row,CurrentCol] = OldCells[Row,CurrentCol] then
              begin

                { count unchanged cell; stop scanning if six unchanged
                  cells have been found }

                UnchangedCells := UnchangedCells + 1;
                if UnchangedCells = 6 then
                  EndOfSegment := true
              end  { if cell is unchanged }
            else
              begin

                { remember where the changed cell was found and reset the
                  count of consecutive unchanged cells to zero }

                LastChangedCol := CurrentCol;
                UnchangedCells := 0
              end  { else -- cell is changed }
          end  { else -- not at end of row }
      until EndOfSegment;

      { position cursor and display segment, ending with last changed cell
        that was found }

      MoveCursor(StartLine + Row - 1,Col);
      for WriteCol := Col to LastChangedCol do
        write(Cells[Row,WriteCol]);
      writeln;

      { set length of segment that was displayed }

      Length := LastChangedCol - Col + 1

    end;  { DisplaySegment }

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

  begin  { Display }

    { scan each row of the cell array, looking for a cell that has
      changed from the previous generation }

    for Row := 1 to CellRows do
      begin
        Col := 1;
        EndOfRow := false;
        repeat  { until EndOfRow }
          if Cells[Row,Col] = OldCells[Row,Col] then
            begin

              { cell is unchanged -- go on to next cell, unless at end of row }

              if Col = CellCols then
                EndOfRow := true
              else
                Col := Col + 1

            end  { if cell is unchanged }
          else
            begin

              { a changed cell was found -- display a segment starting with
                that cell and skip over the segment displayed }

              DisplaySegment(Cells,OldCells,Row,Col,Length,StartLine);
              if (Col + Length) > CellCols then
                EndOfRow := true
              else
                Col := Col + Length
            end  { else -- cell is changed }
        until EndOfRow
      end;  { for }

    { introduce a delay to adjust the pace of the display for fast machines }

    { ## Non-standard statements -- Turbo Pascal version }

    Delay(DelayCount)

    { End Turbo Pascal version }

  end;  { Display }

{*********************************************************************}

function RandomNumber : real;

  {-------------------------------------------------------------------}
  {  RandomNumber returns a random real number between 0 and 1.  The  }
  {  random numbers generated are uniformly distributed in this       }
  {  range.  Before calling RandomNumber, the random-number generator }
  {  should be initialized by a call to procedure InitRandom.         }
  {                                                                   }
  {  RandomNumber calls a non-standard Pascal library routine to      }
  {  generate the random numbers required.                            }
  {                                                                   }
  {  Function value: a random real number between 0 and 1             }
  {-------------------------------------------------------------------}

  begin  { RandomNumber }

    { ## Non-standard statements -- Turbo Pascal version }

    RandomNumber := Random

    { End Turbo Pascal version }

    { ## Non-standard statements -- Berkeley Pascal version }
    {
    RandomNumber := random(1.0)
    }
    { End Berkeley Pascal version }

  end;  { RandomNumber }

{*********************************************************************}

procedure InitRandom(SeedValue : integer);

  {-------------------------------------------------------------------}
  {  InitRandom initializes the random-number generator.  The         }
  {  sequence of random numbers generated is determined by the        }
  {  parameter SeedValue, if this value is not zero.  If SeedValue    }
  {  is zero, the random-number generator is initialized with a       }
  {  value derived from the system clock.  This causes a different    }
  {  sequence of random numbers to be generated each time the program }
  {  is executed.  The procedure calls a non-standard Pascal library  }
  {  routine to perform this operation.                               }
  {                                                                   }
  {  Input parameters:                                                }
  {    SeedValue - a value to be used in initializing the random      }
  {           number generator, or zero                               }
  {                                                                   }
  {  Output parameters: none                                          }
  {                                                                   }
  {  Side effect:  The random number generator is initialized,        }
  {                depending upon the value of SeedValue              }
  {-------------------------------------------------------------------}

  var
    Dummy : integer;        { a dummy variable required by the syntax of
                              Berkeley Pascal -- value is never used }

  begin  { InitRandom }

    { ## Non-standard statements -- Turbo Pascal version }

    if SeedValue = 0 then
      Randomize
    else
      RandSeed := SeedValue

    { End Turbo Pascal version }

    { ## Non-standard statements -- Berkeley Pascal version }
    {
    if SeedValue = 0 then
      Dummy := seed(wallclock)
    else
      Dummy := seed(SeedValue)
    }
    { End Berkeley Pascal version }

  end;  { InitRandom }

{*********************************************************************}

{---------------------------------------------------------------------}
{  The following procedures and function implement operations on      }
{  the abstract data type ListOfRecords.  This data type can be used  }
{  to store and retrieve a collection of records. The fields          }
{  contained in these records can be defined by the programmer.       }
{                                                                     }
{  Procedures:                                                        }
{    InitList - initialize a new (empty) list                         }
{    GetFirst - get the first record (if any) in a list               }
{    GetNext - get the next record (if any) in a list                 }
{    UpdateCurrent - update the last record read from a list          }
{    DeleteCurrent - delete the last record read from a list          }
{    AddNew - add a new record to a list                              }
{                                                                     }
{  Function:                                                          }
{    NumberIn - returns the number of records in a list               }
{                                                                     }
{  The records in a list can be retrieved, one at a time, by using    }
{  the operations GetFirst and GetNext.  GetFirst begins a new scan   }
{  of the records in the list and retrieves the first record.  Each   }
{  record after the first is retrieved with GetNext.  When no more    }
{  records remain to be retrieved, the parameter EndOfList is set to  }
{  true.                                                              }
{                                                                     }
{  This sequence of operations is guaranteed to retrieve every record }
{  in the list before terminating.  However, the order in which the   }
{  records are retrieved is not defined, and may vary from one scan   }
{  of the list to the next.                                           }
{                                                                     }
{  The record most recently retrieved is designated as the current    }
{  record of the list.  This record can be modified by using the      }
{  operation UpdateCurrent, or deleted from the list by using the     }
{  operation DeleteCurrent.  It is not possible to modify or delete   }
{  records other than the current record.                             }
{                                                                     }
{  New records can be added to the list at any time with the          }
{  operation AddNew.  If a new record is added during a scan of the   }
{  list, it will be retrieved later during that same scan.            }
{                                                                     }
{  In order to use these routines, the main program must contain the  }
{  following constant and type definitions.  Items enclosed in        }
{  <<...>> are to be supplied by the programmer.                      }
{                                                                     }
{    const                                                            }
{      MaximumListSize = << maximum size of a list >>                 }
{                                                                     }
{    type                                                             }
{      ListEntry = record                                             }
{                    << definitions of fields in a record >>          }
{                  end;                                               }
{                                                                     }
{      ListOfRecords = record                                         }
{                MaximumSize : 1..MaximumListSize;                    }
{                CurrentSize : integer;                               }
{                Entries : array[1..MaximumListSize] of ListEntry     }
{                CurrentEntry : integer;                              }
{                NextEntry : integer;                                 }
{              end;                                                   }
{                                                                     }
{  In this implementation of ListOfRecords, a list is represented by  }
{  a record that contains the following fields:                       }
{                                                                     }
{    MaximumSize - the maximum number of records in the list          }
{    CurrentSize - the number of records currently in the list        }
{    Entries - an array containing the records in the list            }
{    CurrentEntry - the subscript of the current record in Entries    }
{    NextEntry - the subscript of the next record in Entries          }
{                                                                     }
{  CurrentEntry is 0 if the current entry has been deleted.           }
{  NextEntry is 0 if the current record is the last one in the list.  }
{  Both CurrentEntry and NextEntry are 0 if no records have yet been  }
{  retrieved.                                                         }
{                                                                     }
{  The records in the list always occupy contiguous elements of the   }
{  array Entries.  The first record occupies the element with         }
{  subscript 1, and the last record occupies the element with         }
{  subscript CurrentSize.  When a record is deleted from the list,    }
{  the remaining records are rearranged as needed to maintain this    }
{  arrangement.                                                       }
{---------------------------------------------------------------------}


procedure InitList (var List : ListOfRecords; ListLimit : integer);

  {-------------------------------------------------------------------}
  {  InitList initializes a new (empty) list, setting its maximum     }
  {  size to the value specified in ListLimit.  If the list size      }
  {  specified is too large, an error message is displayed and the    }
  {  maximum allowable list size is used.                             }
  {                                                                   }
  {  Input parameters:                                                }
  {    List - the list to be initialized                              }
  {    ListLimit - the maximum size of the list                       }
  {                                                                   }
  {  Output parameters:                                               }
  {    List - the initialized list                                    }
  {-------------------------------------------------------------------}

  begin  { InitList }

    if ListLimit > MaximumListSize then
      begin
        writeln('*** InitList: specified size is greater than maximum');
        List.MaximumSize := MaximumListSize
      end  { if }
    else
      List.MaximumSize := ListLimit;
    List.CurrentSize := 0;
    List.CurrentEntry := 0;
    List.NextEntry := 0

  end;  { InitList }

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

procedure GetFirst(var List : ListOfRecords; var Entry : ListEntry;
                    var EndOfList : boolean);

  {-------------------------------------------------------------------}
  {  GetFirst gets the first entry (if any) in List, returning it in  }
  {  the parameter Entry.  The entry retrieved is made the current    }
  {  entry for the list.  The following entry (if any) is made the    }
  {  next entry.  If the list contains no entries, EndOfList is set   }
  {  to true; otherwise, EndOfList is set to false.                   }
  {                                                                   }
  {  Input parameters:                                                }
  {    List - the list whose first entry is to be retrieved           }
  {                                                                   }
  {  Output parameters:                                               }
  {    List - current entry set to the entry retrieved, and next      }
  {           entry set to the following entry (if any)               }
  {    Entry - the first entry in List; this parameter remains        }
  {            unchanged if the list is empty                         }
  {    EndOfList - true if List is empty; false otherwise             }
  {-------------------------------------------------------------------}

  begin  { GetFirst }

    if List.CurrentSize = 0 then
      EndOfList := true
    else
      begin
        List.CurrentEntry := 1;
        if List.CurrentEntry = List.CurrentSize then
          List.NextEntry := 0
        else
          List.NextEntry := List.CurrentEntry + 1;
        Entry := List.Entries[List.CurrentEntry];
        EndOfList := false
      end  { else }

  end;  { GetFirst }

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

procedure GetNext(var List : ListOfRecords; var Entry : ListEntry;
                    var EndOfList : boolean);

  {-------------------------------------------------------------------}
  {  GetNext gets the next entry in List (if any), returning it in    }
  {  the parameter Entry.  If there is no next entry in the list,     }
  {  EndOfList is set to true; otherwise, EndOfList is set to false.  }
  {                                                                   }
  {  The entry retrieved (if any) is made the current entry for the   }
  {  list. The following entry (if any) is made the next entry.       }
  {                                                                   }
  {  Input parameters:                                                }
  {    List - the list whose next entry is to be retrieved            }
  {                                                                   }
  {  Output parameters:                                               }
  {    List - current entry set to the entry retrieved, and next      }
  {           entry set to the following entry (if any)               }
  {    Entry - the next entry in List; this parameter remains         }
  {            unchanged if there is no next entry to retrieve        }
  {    EndOfList - true if there was no next entry to retrieve;       }
  {                false otherwise                                    }
  {-------------------------------------------------------------------}

  begin  { GetNext }

    if List.NextEntry = 0 then
      EndOfList := true
    else
      begin
        List.CurrentEntry := List.NextEntry;
        if List.CurrentEntry = List.CurrentSize then
          List.NextEntry := 0
        else
          List.NextEntry := List.CurrentEntry + 1;
        Entry := List.Entries[List.CurrentEntry];
        EndOfList := false
      end

  end;  { GetNext }

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

procedure AddNew(var List : ListOfRecords; NewEntry : ListEntry);

  {-------------------------------------------------------------------}
  {  AddNew adds the entry NewEntry to List.  If the list is already  }
  {  at its maximum size, an error message is displayed and the       }
  {  new entry is not added.  The current list entry is not changed   }
  {  by this operation. If the next entry in the list was previously  }
  {  undefined, the new entry becomes the next entry.                 }
  {                                                                   }
  {  Input parameters:                                                }
  {    List - the list to which NewEntry is to be added               }
  {    NewEntry - the entry to add                                    }
  {                                                                   }
  {  Output parameters:                                               }
  {    List - with NewEntry added at the end of the list              }
  {-------------------------------------------------------------------}

  begin  { AddNew }

    if List.CurrentSize = List.MaximumSize then
      writeln('*** AddNew: list overflow')
    else
      begin
        List.CurrentSize := List.CurrentSize + 1;
        List.Entries[List.CurrentSize] := NewEntry;
        if List.NextEntry = 0 then
          List.NextEntry := List.CurrentSize
      end  { else }

  end;  { AddNew }

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

procedure DeleteCurrent(var List : ListOfRecords);

  {-------------------------------------------------------------------}
  {  DeleteCurrent deletes the current entry from List.  If there is  }
  {  no current entry defined, an error message is displayed. The     }
  {  current entry for the list becomes undefined after this          }
  {  operation.                                                       }
  {                                                                   }
  {  If the entry being deleted is not the last one in the list, the  }
  {  last entry in the list is shifted to replace the entry that was  }
  {  deleted.  (This is done to avoid leaving unused entries within   }
  {  the list.)  This entry becomes the new next entry in the list.   }
  {                                                                   }
  {  Input parameters:                                                }
  {    List - the list whose current entry is to be deleted           }
  {                                                                   }
  {  Output parameters:                                               }
  {    List - with the entry deleted                                  }
  {-------------------------------------------------------------------}

  begin  { DeleteCurrent }

    if List.CurrentEntry = 0 then
      writeln('*** DeleteCurrent: no current entry to delete')
    else
      begin
        if List.CurrentEntry < List.CurrentSize then
          begin
            List.Entries[List.CurrentEntry] := List.Entries[List.CurrentSize];
            List.NextEntry := List.CurrentEntry
          end; { if }
        List.CurrentSize := List.CurrentSize - 1;
        List.CurrentEntry := 0
      end  { else }

  end;  { DeleteCurrent }

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

procedure UpdateCurrent(var List : ListOfRecords; Entry : ListEntry);

  {-------------------------------------------------------------------}
  {  UpdateCurrent updates the current entry in List, replacing it    }
  {  with Entry.  If there is no current entry defined, an error      }
  {  message is displayed.  After this operation, the current entry   }
  {  remains the one that was updated (with the new contents).        }
  {                                                                   }
  {  Input parameters:                                                }
  {    List - the list whose current entry is to be updated           }
  {    Entry - the new contents for the current entry                 }
  {                                                                   }
  {  Output parameters:                                               }
  {    List - with the current entry updated                          }
  {-------------------------------------------------------------------}

  begin  { UpdateCurrent }

    if List.CurrentEntry = 0 then
      writeln('*** UpdateCurrent: no current entry to update')
    else
      List.Entries[List.CurrentEntry] := Entry

  end;  { UpdateCurrent }

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

function NumberIn(var List : ListOfRecords) : integer;

  {-------------------------------------------------------------------}
  {  NumberIn returns a function value equal to the current number of }
  {  entries in List.                                                 }
  {                                                                   }
  {  Input parameters:                                                }
  {    List - the list whose number of entries is to be returned      }
  {      Note: List is a var parameter for efficiency in parameter    }
  {            passing.                                               }
  {                                                                   }
  {  Output parameters: None                                          }
  {                                                                   }
  {  Function value: the current number of entries in List            }
  {-------------------------------------------------------------------}

  begin  { NumberIn }

    NumberIn := List.CurrentSize

  end;  { NumberIn }

{*********************************************************************}

procedure GetParameters (var FishInitial : integer; var FishBreed : integer;
                var FishLimit : integer; var SharkInitial : integer;
                var SharkBreed : integer; var SharkStarve : integer);

  {-------------------------------------------------------------------}
  {  GetParameters is called once at the beginning of the simulation. }
  {  Its purpose is to get from the user the parameter values that    }
  {  are to be used in the simulation.  For each parameter, a         }
  {  message is displayed that gives a description of the parameter,  }
  {  the allowable range of values it can take, and its default       }
  {  value.  The user may enter a carriage return to accept the       }
  {  default value.  If the user enters a parameter value to use      }
  {  instead of the default, the value entered is checked against the }
  {  maximum and minimum allowable values.  If the value entered is   }
  {  outside of the allowed range, an error message is displayed;     }
  {  the prompting is repeated until a valid parameter value is       }
  {  obtained.                                                        }
  {                                                                   }
  {  After all of the parameter values are obtained, GetParameters    }
  {  clears the terminal screen and displays a header line that shows }
  {  the parameter values being used for the current simulation.      }
  {                                                                   }
  {  An internal procedure GetParam is used to perform the prompting, }
  {  reading of parameter values, and range checking.                 }
  {                                                                   }
  {  Input parameters: None                                           }
  {                                                                   }
  {  Output parameters:                                               }
  {    FishInitial - the initial number of fish                       }
  {    FishBreed - the age at which a fish breeds                     }
  {    FishLimit - the maximum number of fish                         }
  {    SharkInitial - the initial number of sharks                    }
  {    SharkBreed - the age at which a shark breeds                   }
  {    SharkStarve - the length of time a shark can go without eating }
  {-------------------------------------------------------------------}

  type
    PromptString = packed array[1..30] of char;

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

  procedure GetParam (Prompt : PromptString; Minimum : integer;
                      Maximum : integer; Default : integer;
                      var ParameterValue : integer);

  {-------------------------------------------------------------------}
  {  GetParam prompts the user with a parameter description and the   }
  {  allowed range of values and default value for that parameter.    }
  {  It then reads input from the terminal.  If the user enters a     }
  {  carriage return, GetParam sets the parameter value to its        }
  {  default.  If the user enters a value, GetParam checks the value  }
  {  entered against the allowed range.  The prompting and reading    }
  {  are repeated, if necessary, until a valid parameter value is     }
  {  obtained.                                                        }
  {                                                                   }
  {  Input parameters:                                                }
  {    Prompt - the parameter description to be displayed as a prompt }
  {    Minimum - the minimum value for the parameter                  }
  {    Maximum - the maximum value for the parameter                  }
  {    Default - the default value for the parameter                  }
  {                                                                   }
  {  Output parameters:                                               }
  {    ParameterValue - the value obtained from the user, or the      }
  {                     default if no value was entered               }
  {-------------------------------------------------------------------}

    var
      ValidParameter : boolean;    { a valid parameter value was obtained }
      Bell : char;                 { character that causes the terminal
                                     bell to sound }

    begin

      { if ASCII codes are being used, the character with ordinal value 7
        will ring the terminal bell }

      Bell := chr(7);

      { display the prompt string and read input from the terminal, repeating
        until a valid parameter value is obtained }

      repeat
        write(Prompt,'(',Minimum:3,' - ',Maximum:3,')  [',Default:3,']  :');
        if eoln(input) then
          begin

            { user entered a carriage return -- accept the default value }

            readln;
            ParameterValue := Default;
            ValidParameter := true
          end  { if default value accepted }
        else
          begin

            { user entered a value -- read it and check it for validity }

            readln(ParameterValue);
            if (ParameterValue < Minimum) or (ParameterValue > Maximum) then
              begin
                ValidParameter := false;
                write(Bell)
              end  { if value entered was out of range }
            else
              ValidParameter := true
          end  { else -- user entered a parameter value }
      until ValidParameter

    end;  { GetParam }

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

  begin  { GetParameters }

    { clear the terminal screen and display instructions }

    ClearScreen;
    writeln('               WaTor Parameter Selection');
    writeln;
    writeln('Enter desired values for each parameter (range of possible');
    writeln('values is given in parentheses), or press return to accept the');
    writeln('default value given in [...].');
    writeln;

    { call GetParam to get a value for each parameter }

    GetParam('FI: Initial number of fish    ',
             1,999,FishInitialDefault,FishInitial);
    GetParam('FB: Fish breeding age         ',
             1,100,FishBreedDefault,FishBreed);
    GetParam('FL: Limit on number of fish   ',
             FishInitial,999,FishLimitDefault,FishLimit);
    GetParam('SI: Initial number of sharks  ',
             1,999,SharkInitialDefault,SharkInitial);
    GetParam('SB: Shark breeding age        ',
             1,100,SharkBreedDefault,SharkBreed);
    GetParam('SS: Shark starvation time     ',
             1,100,SharkStarveDefault,SharkStarve);

    { display the header message showing parameter values being used }

    ClearScreen;
    writeln('FI = ',FishInitial:3,'     FB = ',FishBreed:3,
            '     FL = ',FishLimit:3,'     SI = ',SharkInitial:3,
            '     SB = ',SharkBreed:3,'     SS = ',SharkStarve:3);

  end;  { GetParameters }

{*********************************************************************}

procedure Initialize(var Ocean : CellArray; var OldOcean : CellArray;
                var FishList : ListOfRecords; var SharkList : ListOfRecords);

  {-------------------------------------------------------------------}
  {  Initialize is called after the parameter values are obtained.    }
  {  It places the initial number of sharks and fish (given by the    }
  {  parameters FishInitial and SharkInitial) at random into Ocean    }
  {  and sets OldOcean to blanks.  Information about the sharks and   }
  {  fish generated is stored in the lists SharkList and FishList.    }
  {                                                                   }
  {  Input parameters: None                                           }
  {                                                                   }
  {  Output parameters:                                               }
  {    Ocean - initialized with specified numbers of sharks and fish  }
  {    OldOcean - set to blanks                                       }
  {    FishList - list containing information about fish generated    }
  {    SharkList - list containing information about sharks generated }
  {-------------------------------------------------------------------}

  var
    Row : RowSubscript;       { Row and Col are used as subscripts }
    Col : ColSubscript;       {    for Ocean and OldOcean          }
    NewRow : RowSubscript;    { NewRow and NewCol indicate the position }
    NewCol : ColSubscript;    {    to generate a new shark or fish      }
    Count : integer;          { number of sharks or fish generated so far }
    Fish : ListEntry;         { fish being generated }
    Shark : ListEntry;        { shark being generated }

  begin  { Initialize }

    { initialize the random number generator }

    InitRandom(0);

    { set all cells in Ocean and OldOcean to blank }

    for Row := 1 to CellRows do
      for Col := 1 to CellCols do
        begin
          Ocean[Row,Col] := ' ';
          OldOcean[Row,Col] := ' '
        end;  { for }

    { initialize fish list and generate initial number of fish }

    InitList(FishList,FishLimit);
    for Count := 1 to FishInitial do
      begin

        { select cells of Ocean at random until a blank cell is found  --
          note that RandomNumber is a function that generates a different
          random number each time it is called }

        repeat
          NewRow := trunc(RandomNumber * CellRows) + 1;
          NewCol := trunc(RandomNumber * CellCols) + 1
        until Ocean[NewRow,NewCol] = ' ';

        { place fish in the ocean and enter it in FishList, setting its age
          at random in the allowed range }

        Ocean[NewRow,NewCol] := FishSymbol;
        Fish.Row := NewRow;
        Fish.Col := NewCol;
        Fish.Age := trunc(RandomNumber * FishBreed);
        AddNew(FishList,Fish)
      end;  { for }

    { initialize shark list and generate initial number of sharks }

    InitList(SharkList,MaximumListSize);
    for Count := 1 to SharkInitial do
      begin

        { select cells of Ocean at random until a blank cell is found  --
          note that RandomNumber is a function that generates a different
          random number each time it is called }

        repeat
          NewRow := trunc(RandomNumber * CellRows) + 1;
          NewCol := trunc(RandomNumber * CellCols) + 1
        until Ocean[NewRow,NewCol] = ' ';

        { place shark in the ocean and enter it in SharkList, setting its age
          and time since eating at random in the allowed ranges }

        Ocean[NewRow,NewCol] := SharkSymbol;
        Shark.Row := NewRow;
        Shark.Col := NewCol;
        Shark.Age := trunc(RandomNumber * SharkBreed);
        Shark.TimeSinceFed := trunc(RandomNumber * SharkStarve);
        AddNew(SharkList,Shark)
      end  { for }

  end;  { Initialize }

{*********************************************************************}

procedure FindMove (var Ocean : CellArray; OldRow : RowSubscript;
               OldCol : ColSubscript; Target : char; var MoveFound : boolean;
               var NewRow : RowSubscript; var NewCol : ColSubscript);

  {-------------------------------------------------------------------}
  {  FindMove finds a cell in Ocean that is adjacent to the cell      }
  {  at [OldRow,OldCol] and contains the character specified in       }
  {  Target.  If such a cell exists, the parameter MoveFound is set   }
  {  to true and the location of the cell is returned in the          }
  {  parameters NewRow and NewCol. If no such cell exists, MoveFound  }
  {  is set to false and NewRow and NewCol are unchanged.  If more    }
  {  than one such cell exists, one of them is selected at random.    }
  {                                                                   }
  {  In examining cells adjacent to [OldRow,OldCol], the left and     }
  {  right edges of the ocean are considered to be adjacent, and the  }
  {  top and bottom edges are considered to be adjacent.              }
  {                                                                   }
  {  FindMove is used to find moves for sharks and fish, selecting a  }
  {  move at random if more than one is available.  For example, it   }
  {  is called with Target = ' ' to find an empty cell for a fish to  }
  {  swim to; it is called with Target = FishSymbol to find an        }
  {  adjacent fish for a shark to eat.                                }
  {                                                                   }
  {  Input parameters:                                                }
  {    Ocean - the current generation of sharks and fish              }
  {        Note: Ocean is a var parameter for efficiency in           }
  {              parameter passing                                    }
  {    OldRow,OldCol - the row and column of the cell from which a    }
  {                    move is to be found                            }
  {    Target - the character to be sought in an adjacent cell        }
  {                                                                   }
  {  Output parameters:                                               }
  {    MoveFound - true if a move is found, and false otherwise       }
  {    NewRow,NewCol - the row and column of the cell selected for    }
  {                    the move.  If no move is found, NewRow and     }
  {                    NewCol are unchanged.                          }
  {-------------------------------------------------------------------}

  type
    Move = record                     { entry in array of possible moves }
             MoveRow : RowSubscript;    { MoveRow and MoveCol are the row and }
             MoveCol : ColSubscript     {    column of a possible move        }
           end;

  var
    PossibleMoves : array[1..8] of Move; { possible moves that are found }
    NumberOfMoves : integer;             { number of possible moves found }

    TryRow : integer;                    { the row and column of a cell to be }
    TryCol : integer;                    {    examined, before adjusting for  }
                                         {    the edges of the ocean          }

    Row : RowSubscript;                  { the row and column of a cell to be }
    Col : ColSubscript;                  {    examined, after adjusting for   }
                                         {    the edges of the ocean          }

    I : integer;                         { subscript for PossibleMoves }

  begin  { FindMove }

    NumberOfMoves := 0;

    { examine cells in the neighborhood of the cell specified }

    for TryRow := OldRow - 1 to OldRow + 1 do
      for TryCol := OldCol - 1 to OldCol + 1 do
        begin

          { adjust row and column numbers for the edges of the ocean }

          if TryRow < 1 then Row := CellRows
            else if TryRow > CellRows then Row := 1
            else Row := TryRow;
          if TryCol < 1 then Col := CellCols
            else if TryCol > CellCols then Col := 1
            else Col := TryCol;

          { if the cell being examined is not the original cell, and contains
            the target being sought, enter it in the array of possible moves }

          if (Row <> OldRow) or (Col <> OldCol) then
            if Ocean[Row,Col] = Target then
              begin
                NumberOfMoves := NumberOfMoves + 1;
                PossibleMoves[NumberOfMoves].MoveRow := Row;
                PossibleMoves[NumberOfMoves].MoveCol := Col
              end  { if }
        end;

    { if the array contains no possible moves, set MoveFound to false.
      otherwise, set MoveFound to true and select a move at random from the
      array.  note that RandomNumber is a function that generates a
      different number each time it is called }

    if NumberOfMoves = 0 then
      MoveFound := false
    else
      begin
        MoveFound := true;
        I := trunc(RandomNumber * NumberOfMoves) + 1;
        NewRow := PossibleMoves[I].MoveRow;
        NewCol := PossibleMoves[I].MoveCol
      end  { else }

  end;  { FindMove }

{*********************************************************************}

procedure MoveFish(var Ocean : CellArray; var FishList : ListOfRecords);

  {-------------------------------------------------------------------}
  {  MoveFish simulates the swimming and breeding of fish during one  }
  {  chronon of the simulation.                                       }
  {                                                                   }
  {  For each fish, FindMove is called to search for an empty cell    }
  {  that it can swim to.  If such a cell is found, the fish is moved }
  {  to that location.  If the fish being moved is of breeding age,   }
  {  and the number of fish is not at the maximum allowed, the fish   }
  {  produces an offspring by cloning.  The offspring is added to the }
  {  end of the list of fish, and temporarily remains at the original }
  {  location of the parent.  (It will be examined for a possible     }
  {  swimming move later during the scanning of the list.)  The ages  }
  {  of both the parent and the offspring are set to 0.               }
  {                                                                   }
  {  If there is no adjacent empty cell, the fish does not swim or    }
  {  breed.                                                           }
  {                                                                   }
  {  Input parameters:                                                }
  {    Ocean - the current generation of sharks and fish              }
  {    FishList - the list of fish currently in existence             }
  {                                                                   }
  {  Output parameters:                                               }
  {    Ocean - the current generation of sharks and fish, reflecting  }
  {            the swimming and breeding of fish                      }
  {    FishList - the list of fish after swimming and breeding        }
  {-------------------------------------------------------------------}

  var
    Fish : ListEntry;             { the fish currently being examined }
    MoveFound : boolean;          { a move was found for the fish }
    NewRow : RowSubscript;        { the row and column of the cell (if any) }
    NewCol : ColSubscript;        {    to which the fish will move          }
    NewFish : ListEntry;          { a new fish being generated by breeding }
    EndOfFish : boolean;          { all fish have been examined }

  begin  { MoveFish }

    { scan the list of fish, generating swimming and breeding moves }

    GetFirst(FishList,Fish,EndOfFish);
    while not EndOfFish do
      begin

        { add one chronon to the age of the fish }

        Fish.Age := Fish.Age + 1;

        { find a swimming move for the fish, if one exists }

        FindMove(Ocean,Fish.Row,Fish.Col,' ',MoveFound,NewRow,NewCol);
        if MoveFound then
          begin

            { a swimming move was found.  check to see if the fish can breed }

            if (Fish.Age >= FishBreed)
                      and (NumberIn(FishList) < FishLimit) then
              begin

                { the fish can breed.  generate the offspring and add it to the
                  list, setting the ages of both fish to 0.  the offspring
                  remains at the original location of the parent }

                NewFish := Fish;
                NewFish.Age := 0;
                AddNew(FishList,NewFish);
                Fish.Age := 0
              end  { if }
            else

              { the fish cannot breed; its original location becomes empty }

              Ocean[Fish.Row,Fish.Col] := ' ';

            { move the fish to its new location }

            Ocean[NewRow,NewCol] := FishSymbol;
            Fish.Row := NewRow;
            Fish.Col := NewCol
          end;  { if swimming move was found }

        { update current fish and get the next fish in the list }

        UpdateCurrent(FishList,Fish);
        GetNext(FishList,Fish,EndOfFish)
      end  { while }

  end;  { MoveFish }

{*********************************************************************}

procedure MoveSharks(var Ocean : CellArray; var SharkList : ListOfRecords;
                     var FishList : ListOfRecords);

  {-------------------------------------------------------------------}
  {  MoveSharks simulates the hunting, swimming, and breeding of      }
  {  sharks during one chronon of the simulation.  It first scans the }
  {  list of sharks, removing the entries for sharks that have        }
  {  starved during the previous chronon.  It then scans the list a   }
  {  second time, generating the hunting, swimming, and breeding      }
  {  moves for the sharks that remain.  After all of the sharks have  }
  {  moved, the fish that were eaten are removed from the list of     }
  {  fish.                                                            }
  {                                                                   }
  {  For each shark, FindMove is called to search for an adjacent     }
  {  cell that contains a fish.  If such a cell is found, the shark   }
  {  swims to that location and eats the fish.  If no adjacent fish   }
  {  is found, FindMove is called again to search for an empty cell   }
  {  that the shark can swim to.  If the shark being moved is of      }
  {  breeding age, it produces an offspring by cloning.  The          }
  {  offspring is added to the end of the list of sharks, and         }
  {  temporarily remains at the original location of the parent.      }
  {  (It will be examined for possible hunting and swimming moves     }
  {  later during the scanning of the list.)  The ages  of both the   }
  {  parent and the offspring are set to 0.                           }
  {                                                                   }
  {  If there is no adjacent empty cell, the shark does not swim or   }
  {  breed.                                                           }
  {                                                                   }
  {  Input parameters:                                                }
  {    Ocean - the current generation of sharks and fish              }
  {    SharkList - the list of sharks currently in existence          }
  {    FishList - the list of fish currently in existence             }
  {                                                                   }
  {  Output parameters:                                               }
  {    Ocean - the current generation of sharks and fish, reflecting  }
  {            the hunting, swimming, and breeding of sharks          }
  {    SharkList - the list of sharks after swimming and breeding     }
  {    FishList - the list of fish, after removing those eaten by     }
  {               sharks                                              }
  {-------------------------------------------------------------------}

  var
    Shark : ListEntry;            { the shark currently being examined }
    MoveFound : boolean;          { a move was found for the shark }
    NewRow : RowSubscript;        { the row and column of the cell (if any) }
    NewCol : ColSubscript;        {    to which the shark will move         }
    NewShark : ListEntry;         { a new shark being generated by breeding }
    EndOfSharks : boolean;        { all sharks have been examined }
    Fish : ListEntry;             { the fish currently being examined }
    EndOfFish : boolean;          { all fish have been examined }

  begin  { MoveSharks }

    { scan the list of sharks, removing those that have starved }

    GetFirst(SharkList,Shark,EndOfSharks);
    while not EndOfSharks do
      begin
        if Shark.TimeSinceFed > SharkStarve then
          begin
            Ocean[Shark.Row,Shark.Col] := ' ';
            DeleteCurrent(SharkList)
          end;  { if shark has starved }
        GetNext(SharkList,Shark,EndOfSharks)
      end;  { while }

    { scan the list again, generating hunting, swimming, and breeding moves }

    GetFirst(SharkList,Shark,EndOfSharks);
    while not EndOfSharks do
      begin

        { add one chronon to the age of the shark and the time since fed }

        Shark.Age := Shark.Age + 1;
        Shark.TimeSinceFed := Shark.TimeSinceFed + 1;

        { find a fish for the shark to eat, if one exists }

        FindMove(Ocean,Shark.Row,Shark.Col,FishSymbol,MoveFound,NewRow,NewCol);

        { if a fish is found, set time since fed to 0; otherwise, find a
          swimming move for the shark }

        if MoveFound then
          Shark.TimeSinceFed := 0
        else
          FindMove(Ocean,Shark.Row,Shark.Col,' ',MoveFound,NewRow,NewCol);
        if MoveFound then
          begin

            { a hunting or swimming move was found.  check to see if the
              shark can breed }

            if (Shark.Age >= SharkBreed) then
              begin

                { the shark can breed.  generate the offspring and add it to
                  the list, setting the ages of both sharks to 0.  the
                  offspring remains at the original location of the parent }

                NewShark := Shark;
                NewShark.Age := 0;
                AddNew(SharkList,NewShark);
                Shark.Age := 0
              end  { if }
            else

              { the shark cannot breed; its original location becomes empty }

              Ocean[Shark.Row,Shark.Col] := ' ';

            { move the shark to its new location }

            Ocean[NewRow,NewCol] := SharkSymbol;
            Shark.Row := NewRow;
            Shark.Col := NewCol
          end;  { if move was found }

        { update current shark and get next shark in the list }

        UpdateCurrent(SharkList,Shark);
        GetNext(SharkList,Shark,EndOfSharks)
      end;  { while }

    { scan the list of fish, removing those eaten by sharks }

    GetFirst(FishList,Fish,EndOfFish);
    while not EndOfFish do
      begin

        { if the position of the fish is now occupied by a shark, remove it
          from the list }

        if Ocean[Fish.Row,Fish.Col] = SharkSymbol then
          DeleteCurrent(FishList);
        GetNext(FishList,Fish,EndOfFish)
      end { while }

  end;  { MoveSharks }

{*********************************************************************}

procedure DisplayOcean(var Ocean : CellArray; var OldOcean : CellArray;
               var FishList : ListOfRecords; var SharkList : ListOfRecords);

  {-------------------------------------------------------------------}
  {  DisplayOcean is called to display the current ocean on the       }
  {  terminal.  It calls the procedure Display (from the file         }
  {  "display.i") to display the contents of Ocean, beginning on      }
  {  line 3 of the screen.  It then displays the total numbers of     }
  {  fish and sharks on a line below the ocean.                       }
  {                                                                   }
  {  The procedure Display attempts to avoid re-displaying parts of   }
  {  the ocean that have not changed from the previous version        }
  {  (OldOcean).  On the first call to DisplayOcean, it is assumed    }
  {  that the terminal screen has been cleared to blanks and that     }
  {  OldOcean contains all blanks.                                    }
  {                                                                   }
  {  After Ocean is displayed, its contents are copied to OldOcean.   }
  {                                                                   }
  {  Input parameters:                                                }
  {    Ocean - the current ocean to be displayed                      }
  {    OldOcean - the previous version of the ocean, or blanks if     }
  {               this is the first version displayed                 }
  {    FishList - the list of fish currently in existence             }
  {    SharkList - the list of sharks currently in existence          }
  {                                                                   }
  {      Note: All parameters are var parameters for efficiency in    }
  {            parameter passing.                                     }
  {                                                                   }
  {  Output parameters:                                               }
  {    OldOcean - set to the contents of Ocean                        }
  {-------------------------------------------------------------------}

  begin  { DisplayOcean }

    { display the current ocean and totals of fish and sharks }

    Display(Ocean,OldOcean,3);
    MoveCursor(22,1);
    writeln('Fish = ',NumberIn(FishList):4,
            '            Sharks = ',NumberIn(SharkList):4);

    { copy ocean displayed to previous ocean }

    OldOcean := Ocean

  end;  { DisplayOcean }

{*********************************************************************}

procedure AskUser(var WantsToContinue : boolean);

  {-------------------------------------------------------------------}
  {  AskUser is called after the simulation has run for a number      }
  {  of chronons given by the constant PauseAfter.  It asks the       }
  {  user if he or she wants to continue the simulation, and sets the }
  {  parameter WantsToContinue according to the answer received.      }
  {                                                                   }
  {  If the first character entered by the user is 'y' or 'Y',        }
  {  WantsToContinue is set to true; if the first character is        }
  {  'n' or 'N', WantsToContinue is set to false.  The prompting is   }
  {  repeated if necessary until one of these two results is obtained.}
  {                                                                   }
  {  Input parameters: None                                           }
  {                                                                   }
  {  Output parameters:                                               }
  {    WantsToContinue - true if the user wants to continue the       }
  {                      simulation                                   }
  {-------------------------------------------------------------------}

  var
    InputCharacter : char;     { the first character entered by the user }

  begin  { AskUser }

    { position the cursor on a line below the ocean display.  prompt and
      read from the terminal until one of the characters 'y','Y','n','N'
      is obtained }

    repeat
      MoveCursor(23,1);
      write('Continue? (y/n) : ');
      readln(InputCharacter)
    until InputCharacter in ['y','Y','n','N'];

    { set WantsToContinue according to the input received }

    WantsToContinue := (InputCharacter in ['y','Y'])

  end;  { AskUser }

{*********************************************************************}

procedure CleanUp;

  {-------------------------------------------------------------------}
  {  CleanUp is called once at the end of the program.  It moves the  }
  {  cursor to a line below the ocean display, and waits for the user }
  {  to enter a carriage return before exiting the program.           }
  {-------------------------------------------------------------------}

  begin  { CleanUp }

    MoveCursor(23,1);
    ClearLine;
    write('Press return to exit ');
    readln

  end;  { CleanUp }

{*********************************************************************}

begin  { main program }

  { get the simulation parameters from the user and initialize the ocean }

  GetParameters(FishInitial,FishBreed,FishLimit,SharkInitial,SharkBreed,
                SharkStarve);
  Initialize(Ocean,OldOcean,FishList,SharkList);

  { run the simulation until the number of sharks or the number of fish
    falls to zero, or until the user does not want to continue }

  CycleNumber := 0;
  WantsToContinue := true;
  repeat  { until the user wants to stop, or all the sharks or fish are gone }

    { move the fish and display the result }

    MoveFish(Ocean,FishList);
    DisplayOcean(Ocean,OldOcean,FishList,SharkList);

    { move the sharks and display the result }

    MoveSharks(Ocean,SharkList,FishList);
    DisplayOcean(Ocean,OldOcean,FishList,SharkList);

    { ask the user for instructions after each PauseAfter chronons }

    CycleNumber := CycleNumber + 1;
    if (CycleNumber mod PauseAfter) = 0 then
        AskUser(WantsToContinue)

  until (WantsToContinue = false) or (NumberIn(FishList) = 0)
         or (NumberIn(SharkList) = 0);

  CleanUp

end.
