program Index3(input,output,IDATA,TERMS,INDEXLST);

{---------------------------------------------------------------------}
{  This program searches a file IDATA for occurrences of index        }
{  words listed in the file TERMS.  A list of the pages on which each }
{  index word was found is written to the file INDEXLST. A maximum of }
{  20 words can be searched for in one execution of the program.      }
{  Pages in IDATA  are identified by an entry  "@ n"  at the          }
{  beginning of a page, where  "n"  is the page number.  The largest  }
{  page number that can be processed in this version of the program   }
{  is 31.  If IDATA contains pages numbered higher than 31, they      }
{  are ignored and a warning message is displayed.                    }
{                                                                     }
{  Index words may contain any of the characters 'a' through 'z'      }
{  and apostrophe.  The maximum length for an index word is given by  }
{  the constant MaximumWordLength.  Index words in TERMS that exceed  }
{  the maximum length are truncated with no error message.  If TERMS  }
{  contains too many index words, a warning message is displayed. The }
{  case of letters is not considered when searching for index words.  }
{                                                                     }
{  Input:  Index words to be searched for are contained in the file   }
{          TERMS, one word per line.  The text to be searched, with   }
{          page numbers identified, is read from file IDATA.          }
{                                                                     }
{  Output: A list of page numbers on which each specified index word  }
{          occurs is written to the file INDEXLST.                    }
{                                                                     }
{  This program contains non-standard Pascal statements to support    }
{  Turbo Pascal and Berkeley Pascal versions.  Non-standard Pascal    }
{  statements appear in the following locations:                      }
{    1. in the compiler directives to include "cases.i"               }
{    2. in procedures Initialize and CleanUp                          }
{  These non-standard statements are preceded by a comment that       }
{  begins with '## N'.                                                }
{                                                                     }
{  Modification history:                                              }
{    This program performs the same indexing operations as the        }
{    previous version (Index2).  However, the list of page references }
{    for each index term is implemented as an abstract data type      }
{    ListOfValues.  This data type is operated on by the procedures   }
{    InitializeList and AddToList, and the function  IsInList.        }
{                                                                     }
{    The index produced by this program is written to a file          }
{    INDEXLST, instead of being displayed on the terminal as in the   }
{    previous version.                                                }
{---------------------------------------------------------------------}

const
  MaximumListValue = 31;                { required for data type ListOfValues }
  Quote = '''';                         { single quote (apostrophe) }
  PageFlag = '@';                       { marks the page number in IDATA }
  MaximumWordLength = 20;               { longest word to be considered }
  BlankWord = '                    ';   { blanks to maximum word length }
  MaximumIndexWords = 20;               { maximum number of index words }
  MaximumPageNumber = MaximumListValue; { largest allowable page number }

type
  ListOfValues = set of 0..MaximumListValue;
  AlphaString = packed array [1..MaximumWordLength] of char;
  CharacterSet = set of char;

  IndexEntry = record                      { entry in table of index words }
                 Term : AlphaString;         { index word }
                 PagesFound : ListOfValues   { pages on which word appears }
               end;

  IndexType = array[1..MaximumIndexWords] of IndexEntry;

var
  LegalCharacters : CharacterSet; { characters legal in index word }

     { Note: LegalCharacters is a constant during the execution of this
       program.  It is set by the procedure Initialize to the value
       ['a'..'z',Quote].  As usual with constants, LegalCharacters is used
       by other procedures without being passed as a parameter. }

  IDATA : text;                   { file containing text to be searched }
  TERMS : text;                   { file containing words to be indexed }
  INDEXLST : text;                { file on which index is to be written }
  TextWord : AlphaString;         { word from IDATA }
  PageNumber : integer;           { current page number in IDATA }
  EndOfText : boolean;            { end of IDATA has been reached }
  IndexTable : IndexType;         { table of index words and references }
  TableSize : integer;            { number of words in index table }
  TableSubscript : integer;       { subscript for IndexTable }

{*********************************************************************}

{---------------------------------------------------------------------}
{  The following compiler directives include the file "cases.i",      }
{  which contains the following functions:                            }
{    Lowercase - convert letters to lower case                        }
{    Uppercase - convert letters to upper case                        }
{                                                                     }
{  Note: Uppercase is not used in this program.                       }
{---------------------------------------------------------------------}

{$I cases.i} { for Turbo Pascal; treated as comment in standard Pascal }

{ ## Non-standard statements -- Berkeley Pascal version }
{
#include "cases.i"
}
{ End Berkeley Pascal version }

{*********************************************************************}

{---------------------------------------------------------------------}
{  The following procedures and functions implement operations on the }
{  abstract data type ListOfValues.  This data type can be used to    }
{  represent a list of integers with values between 0 and             }
{  MaximumListValue.  The following procedures and functions are      }
{  provided:                                                          }
{     InitializeList - initializes a list to contain no values        }
{     AddToList - adds a specified value to a list                    }
{     IsInList - tests whether a specified value is in a list         }
{                                                                     }
{  In this implementation of the type ListOfValues, a list is         }
{  represented as a set of integers with values between 0 and 31.     }
{                                                                     }
{  In order to use this abstract data type, the main program must     }
{  include the constant definition                                    }
{     MaximumListValue = 31                                           }
{  and the type definition                                            }
{     ListOfValues = set of 0..MaximumListValue                       }
{---------------------------------------------------------------------}


{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

procedure InitializeList(var List : ListOfValues);

  {-------------------------------------------------------------------}
  {  InitializeList initializes a list so that it is empty (that is,  }
  {  contains no values).                                             }
  {                                                                   }
  {  Input Parameters: None                                           }
  {                                                                   }
  {  Output Parameters:                                               }
  {    List - the initialized list                                    }
  {-------------------------------------------------------------------}

  begin  { InitializeList }

    List := []

  end;  { InitializeList }

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

procedure AddToList(var List: ListOfValues; Value : integer);

  {-------------------------------------------------------------------}
  {  AddToList adds a new value to a list.  If the value being added  }
  {  is already in the list, the list remains unchanged.  If the      }
  {  value being added is outside of the allowed range of values for  }
  {  the list, an error message is displayed.                         }
  {                                                                   }
  {  Input Parameters:                                                }
  {    List - the list to be processed                                }
  {    Value - the new value to be added to the list                  }
  {                                                                   }
  {  Output Parameters:                                               }
  {    List - the list with the new value added                       }
  {-------------------------------------------------------------------}

  begin  { AddToList }

    if (Value >= 0) and (Value <= MaximumListValue) then
      List := List + [Value]
    else
      writeln('*** AddToList: value ',Value:1,' is out of range.')

  end;  { AddToList }

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

function IsInList(List : ListOfValues; Value : integer) : boolean;

  {-------------------------------------------------------------------}
  {  IsInList tests whether or not a specified value is in a list.    }
  {  If the value is in the list, IsInList returns the Boolean value  }
  {  true; otherwise, it returns the value false.  If the value       }
  {  is already in the list, the list remains unchanged.  If the      }
  {  value being tested is outside of the allowed range of values for }
  {  the list, an error message is displayed.                         }
  {                                                                   }
  {  Input Parameters:                                                }
  {    List - the list to be processed                                }
  {    Value - the value to be sought in the list                     }
  {                                                                   }
  {  Output Parameters: None                                          }
  {                                                                   }
  {  Function value: true if Value is in List; false otherwise        }
  {-------------------------------------------------------------------}

  begin  { IsInList }

    if (Value >= 0) and (Value <= MaximumListValue) then
      begin

        { Value is in range.  set the function value IsInList to true
          if Value is in List, and false otherwise }

        IsInList := (Value in List)

      end  { if Value is in range }
    else
      begin

        { Value is out of range -- display an error message and return
          false as the function value }

        writeln('*** IsInList: value ',Value:1,' is out of range.');
        IsInList := false
      end  { else }

  end;  { IsInList }

{*********************************************************************}

procedure Initialize(var LegalCharacters : CharacterSet);

  {-------------------------------------------------------------------}
  {  Initialize is called once at the beginning of the program.       }
  {  It opens the files IDATA, TERMS, and INDEXLST, and initializes   }
  {  the set of characters considered legal as part of a word.        }
  {                                                                   }
  {  Input Parameters: None                                           }
  {                                                                   }
  {  Output Parameters:                                               }
  {    LegalCharacters - the set of characters legal as part of a     }
  {                      word, initialized to ['a'..'z',Quote]        }
  {-------------------------------------------------------------------}

  begin  { Initialize }

    { ## Non-standard statements -- Turbo Pascal version }

    assign(IDATA,'IDATA');
    assign(TERMS,'TERMS');
    assign(INDEXLST,'INDEXLST');

    { End Turbo Pascal version }

    reset(IDATA);
    reset(TERMS);
    rewrite(INDEXLST);
    LegalCharacters := ['a'..'z',Quote]

  end;  { Initialize }

{*********************************************************************}

procedure GetIndexWords(var IndexTable : IndexType; var TableSize : integer);

  {-------------------------------------------------------------------}
  {  GetIndexWords reads the words to be indexed from the file TERMS  }
  {  and loads them into the IndexTable.  The list of pages on which  }
  {  each term occurs is initialized to empty.  The number of words   }
  {  loaded into IndexTable is stored in TableSize.                   }
  {                                                                   }
  {  An internal procedure ReadIndexWord is used to read index words  }
  {  from TERMS.  Characters in the words read from TERMS are not     }
  {  checked for validity.  However, upper case letters are converted }
  {  to lower case.  If an index word contains more than the maximum  }
  {  allowed number of characters (MaximumWordLength), it is          }
  {  truncated with no error message.  If TERMS contains more than    }
  {  MaximumIndexWords index words, the excess words are discarded    }
  {  and a warning message is displayed on the terminal.              }
  {                                                                   }
  {  Input parameters: None                                           }
  {                                                                   }
  {  Output parameters:                                               }
  {    IndexTable - table of words to be indexed, with lists of page  }
  {            numbers initialized to be empty                        }
  {    TableSize - the number of index words                          }
  {-------------------------------------------------------------------}

  var
    IndexWord : AlphaString;       { current index word }

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

procedure ReadIndexWord(var IndexWord : AlphaString);

  {-------------------------------------------------------------------}
  {  ReadIndexWord reads the next index word from the file TERMS,     }
  {  converting alphabetic characters to lower case.  Index words     }
  {  longer than MaximumWordLength are truncated with no error        }
  {  message.                                                         }
  {                                                                   }
  {  ReadIndexWord is similar to the procedure GetIndexWord in        }
  {  program Index1 (the original version of this program).  However, }
  {  ReadIndexWord reads from the file TERMS instead of from the      }
  {  terminal, and does not perform error checking.                   }
  {                                                                   }
  {  Input parameters: None                                           }
  {                                                                   }
  {  Output parameters:                                               }
  {    IndexWord - the valid index word obtained                      }
  {-------------------------------------------------------------------}

  var
    LengthCounter : integer;             { number of characters read }
    CurrentCharacter : char;             { current character }

  begin  { ReadIndexWord }

    { initialize IndexWord to all blanks and length read to zero }

    LengthCounter := 0;
    IndexWord := BlankWord;

    { read characters from TERMS until end-of-line is detected or
      the maximum word length is reached }

    while (LengthCounter < MaximumWordLength) and (not eoln(TERMS)) do
      begin
        read(TERMS,CurrentCharacter);
        LengthCounter := LengthCounter + 1;
        IndexWord[LengthCounter] := Lowercase(CurrentCharacter);
      end;  { while }

    { skip any unread characters and end-of-line marker }

    readln(TERMS)

  end;  { ReadIndexWord }

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

  begin  { GetIndexWords }

    { read index words from TERMS until end-of-file is detected, or until
      the maximum number of index words is reached }

    TableSize := 0;
    while (not eof(TERMS)) and (TableSize < MaximumIndexWords) do
      begin

        { read an index word from TERMS }

        ReadIndexWord(IndexWord);

        { store index word in table and initialize list of pages }

        TableSize := TableSize + 1;
        IndexTable[TableSize].Term := IndexWord;
        InitializeList(IndexTable[TableSize].PagesFound)
      end;  { while }

    { if end-of-file was not detected, TERMS contains too many index
      words -- display warning message }

    if not eof(TERMS) then
      writeln('*** TERMS contains more than ',MaximumIndexWords:2,
              ' index words -- excess words discarded.')

  end;  { GetIndexWords }

{*********************************************************************}

procedure GetTextWord(var TextWord : AlphaString; var EndOfText : boolean;
                      var PageNumber : integer);

  {-------------------------------------------------------------------}
  {  GetTextWord is called to read the next word from IDATA.  It      }
  {  returns the word read and the page number on which the word      }
  {  appears.  If no word is found (because the end of IDATA has      }
  {  been reached), EndOfText is set to true.                         }
  {                                                                   }
  {  A word read by this procedure consists of a string of characters }
  {  all of which are 'a' through 'z' or apostrophe.  The constant    }
  {  set LegalCharacters is used to make this check.  The maximum     }
  {  length of a word is given by the constant MaximumWordLength.     }
  {  Words longer than this length are truncated to the maximum       }
  {  length with no error message.  Alphabetic characters in TextWord }
  {  are converted to lower case.                                     }
  {                                                                   }
  {  An internal procedure FindNextWord is used to locate the         }
  {  beginning of the word to be read (if one is found), and to       }
  {  indicate the page number on which the word appears.              }
  {                                                                   }
  {  Input parameters:                                                }
  {    PageNumber - the page number on which the previous text word   }
  {                 was found, or 0 if this is the first call to      }
  {                 GetTextWord                                       }
  {                                                                   }
  {  Output parameters:                                               }
  {    TextWord - the word read from IDATA (unless end-of-file        }
  {               was detected), with alphabetic characters converted }
  {               to lower case                                       }
  {    EndOfText - set to true if the end of IDATA was found          }
  {                (and no word was read); set to false otherwise     }
  {    PageNumber - the number of the page on which TextWord was      }
  {                 found (unless end-of-file was detected)           }
  {-------------------------------------------------------------------}

  var
    LengthCounter : integer;    { number of characters in the word so far }
    CurrentCharacter : char;    { current character }

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

  procedure FindNextWord(var CurrentCharacter : char; var EndOfText : boolean;
                         var PageNumber : integer);

  {-------------------------------------------------------------------}
  {  FindNextWord locates the beginning of the next word to be read   }
  {  from IDATA, unless end-of-file is detected first.  It scans      }
  {  IDATA until it finds a character that begins a legal text        }
  {  word (one that is in the constant set LegalCharacters).  This    }
  {  character is converted to lower case (if it is alphabetic) and   }
  {  returned in the parameter CurrentCharacter.  If  the end of      }
  {  IDATA is reached before such a character is found, EndOfText is  }
  {  set to true; otherwise, EndOfText is set to false.               }
  {                                                                   }
  {  If a page number flag ( "@ n" , where "n" is the page number) is }
  {  found while scanning for the next text word, the value of "n" is }
  {  stored in the parameter PageNumber.                              }
  {                                                                   }
  {  Input parameters:                                                }
  {    PageNumber - the current page number in IDATA                  }
  {                                                                   }
  {  Output parameters:                                               }
  {    CurrentCharacter - the character that begins the next text     }
  {                word (unless end-of-file was detected), converted  }
  {                to lower case if alphabetic                        }
  {    EndOfText - set to true if the end of IDATA was reached        }
  {                without finding a word; set to false otherwise     }
  {    PageNumber - the current page number in IDATA; equal to the    }
  {                 value of the input parameter unless a page number }
  {                 flag was found while reading IDATA                }
  {                                                                   }
  {  Side effect:  the file IDATA is positioned after the first       }
  {                character of the next word (unless end-of-file     }
  {                is detected)                                       }
  {-------------------------------------------------------------------}

    var
      Scanning : boolean;     { searching for beginning of next word }

    begin  { FindNextWord }

      { read IDATA until the beginning of a word is found, or until
        end-of-file is detected }

      Scanning := true;
      while (Scanning) and (not eof(IDATA)) do
        begin
          if eoln(IDATA) then

            { if end-of-line is reached, go to next line }

            readln(IDATA)
          else
            begin
              read(IDATA,CurrentCharacter);
              if CurrentCharacter = PageFlag then

                { "@" was found -- read new page number }

                read(IDATA,PageNumber)
              else
                begin

                  { stop scanning if CurrentCharacter begins a word }

                  CurrentCharacter := Lowercase(CurrentCharacter);
                  if CurrentCharacter in LegalCharacters then
                    Scanning := false
                end  { if not PageFlag }
            end  { if not eoln }
        end;  { while }

      { set EndOfText according to whether end-of-file was detected }

      EndOfText := eof(IDATA)

    end;  { FindNextWord }

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

  begin  { GetTextWord }

    { find the beginning of the next word (if any) and the page number
      on which it occurs }

    FindNextWord(CurrentCharacter,EndOfText,PageNumber);

    { if end-of-file was not reached, read the next word }

    if not EndOfText then
      begin

        { initialize text word to all blanks }

        LengthCounter := 0;
        TextWord := BlankWord;

        { read characters until one is found that is not in LegalCharacters.
          store the characters read in TextWord, with letters converted to
          lower case.  discard characters in excess of maximum word length }

        while CurrentCharacter in LegalCharacters do
          begin
            if LengthCounter < MaximumWordLength then
              begin
                LengthCounter := LengthCounter + 1;
                TextWord[LengthCounter] := CurrentCharacter
              end;  { if }
            read(IDATA,CurrentCharacter);
            CurrentCharacter := Lowercase(CurrentCharacter)
          end  { while }
      end  { if not EndOfText }

  end;  { GetTextWord }

{*********************************************************************}

procedure SearchTable(TextWord : AlphaString; IndexTable : IndexType;
                      TableSize : integer; var TableSubscript : integer);

  {-------------------------------------------------------------------}
  {  SearchTable searches IndexTable for an entry that matches        }
  {  TextWord, using a linear search starting at the beginning of the }
  {  table.  If a matching entry is found, TableSubscript is set to   }
  {  the subscript of the matching entry; if no matching entry is     }
  {  found, TableSubscript is set to 0.                               }
  {                                                                   }
  {  Input parameters:                                                }
  {    TextWord - the word to be searched for                         }
  {    IndexTable - the index table                                   }
  {    TableSize - the number of entries in the index table           }
  {                                                                   }
  {  Output parameters:                                               }
  {    TableSubscript - the subscript of the entry in IndexTable that }
  {                     matches TextWord, or 0 if no matching entry   }
  {                     exists                                        }
  {-------------------------------------------------------------------}

  var
    I : integer;           { used as subscript for IndexTable }
    Found : boolean;       { TextWord has been found in IndexTable }

  begin  { SearchTable }

    { examine entries in IndexTable until a matching entry is found, or
      until the end of the table is reached }

    I := 1;
    Found := false;
    while (not Found) and (I <= TableSize) do
      begin
        if IndexTable[I].Term = TextWord then

          { matching entry has been found }

          Found := true
        else

          { entry does not match -- go on to the next entry }

          I := I + 1
      end;  { while }

    { set TableSubscript according to whether or not a match was found }

    if Found then
      TableSubscript := I
    else
      TableSubscript := 0

  end;  { SearchTable }

{*********************************************************************}

procedure WriteIndexEntry(Entry : IndexEntry);

  {-------------------------------------------------------------------}
  {  WriteIndexEntry is called with each entry in the index table,    }
  {  after the lists of page numbers have been constructed.  It       }
  {  writes to INDEXLST the index word and the list of pages on which }
  {  that word occurs.  The maximum number of page numbers to be      }
  {  written per line is given by the constant MaxNumbersPerLine.     }
  {                                                                   }
  {  Input parameters:                                                }
  {    Entry - the entry in the index table that is to be displayed   }
  {                                                                   }
  {  Output parameters: None                                          }
  {-------------------------------------------------------------------}

  const
    MaxNumbersPerLine = 11;     { maximum page numbers to write per line }

  var
    Page : integer;             { page number }
    NumbersOnLine : integer;    { count of numbers written on current line }

  begin  { WriteIndexEntry }

    { write the index word }

    write(INDEXLST,Entry.Term);
    NumbersOnLine := 0;

    { write each possible page number that is in the list of pages for
      the index word.  count the number of page numbers written }

    for Page := 0 to MaximumPageNumber do
      begin
        if IsInList(Entry.PagesFound,Page) then
          begin

            { if MaxNumbersPerLine page numbers have already been written
              on the current line, start a new line }

            if NumbersOnLine = MaxNumbersPerLine then
              begin
                writeln(INDEXLST);
                write(INDEXLST,BlankWord);
                NumbersOnLine := 0
              end;  { if line is full }

            { write page number and add to count of numbers on line }

            write(INDEXLST,Page:5);
            NumbersOnLine := NumbersOnLine + 1
          end  { if Page in PagesFound }
      end;  { for }

    { write out the last line }

    writeln(INDEXLST)

  end;  { WriteIndexEntry }

{*********************************************************************}

procedure CleanUp;

  {-------------------------------------------------------------------}
  {  CleanUp is called once at the end of the program.  It closes     }
  {  the files IDATA, TERMS, and INDEXLST, if this is required in the }
  {  Pascal version being used, and waits for the user to enter a     }
  {  carriage return before exiting the program.                      }
  {-------------------------------------------------------------------}

  begin  { CleanUp }

    { ## Non-standard statements -- Turbo Pascal version }

    close(IDATA);
    close(TERMS);
    close(INDEXLST);

    { End Turbo Pascal version }

    write('Press return to exit program');
    readln

  end;  { CleanUp }

{*********************************************************************}

begin  { main program }

  Initialize(LegalCharacters);

  { read index words from TERMS into index table }

  TableSize := 0;
  GetIndexWords(IndexTable,TableSize);
  PageNumber := 0;

  { get first word from IDATA.  if end-of-file is detected, write an
    error message. }

  GetTextWord(TextWord,EndOfText,PageNumber);
  if EndOfText then
    writeln('*** No text data supplied.')
  else
    writeln('Scanning text...');

  { read words from IDATA until end-of-file is reached, or until a
    page number greater than MaximumPageNumber is found }

  while (not EndOfText) and (PageNumber <= MaximumPageNumber) do
    begin

      { search the index table for each word read from IDATA.  if
        a matching entry is found, add the current page number to the
        list of pages on which the index word occurs }

      SearchTable(TextWord,IndexTable,TableSize,TableSubscript);
      if TableSubscript <> 0 then
        AddToList(IndexTable[TableSubscript].PagesFound,PageNumber);
      GetTextWord(TextWord,EndOfText,PageNumber)
    end;  { while }

  { if a page number was too large, display a warning message }

  if PageNumber > MaximumPageNumber then
    writeln('*** Pages in IDATA numbered higher than ',
            MaximumPageNumber:2,' have been ignored.');

  { write the index words and page numbers to the file INDEXLST }

  writeln('Writing index...');
  writeln(INDEXLST,'Index Terms         Pages Found');
  writeln(INDEXLST);
  for TableSubscript := 1 to TableSize do
    WriteIndexEntry(IndexTable[TableSubscript]);
  CleanUp

end.
