program Anagram1(input,output,WORDLIST);

{---------------------------------------------------------------------}
{  This program uses a dictionary to find anagrams for a target word  }
{  entered by the user, and displays these anagrams on the terminal.  }
{  Two types of anagrams are recognized.  Full anagrams are words     }
{  from the dictionary that contain exactly the same letters as the   }
{  target word.  Partial anagrams are words from the dictionary that  }
{  can be formed from the letters in the target word, but do not use  }
{  all of those letters.  For example, if the target word is "stop",  }
{  then "post" and "spot" are full anagrams; "pot" and "to" are       }
{  partial anagrams.  Full anagrams are preceded by "*" when they     }
{  are displayed.                                                     }
{                                                                     }
{  The case of letters is not considered when searching for anagrams, }
{  and non-alphabetic characters are ignored.  Thus, for example,     }
{  "sit" would be considered a full anagram for "It's".               }
{                                                                     }
{  The maximum length for a target word is 20 characters.  The target }
{  word may contain blanks (thus making it appear to be more than one }
{  word).  However, all alphabetic characters entered are considered  }
{  to be part of the same word when searching for anagrams.  For      }
{  example, if the user entered "It's me", the program would find     }
{  "smite" as a full anagram.                                         }
{                                                                     }
{  Input:  The target word is entered from the terminal.  Dictionary  }
{          words are read from the file WORDLIST.                     }
{                                                                     }
{  Output: The anagrams found (both full and partial) are displayed   }
{          on the terminal.                                           }
{                                                                     }
{  This program contains non-standard Pascal statements to support    }
{  Turbo Pascal and Berkeley Pascal versions.  Non-standard Pascal    }
{  statements appear in the following locations:                      }
{    1. in the compiler directives to include "cases.i"               }
{    2. in procedures Initialize and CleanUp                          }
{  These non-standard statements are preceded by a comment that       }
{  begins with '## N'.                                                }
{---------------------------------------------------------------------}

const
  MaxWordLength = 20;        { maximum length of target or dictionary word }

type
  AnagramType = (FullAnagram,PartialAnagram,NotAnagram);
  WordArray = array[1..MaxWordLength] of char;
  CountOfLetters = array['a'..'z'] of integer;
  SetOfLetters = set of 'a'..'z';

  WordInfo = record                         { target or dictionary word }
               Letters : WordArray;           { the word itself }
               Length : integer;              { length of the word }
               LetterCounts : CountOfLetters; { count of letters 'a' to 'z' }
               LetterSet : SetOfLetters       { letters contained in the word }
             end;  { WordInfo }

var
  WORDLIST : text;                { file containing dictionary words }
  TargetWord : WordInfo;          { target word entered by the user }
  DictionaryWord : WordInfo;      { word read from dictionary }
  AnagramResult : AnagramType;    { anagram relationship between TargetWord and
                                    DictionaryWord -- full, partial, or none }
  AnagramCount : integer;         { number of anagrams found so far }
  WantsToContinue : boolean;      { user wants to enter another target word }

{*********************************************************************}

{---------------------------------------------------------------------}
{  The following compiler directives include the file "cases.i",      }
{  which contains the following functions:                            }
{    Lowercase - convert letters to lower case                        }
{    Uppercase - convert letters to upper case                        }
{                                                                     }
{  Note: Uppercase is not used in this program.                       }
{---------------------------------------------------------------------}

{$I cases.i} { for Turbo Pascal; treated as comment in standard Pascal }

{ ## Non-standard statements -- Berkeley Pascal version }
{
#include "cases.i"
}
{ End Berkeley Pascal version }

{*********************************************************************}

procedure Initialize;

  {-------------------------------------------------------------------}
  {  Initialize is called once at the beginning of the program.  It   }
  {  assigns an external file name to WORDLIST, if this is required   }
  {  in the Pascal version being used.                                }
  {                                                                   }
  {  Note: WORDLIST is opened (using reset) in the main program,      }
  {  because it may be read more than once during an execution of     }
  {  the program.                                                     }
  {-------------------------------------------------------------------}

  begin  { Initialize }

    { ## Non-standard statements -- Turbo Pascal version }

    assign(WORDLIST,'WORDLIST')

    { End Turbo Pascal version }

  end;  { Initialize }

{*********************************************************************}

procedure CountLetters(var InputWord : WordInfo);

  {-------------------------------------------------------------------}
  {  CountLetters is called to count the number of times each         }
  {  letter of the alphabet occurs in a target word or a dictionary   }
  {  word.  The word whose letters are to be counted is passed in     }
  {  InputWord.Letters.  The counts of letters are stored by the      }
  {  procedure in InputWord.LetterCounts. This procedure also builds  }
  {  the set InputWord.LetterSet, which indicates the letters that    }
  {  occur at least once in the input word.                           }
  {                                                                   }
  {  Upper case letters are converted to lower case before being      }
  {  counted.                                                         }
  {                                                                   }
  {  Input parameters:                                                }
  {    InputWord - contains the word whose characters are to be       }
  {                counted in the field Letters.  The length of       }
  {                the word is contained in the field Length.         }
  {                                                                   }
  {  Output parameters:                                               }
  {    InputWord - with the counts of alphabetic characters stored    }
  {                in the array LetterCounts, and the set of letters  }
  {                stored in LetterSet                                }
  {-------------------------------------------------------------------}

  var
    Letter : char;         { used as subscript for InputWord.LetterCounts }
    I : integer;           { used as subscript for InputWord.Letters }
    Character : char;      { character from InputWord.Letters }

  begin  { CountLetters }

    { initialize all letter counts to zero, and the set of letters to empty }

    for Letter := 'a' to 'z' do
      InputWord.LetterCounts[Letter] := 0;
    InputWord.LetterSet := [];

    { examine each character in the input word, converting it to lower case
      if alphabetic.  if the converted character is 'a' through 'z', add one
      to the corresponding count and enter the letter into LetterSet }

    for I := 1 to InputWord.Length do
      begin
        Character := Lowercase(InputWord.Letters[I]);
        if Character in ['a'..'z'] then
          begin
            InputWord.LetterCounts[Character] :=
                 InputWord.LetterCounts[Character] + 1;
            InputWord.LetterSet := InputWord.LetterSet + [Character]
          end  { if }
      end  { for }

  end;  { CountLetters }

{*********************************************************************}

procedure GetTargetWord(var TargetWord : WordInfo);

  {-------------------------------------------------------------------}
  {  GetTargetWord is called to get from the user a target word for   }
  {  which anagrams are to be sought.  It prompts the user to enter   }
  {  the target word, and then reads characters from the terminal     }
  {  until end-of-line is detected or the maximum length word has     }
  {  been read.  Non-alphabetic characters entered by the user are    }
  {  discarded, and letters are converted to lower case.              }
  {                                                                   }
  {  If the user enters input that is longer than the maximum word    }
  {  length, or if the word entered does not contain at least one     }
  {  alphabetic character, a warning message is displayed.  The       }
  {  prompting and reading of input are repeated until a valid target }
  {  word is obtained.                                                }
  {                                                                   }
  {  Input parameters: None                                           }
  {                                                                   }
  {  Output parameters:                                               }
  {    TargetWord - the target word read                              }
  {-------------------------------------------------------------------}

  var
    CharacterCount : integer;            { count of characters read so far }
    CurrentChar : char;                  { the current character }
    InputIsCorrect : boolean;            { target word is valid }

  begin  { GetTargetWord }

    { prompt and read input until a valid word is obtained }

    repeat  { until input is correct }
      write('Enter target word(s): ');
      CharacterCount := 0;

      { read characters of the target word until end-of-line is reached,
        or until the maximum length word is read.  discard non-alphabetic
        characters and convert letters to lower case }

      while (not eoln(input)) and (CharacterCount < MaxWordLength) do
        begin
          read(CurrentChar);
          if CurrentChar in ['a'..'z','A'..'Z'] then
            begin
              CharacterCount := CharacterCount + 1;
              TargetWord.Letters[CharacterCount] := Lowercase(CurrentChar);
            end  { if }
        end;  { while }

      { display an error message if no letter was found }

      if CharacterCount = 0 then
        writeln('Target word must contain at least one alphabetic character.');

      { if end-of-line was not found, the input was too long -- display an
        error message }

      if not eoln(input) then
          writeln('Maximum length of target word is ',
                  MaxWordLength:2,' letters.');

      { input is correct if end-of-line was reached and at least one letter
        was found }

      InputIsCorrect := (CharacterCount > 0) and (eoln(input));

      { skip any unread characters and the end-of-line marker }

      readln
    until InputIsCorrect;

    TargetWord.Length := CharacterCount;

    { pad the target word with blanks to its maximum length }

    for CharacterCount := TargetWord.Length + 1 to MaxWordLength do
      TargetWord.Letters[CharacterCount] := ' ';

    { count letters in target word and build set of letters contained in it }

    CountLetters(TargetWord)

  end;  { GetTargetWord }

{*********************************************************************}

procedure  GetDictionaryWord(var DictionaryWord : WordInfo);

  {-------------------------------------------------------------------}
  {  GetDictionaryWord is called to read a word from the dictionary.  }
  {  It reads characters from WORDLIST until end-of-line is detected  }
  {  or the maximum length word has been read.  If a word being read  }
  {  from the dictionary is longer than the maximum allowable word    }
  {  length, the excess characters are discarded with no error        }
  {  indication.                                                      }
  {                                                                   }
  {  Non-alphabetic and upper case letters in the dictionary word are }
  {  retained without change.  This allows the word to be displayed   }
  {  in its proper form when it appears as part of an anagram.        }
  {                                                                   }
  {  Input parameters: None                                           }
  {                                                                   }
  {  Output parameters:                                               }
  {    DictionaryWord - the dictionary word read                      }
  {-------------------------------------------------------------------}

  var
    CharacterCount : integer;            { count of characters read so far }

  begin  { GetDictionaryWord }

    { read characters of the dictionary word until end-of-line is reached,
      or until the maximum length word is read }

    CharacterCount := 0;
    while (not eoln(WORDLIST)) and (CharacterCount < MaxWordLength) do
      begin
        CharacterCount := CharacterCount + 1;
        read(WORDLIST,DictionaryWord.Letters[CharacterCount])
      end;  { while }

    { skip any unread characters and the end-of-line marker }

    readln(WORDLIST);
    DictionaryWord.Length := CharacterCount;

    { pad the dictionary word with blanks to its maximum length }

    for CharacterCount := DictionaryWord.Length + 1 to MaxWordLength do
      DictionaryWord.Letters[CharacterCount] := ' ';

    { count letters in dictionary word, and build set of letters in it }

    CountLetters(DictionaryWord)

  end;  { GetDictionaryWord }

{*********************************************************************}

function AreAnagrams(TargetWord : WordInfo;
                     DictionaryWord : WordInfo) : AnagramType;

  {-------------------------------------------------------------------}
  {  AreAnagrams tests whether or not the two words passed as         }
  {  parameters are anagrams.  The function value is set to indicate  }
  {  the anagram relationship as follows:                             }
  {      FullAnagram - TargetWord and DictionaryWord contain exactly  }
  {          the same letters                                         }
  {      PartialAnagram - DictionaryWord can be formed from the       }
  {          letters in TargetWord, but they are not full anagrams    }
  {      NotAnagram - neither of the above relationships holds        }
  {                                                                   }
  {  AreAnagrams operates by first comparing the sets of letters      }
  {  contained in TargetWord and DictionaryWord.  If all of the       }
  {  letters in DictionaryWord are present in TargetWord, then the    }
  {  words may be anagrams.  If not, they cannot be anagrams.         }
  {                                                                   }
  {  If the words pass the set-based test just described, the counts  }
  {  of letters are compared.  If the counts are the same for all     }
  {  letters 'a' through 'z', the two words are full anagrams.        }
  {  If any letter count in TargetWord is less than the corresponding }
  {  count in DictionaryWord, the words are not anagrams.  If all     }
  {  letter counts in TargetWord are greater than or equal to the     }
  {  corresponding counts in DictionaryWord, the words are partial    }
  {  anagrams.                                                        }
  {                                                                   }
  {  Input parameters:                                                }
  {    TargetWord - the target word                                   }
  {    DictionaryWord - a dictionary word                             }
  {                                                                   }
  {  Output parameters: None                                          }
  {                                                                   }
  {  Function value: set to indicate the type of anagram relationship }
  {                  between TargetWord and DictionaryWord            }
  {-------------------------------------------------------------------}

  var
    Letter : char;           { used as subscript for LetterCounts }
    Result : AnagramType;    { anagram relationship so far }

  begin  { AreAnagrams }

    { compare letter sets to see whether the words can possibly be anagrams }

    if not (DictionaryWord.LetterSet <= TargetWord.LetterSet) then

      { DictionaryWord contains letters that are not in TargetWord --
        the words cannot be anagrams }

      Result := NotAnagram

    else
      begin

        { words are potential anagrams.  assume that they are full anagrams
          until proven otherwise }

        Result := FullAnagram;

        { compare letter counts until words are found not to be anagrams, or
          until all counts have been examined }

        Letter := 'a';
        while (Result <> NotAnagram) and (Letter <= 'z') do
          begin
            if TargetWord.LetterCounts[Letter]
                            < DictionaryWord.LetterCounts[Letter] then

              { target word contains fewer of some letter than dictionary word;
                the words cannot be anagrams }

              Result := NotAnagram;
            if TargetWord.LetterCounts[Letter]
                            > DictionaryWord.LetterCounts[Letter] then

              { target word contains more of some letter than dictionary word;
                the words are not full anagrams, but may be partial anagrams }

              Result := PartialAnagram;

            { examine the next letter }

            Letter := succ(Letter)

          end  { while }

      end;  { if the words are potential anagrams }

    { set the function value }

    AreAnagrams := Result

  end;  { AreAnagrams }

{*********************************************************************}

procedure WriteAnagram(DictionaryWord : WordInfo; AnagramResult : AnagramType;
                       var AnagramCount : integer);

  {-------------------------------------------------------------------}
  {  WriteAnagram displays on the terminal a dictionary word that has }
  {  been found to be an anagram (full or partial) of the target      }
  {  word.  Full anagrams are indicated by preceding them with ' * '. }
  {                                                                   }
  {  The anagrams being displayed are counted by WriteAnagram.  A     }
  {  header message is written before the first anagram is displayed. }
  {  After every 20 anagrams, the user is prompted to press return    }
  {  before the display continues.                                    }
  {                                                                   }
  {  Input parameters:                                                }
  {    DictionaryWord - the dictionary word to be displayed as an     }
  {                     anagram                                       }
  {    AnagramResult - the type of anagram (full or partial)          }
  {    AnagramCount - the number of anagrams displayed so far         }
  {                                                                   }
  {  Output parameters:                                               }
  {    AnagramCount - the number of anagrams displayed, including the }
  {                   current one                                     }
  {-------------------------------------------------------------------}

  var
    I : integer;        { used as subscript for DictionaryWord.Letters }

  begin  { WriteAnagram }

    { if this is the first anagram displayed, write header message }

    if AnagramCount = 0 then
      writeln('Full anagrams (*) and partial anagrams found are');

    { count anagrams displayed. prompt the user to enter a carriage return
      on every 20th anagram }

    AnagramCount := AnagramCount + 1;
    if (AnagramCount mod 20) = 0 then
      begin
        writeln;
        writeln('Press return to continue');
        readln
      end;

    { write '*' to indicate a full anagram, or leave the space blank to
      indicate a partial anagram }

    if AnagramResult = FullAnagram then
      write(' * ')
    else
      write('   ');

    { display the dictionary word }

    for I := 1 to DictionaryWord.Length do
      write(DictionaryWord.Letters[I]);
    writeln

  end;  { WriteAnagram }

{*********************************************************************}

procedure AskUser(var WantsToContinue : boolean);

  {-------------------------------------------------------------------}
  {  AskUser is called after all anagrams for a given target word     }
  {  are displayed.  It asks the user if he or she wants to enter     }
  {  another target word, and sets the parameter WantsToContinue      }
  {  according to the answer received.                                }
  {                                                                   }
  {  If the first character entered by the user is 'y' or 'Y',        }
  {  WantsToContinue is set to true; if the first character is        }
  {  'n' or 'N', WantsToContinue is set to false.  The prompting is   }
  {  repeated if necessary until one of these two results is obtained.}
  {                                                                   }
  {  Input parameters: None                                           }
  {                                                                   }
  {  Output parameters:                                               }
  {    WantsToContinue - true if user wants to enter another          }
  {                      target word                                  }
  {-------------------------------------------------------------------}

  var
    InputCharacter : char;     { the first character entered by the user }

  begin  { AskUser }

    { prompt and read from terminal until one of the characters
      'y','Y','n','N' is obtained }

    repeat
      writeln;
      write('Another target word? (y/n) : ');
      readln(InputCharacter)
    until InputCharacter in ['y','Y','n','N'];

    { set WantsToContinue according to the input received }

    WantsToContinue := (InputCharacter in ['y','Y'])

  end;  { AskUser }

{*********************************************************************}

procedure CleanUp;

  {-------------------------------------------------------------------}
  {  CleanUp is called once at the end of the program.  It closes the }
  {  file WORDLIST, if this is required in the Pascal version being   }
  {  used.                                                            }
  {-------------------------------------------------------------------}

  begin  { CleanUp }

    { ## Non-standard statements -- Turbo Pascal version }

    close(WORDLIST)

    { End Turbo Pascal version }

  end;  { CleanUp }

{*********************************************************************}

begin  { main program }

  Initialize;

  { read target words and display anagrams as long as the user wants
    to continue }

  repeat  { until WantsToContinue is false }

    { get a target word from the user }

    GetTargetWord(TargetWord);

    { scan the dictionary, finding and displaying anagrams }

    reset(WORDLIST);
    AnagramCount := 0;
    while not eof(WORDLIST) do
      begin

        { get a word from the dictionary.  if it is an anagram of the
          target word, display it on the terminal }

        GetDictionaryWord(DictionaryWord);
        AnagramResult := AreAnagrams(TargetWord,DictionaryWord);
        if AnagramResult <> NotAnagram then
          WriteAnagram(DictionaryWord,AnagramResult,AnagramCount)

      end;  { while not eof }

    { if no anagrams were found, display a message }

    if AnagramCount = 0 then
      writeln('No anagrams found.');

    { does the user want to enter another target word ? }

    AskUser(WantsToContinue)

  until (WantsToContinue = false);
  CleanUp

end.
