program Anagram2(input,output,WORDLIST);

{---------------------------------------------------------------------}
{  This program uses a dictionary to find multi-word anagrams for a   }
{  target word or words entered by the user, and displays these       }
{  anagrams on the terminal.  A multi-word anagram is a sequence of   }
{  one or more words from the dictionary that contains exactly the    }
{  same letters as the target word(s).  For example, if the target    }
{  words are "word one", a multiword anagram would be "new door".     }
{                                                                     }
{  The case of letters is not considered when searching for anagrams, }
{  and non-alphabetic characters are ignored.  Thus, for example,     }
{  "sit" would be considered an anagram for "It's".                   }
{                                                                     }
{  The maximum length for the target word or words is 20 characters.  }
{                                                                     }
{  After a target word or words is obtained, the program scans the    }
{  dictionary.  All dictionary words that can be formed from the      }
{  letters in the target word(s) are loaded into a table.  This table }
{  is then searched by a recursive procedure FindMultiGrams to find   }
{  the multiple-word anagrams.  Only anagrams containing three words  }
{  or less are found by this program.                                 }
{                                                                     }
{  Input:  Target words are entered from the terminal.  Dictionary    }
{          words are read from the file WORDLIST.                     }
{                                                                     }
{  Output: The multi-word anagrams found are displayed on the         }
{          terminal.                                                  }
{                                                                     }
{  This program contains non-standard Pascal statements to support    }
{  Turbo Pascal and Berkeley Pascal versions.  Non-standard Pascal    }
{  statements appear in the following locations:                      }
{    1. in the compiler directives to include "cases.i" and "phrase.i"}
{    2. in procedures Initialize and CleanUp                          }
{  These non-standard statements are preceded by a comment that       }
{  begins with '## N'.                                                }
{                                                                     }
{  Modification history:                                              }
{    This program contains many of the same procedures as the         }
{    original version (program Anagram1).  However, this version      }
{    finds multi-word anagrams, whereas Anagram1 found only           }
{    single-word anagrams.  A recursive procedure FindMultiGrams      }
{    is used to search for the multiple-word anagrams.  An abstract   }
{    data type PhraseType is used to store and display the multi-word }
{    anagrams being generated.                                        }
{---------------------------------------------------------------------}

const
  MaxWordLength = 20;        { maximum length of target or dictionary word }
  MaxTableSize = 800;        { maximum size of dictionary table }
  MaxAnagramWords = 3;       { maximum words in a multiword anagram }
  MaxPhraseLength = 25;      { maximum characters in a PhraseType }

type
  AnagramType = (FullAnagram,PartialAnagram,NotAnagram);
  WordArray = array[1..MaxWordLength] of char;
  CountOfLetters = array['a'..'z'] of integer;
  SetOfLetters = set of 'a'..'z';

  WordInfo = record                         { target or dictionary word }
               Letters : WordArray;           { the word itself }
               Length : integer;              { length of the word }
               LetterCounts : CountOfLetters; { count of letters 'a' to 'z' }
               LetterSet : SetOfLetters       { letters contained in the word }
             end;  { WordInfo }

  TableOfWords = array [1..MaxTableSize] of WordInfo;
  EntriesInTable = 0..MaxTableSize;
  TableIndex = 1..MaxTableSize;

  PhraseType = record                     { abstract data type for phrase }
      CharCount : 0..MaxPhraseLength;       { number of characters in phrase }
      WordCount : integer;                  { number of words in phrase }
      Contents : array[1..MaxPhraseLength] of char
                                            { contents of the phrase }
      end;  { PhraseType }

var
  WORDLIST : text;                { file containing dictionary words }
  TargetWord : WordInfo;          { target word entered by the user }
  DictionaryWord : WordInfo;      { word read from dictionary }
  AnagramResult : AnagramType;    { anagram relationship between TargetWord and
                                    DictionaryWord -- full, partial, or none }
  AnagramCount : integer;         { number of anagrams found so far }
  WantsToContinue : boolean;      { user wants to enter another target word }
  WordTable : TableOfWords;       { table of dictionary words that could be
                                    in a multi-word anagram for TargetWord }
  TableSize : EntriesInTable;     { number of entries in WordTable }
  AnagramPhrase : PhraseType;     { multi-word anagram phrase }
  I : integer;                    { used as subscript for letters in word }

{*********************************************************************}

{---------------------------------------------------------------------}
{  The following compiler directives include the file "cases.i",      }
{  which contains the following functions:                            }
{    Lowercase - convert letters to lower case                        }
{    Uppercase - convert letters to upper case                        }
{                                                                     }
{  Note: Uppercase is not used in this program.                       }
{---------------------------------------------------------------------}

{$I cases.i} { for Turbo Pascal; treated as comment in standard Pascal }

{ ## Non-standard statements -- Berkeley Pascal version }
{
#include "cases.i"
}
{ End Berkeley Pascal version }

{*********************************************************************}

{---------------------------------------------------------------------}
{  The following compiler directives include the file "phrase.i",     }
{  which contains the following procedures and function:              }
{    InitPhrase - initialize a phrase so that it contains no words    }
{    AddToPhrase - add a new word at the end of a phrase              }
{    DisplayPhrase - display a phrase on the terminal                 }
{    WordsInPhrase - returns the current number of words in a phrase  }
{---------------------------------------------------------------------}

{$I phrase.i} { for Turbo Pascal; treated as comment in standard Pascal }

{ ## Non-standard statements -- Berkeley Pascal version }
{
#include "phrase.i"
}
{ End Berkeley Pascal version }

{*********************************************************************}

procedure Initialize;

  {-------------------------------------------------------------------}
  {  Initialize is called once at the beginning of the program.  It   }
  {  assigns an external file name to WORDLIST, if this is required   }
  {  in the Pascal version being used.                                }
  {                                                                   }
  {  Note: WORDLIST is opened (using reset) in the main program,      }
  {  because it may be read more than once during an execution of     }
  {  the program.                                                     }
  {-------------------------------------------------------------------}

  begin  { Initialize }

    { ## Non-standard statements -- Turbo Pascal version }

    assign(WORDLIST,'WORDLIST')

    { End Turbo Pascal version }

  end;  { Initialize }

{*********************************************************************}

procedure CountLetters(var InputWord : WordInfo);

  {-------------------------------------------------------------------}
  {  CountLetters is called to count the number of times each         }
  {  letter of the alphabet occurs in a target word or a dictionary   }
  {  word.  The word whose letters are to be counted is passed in     }
  {  InputWord.Letters.  The counts of letters are stored by the      }
  {  procedure in InputWord.LetterCounts. This procedure also builds  }
  {  the set InputWord.LetterSet, which indicates the letters that    }
  {  occur at least once in the input word.                           }
  {                                                                   }
  {  Upper case letters are converted to lower case before being      }
  {  counted.                                                         }
  {                                                                   }
  {  Input parameters:                                                }
  {    InputWord - contains the word whose characters are to be       }
  {                counted in the field Letters.  The length of       }
  {                the word is contained in the field Length.         }
  {                                                                   }
  {  Output parameters:                                               }
  {    InputWord - with the counts of alphabetic characters stored    }
  {                in the array LetterCounts, and the set of letters  }
  {                stored in LetterSet                                }
  {-------------------------------------------------------------------}

  var
    Letter : char;         { used as subscript for InputWord.LetterCounts }
    I : integer;           { used as subscript for InputWord.Letters }
    Character : char;      { character from InputWord.Letters }

  begin  { CountLetters }

    { initialize all letter counts to zero, and the set of letters to empty }

    for Letter := 'a' to 'z' do
      InputWord.LetterCounts[Letter] := 0;
    InputWord.LetterSet := [];

    { examine each character in the input word, converting it to lower case
      if alphabetic.  if the converted character is 'a' through 'z', add one
      to the corresponding count and enter the letter into LetterSet }

    for I := 1 to InputWord.Length do
      begin
        Character := Lowercase(InputWord.Letters[I]);
        if Character in ['a'..'z'] then
          begin
            InputWord.LetterCounts[Character] :=
                 InputWord.LetterCounts[Character] + 1;
            InputWord.LetterSet := InputWord.LetterSet + [Character]
          end  { if }
      end  { for }

  end;  { CountLetters }

{*********************************************************************}

procedure GetTargetWord(var TargetWord : WordInfo);

  {-------------------------------------------------------------------}
  {  GetTargetWord is called to get from the user a target word for   }
  {  which anagrams are to be sought.  It prompts the user to enter   }
  {  the target word, and then reads characters from the terminal     }
  {  until end-of-line is detected or the maximum length word has     }
  {  been read.  Non-alphabetic characters entered by the user are    }
  {  discarded, and letters are converted to lower case.              }
  {                                                                   }
  {  If the user enters input that is longer than the maximum word    }
  {  length, or if the word entered does not contain at least one     }
  {  alphabetic character, a warning message is displayed.  The       }
  {  prompting and reading of input are repeated until a valid target }
  {  word is obtained.                                                }
  {                                                                   }
  {  Input parameters: None                                           }
  {                                                                   }
  {  Output parameters:                                               }
  {    TargetWord - the target word read                              }
  {-------------------------------------------------------------------}

  var
    CharacterCount : integer;            { count of characters read so far }
    CurrentChar : char;                  { the current character }
    InputIsCorrect : boolean;            { target word is valid }

  begin  { GetTargetWord }

    { prompt and read input until a valid word is obtained }

    repeat  { until input is correct }
      write('Enter target word(s): ');
      CharacterCount := 0;

      { read characters of the target word until end-of-line is reached,
        or until the maximum length word is read.  discard non-alphabetic
        characters and convert letters to lower case }

      while (not eoln(input)) and (CharacterCount < MaxWordLength) do
        begin
          read(CurrentChar);
          if CurrentChar in ['a'..'z','A'..'Z'] then
            begin
              CharacterCount := CharacterCount + 1;
              TargetWord.Letters[CharacterCount] := Lowercase(CurrentChar);
            end  { if }
        end;  { while }

      { display an error message if no letter was found }

      if CharacterCount = 0 then
        writeln('Target word must contain at least one alphabetic character.');

      { if end-of-line was not found, the input was too long -- display an
        error message }

      if not eoln(input) then
          writeln('Maximum length of target word is ',
                  MaxWordLength:2,' letters.');

      { input is correct if end-of-line was reached and at least one letter
        was found }

      InputIsCorrect := (CharacterCount > 0) and (eoln(input));

      { skip any unread characters and the end-of-line marker }

      readln
    until InputIsCorrect;

    TargetWord.Length := CharacterCount;

    { pad the target word with blanks to its maximum length }

    for CharacterCount := TargetWord.Length + 1 to MaxWordLength do
      TargetWord.Letters[CharacterCount] := ' ';

    { count letters in target word and build set of letters contained in it }

    CountLetters(TargetWord)

  end;  { GetTargetWord }

{*********************************************************************}

procedure  GetDictionaryWord(var DictionaryWord : WordInfo);

  {-------------------------------------------------------------------}
  {  GetDictionaryWord is called to read a word from the dictionary.  }
  {  It reads characters from WORDLIST until end-of-line is detected  }
  {  or the maximum length word has been read.  If a word being read  }
  {  from the dictionary is longer than the maximum allowable word    }
  {  length, the excess characters are discarded with no error        }
  {  indication.                                                      }
  {                                                                   }
  {  Non-alphabetic and upper case letters in the dictionary word are }
  {  retained without change.  This allows the word to be displayed   }
  {  in its proper form when it appears as part of an anagram.        }
  {                                                                   }
  {  Input parameters: None                                           }
  {                                                                   }
  {  Output parameters:                                               }
  {    DictionaryWord - the dictionary word read                      }
  {-------------------------------------------------------------------}

  var
    CharacterCount : integer;            { count of characters read so far }

  begin  { GetDictionaryWord }

    { read characters of the dictionary word until end-of-line is reached,
      or until the maximum length word is read }

    CharacterCount := 0;
    while (not eoln(WORDLIST)) and (CharacterCount < MaxWordLength) do
      begin
        CharacterCount := CharacterCount + 1;
        read(WORDLIST,DictionaryWord.Letters[CharacterCount])
      end;  { while }

    { skip any unread characters and the end-of-line marker }

    readln(WORDLIST);
    DictionaryWord.Length := CharacterCount;

    { pad the dictionary word with blanks to its maximum length }

    for CharacterCount := DictionaryWord.Length + 1 to MaxWordLength do
      DictionaryWord.Letters[CharacterCount] := ' ';

    { count letters in dictionary word, and build set of letters in it }

    CountLetters(DictionaryWord)

  end;  { GetDictionaryWord }

{*********************************************************************}

function AreAnagrams(TargetWord : WordInfo;
                     DictionaryWord : WordInfo) : AnagramType;

  {-------------------------------------------------------------------}
  {  AreAnagrams tests whether or not the two words passed as         }
  {  parameters are anagrams.  The function value is set to indicate  }
  {  the anagram relationship as follows:                             }
  {      FullAnagram - TargetWord and DictionaryWord contain exactly  }
  {          the same letters                                         }
  {      PartialAnagram - DictionaryWord can be formed from the       }
  {          letters in TargetWord, but they are not full anagrams    }
  {      NotAnagram - neither of the above relationships holds        }
  {                                                                   }
  {  AreAnagrams operates by first comparing the sets of letters      }
  {  contained in TargetWord and DictionaryWord.  If all of the       }
  {  letters in DictionaryWord are present in TargetWord, then the    }
  {  words may be anagrams.  If not, they cannot be anagrams.         }
  {                                                                   }
  {  If the words pass the set-based test just described, the counts  }
  {  of letters are compared.  If the counts are the same for all     }
  {  letters 'a' through 'z', the two words are full anagrams.        }
  {  If any letter count in TargetWord is less than the corresponding }
  {  count in DictionaryWord, the words are not anagrams.  If all     }
  {  letter counts in TargetWord are greater than or equal to the     }
  {  corresponding counts in DictionaryWord, the words are partial    }
  {  anagrams.                                                        }
  {                                                                   }
  {  Input parameters:                                                }
  {    TargetWord - the target word                                   }
  {    DictionaryWord - a dictionary word                             }
  {                                                                   }
  {  Output parameters: None                                          }
  {                                                                   }
  {  Function value: set to indicate the type of anagram relationship }
  {                  between TargetWord and DictionaryWord            }
  {-------------------------------------------------------------------}

  var
    Letter : char;           { used as subscript for LetterCounts }
    Result : AnagramType;    { anagram relationship so far }

  begin  { AreAnagrams }

    { compare letter sets to see whether the words can possibly be anagrams }

    if not (DictionaryWord.LetterSet <= TargetWord.LetterSet) then

      { DictionaryWord contains letters that are not in TargetWord --
        the words cannot be anagrams }

      Result := NotAnagram

    else
      begin

        { words are potential anagrams.  assume that they are full anagrams
          until proven otherwise }

        Result := FullAnagram;

        { compare letter counts until words are found not to be anagrams, or
          until all counts have been examined }

        Letter := 'a';
        while (Result <> NotAnagram) and (Letter <= 'z') do
          begin
            if TargetWord.LetterCounts[Letter]
                            < DictionaryWord.LetterCounts[Letter] then

              { target word contains fewer of some letter than dictionary word;
                the words cannot be anagrams }

              Result := NotAnagram;
            if TargetWord.LetterCounts[Letter]
                            > DictionaryWord.LetterCounts[Letter] then

              { target word contains more of some letter than dictionary word;
                the words are not full anagrams, but may be partial anagrams }

              Result := PartialAnagram;

            { examine the next letter }

            Letter := succ(Letter)

          end  { while }

      end;  { if the words are potential anagrams }

    { set the function value }

    AreAnagrams := Result

  end;  { AreAnagrams }

{*********************************************************************}

procedure WriteAnagram(Phrase : PhraseType; var AnagramCount : integer);

  {-------------------------------------------------------------------}
  {  WriteAnagram is called whenever FindMultiGrams completes a       }
  {  multi-word anagram.  It displays the completed anagram phrase    }
  {  on the terminal, using the procedure DisplayPhrase from the      }
  {  abstract data type PhraseType.                                   }
  {                                                                   }
  {  The anagrams being displayed are counted by WriteAnagram.  A     }
  {  header message is written before the first anagram is displayed. }
  {  After every 20 anagrams, the user is prompted to press return    }
  {  before the display continues.                                    }
  {                                                                   }
  {  Input parameters:                                                }
  {    Phrase - the multi-word anagram phrase to be displayed         }
  {    AnagramCount - the number of anagrams displayed so far         }
  {                                                                   }
  {  Output parameters:                                               }
  {    AnagramCount - the number of anagrams displayed, including the }
  {                   current one                                     }
  {-------------------------------------------------------------------}

  begin  { WriteAnagram }

    { if this is the first anagram displayed, write header message }

    if AnagramCount = 0 then
      writeln('Anagrams found are');

    { count anagrams displayed. prompt the user to enter a carriage return
      on every 20th anagram }

    AnagramCount := AnagramCount + 1;
    if (AnagramCount mod 20) = 0 then
      begin
        writeln;
        writeln('Press return to continue');
        readln
      end;

    { display the anagram found }

    DisplayPhrase(Phrase)

  end;  { WriteAnagram }

{*********************************************************************}

procedure LoadTable(DictionaryWord : WordInfo; var WordTable : TableOfWords;
                    var TableSize : EntriesInTable);

  {-------------------------------------------------------------------}
  {  LoadTable is called to insert a dictionary word into the table   }
  {  that is used to search for multi-word anagrams.                  }
  {                                                                   }
  {  Input parameters:                                                }
  {    DictionaryWord - the word to be added to the table, with       }
  {                     letter counts and sets already filled in      }
  {    WordTable - the table of words                                 }
  {    TableSize - the number of entries already in WordTable         }
  {                                                                   }
  {  Output parameters:                                               }
  {    WordTable - the table with DictionaryWord added                }
  {    TableSize - the number of entries now in WordTable             }
  {-------------------------------------------------------------------}

  var
    I : integer;             { used as subscript for DictionaryWord.Letters }
    Letter : char;           { subscript for DictionaryWord.LetterCounts }

  begin  { LoadTable }

    TableSize := TableSize + 1;
    for I := 1 to MaxWordLength do
      WordTable[TableSize].Letters[I] := DictionaryWord.Letters[I];
    WordTable[TableSize].Length := DictionaryWord.Length;
    WordTable[TableSize].LetterSet := DictionaryWord.LetterSet;
    for Letter := 'a' to 'z' do
      WordTable[TableSize].LetterCounts[Letter] :=
                       DictionaryWord.LetterCounts[Letter]

  end;  { LoadTable }

{*********************************************************************}

procedure FindMultiGrams(OldPartialPhrase : PhraseType; OldTarget : WordInfo;
               var WordTable : TableOfWords; TableSize : EntriesInTable;
               StartIndex : TableIndex; var AnagramCount : integer);

  {-------------------------------------------------------------------}
  {  FindMultiGrams is a recursive procedure that searches for        }
  {  multi-word anagrams, using the table of dictionary words.  It is }
  {  called with a partially constructed multi-word anagram phrase    }
  {  (OldPartialPhrase), and a target (OldTarget) that contains all   }
  {  letters of the original target word not accounted for in         }
  {  OldPartialPhrase.                                                }
  {                                                                   }
  {  FindMultiGrams searches WordTable for words that will fit into   }
  {  the multi-word anagram being constructed.  Each such word that   }
  {  is found is added to the anagram from OldPartialPhrase.  If      }
  {  any letters of the target word are still left over, the          }
  {  procedure calls itself recursively to attempt to add yet another }
  {  word to the anagram.  If no letters remain in the target, the    }
  {  completed anagram phrase is displayed by calling the procedure   }
  {  WriteAnagram.                                                    }
  {                                                                   }
  {  For example, if the original target words were "word one",       }
  {  FindMultiGrams might be called with a partially constructed      }
  {  anagram  OldPartialPhrase = "do" and remaining letters           }
  {  OldTarget = "wrone".  In such a case, it might find a new word   }
  {  "nor" that could be added to the anagram.  It would then call    }
  {  itself recursively with the partial anagram phrase "do nor" and  }
  {  the remaining letters "we".  During this recursive call, the new }
  {  word "we" would be found.  Since this exhausts all the letters   }
  {  of the target, the full anagram "do nor we" would be displayed.  }
  {                                                                   }
  {  After finding "nor" and making the recursive call, the procedure }
  {  might next attempt to add the word "one" to the partial anagram. }
  {  It would then call itself with the partial anagram "do one" and  }
  {  the remaining letters "wr".  During the recursive call, no new   }
  {  word would be found that could be made from "wr"; thus "do one"  }
  {  would be discarded as the beginning of an anagram.               }
  {                                                                   }
  {  The search for a new word to add to the anagram begins at the    }
  {  point in the word table where the previous word was found.  This }
  {  avoids finding the same anagram words several times in different }
  {  orders -- for example, "do nor we" and then "do we nor".         }
  {                                                                   }
  {  An internal procedure BuildNewTarget is used to generate the     }
  {  list of letters remaining after a new word is added to the       }
  {  partial anagram phrase.  The abstract data type PhraseType is    }
  {  used to represent an anagram phrase.                             }
  {                                                                   }
  {  Input parameters:                                                }
  {    OldPartialPhrase - the partial anagram phrase being formed     }
  {    OldTarget - the letters of the target word(s) not accounted    }
  {                for by the words in OldPartialPhrase               }
  {    WordTable - the table of dictionary words                      }
  {        Note: WordTable is a var parameter for efficiency of       }
  {              parameter passing                                    }
  {    TableSize - the number of entries in WordTable                 }
  {    StartIndex - the starting point in WordTable from which to     }
  {                 search for a new word for the anagram             }
  {    AnagramCount - the number of anagrams found so far             }
  {                                                                   }
  {  Output parameters:                                               }
  {    AnagramCount - the number of anagrams found so far             }
  {-------------------------------------------------------------------}

  var
    I : TableIndex;                 { used as subscript for WordTable }
    AnagramResult : AnagramType;    { anagram relationship between target
                                      and table word }
    NewPartialPhrase : PhraseType;  { partial anagram with new word added }
    NewTarget : WordInfo;           { remaining letters in target }
    NewStartIndex : TableIndex;     { point in table to begin search for
                                      next word }

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

procedure BuildNewTarget(OldTarget : WordInfo; TrialWord : WordInfo;
                         var NewTarget : WordInfo);

  {-------------------------------------------------------------------}
  {  BuildNewTarget is called each time a new word is added to an     }
  {  anagram.  It generates the list of target letters that remain    }
  {  unaccounted for after the new word is added to the anagram.      }
  {  This list, which is returned in NewTarget, contains all of the   }
  {  remaining letters with blanks removed; letter counts and sets of }
  {  letters are generated for the new target, so that it has the     }
  {  same format as an original target word.                          }
  {                                                                   }
  {  Input parameters:                                                }
  {    OldTarget - the target letters remaining before the new word   }
  {                is added to the anagram                            }
  {    TrialWord - the new word being added to the anagram            }
  {                                                                   }
  {  Output parameters:                                               }
  {    NewTarget - the target letters remaining after the new word    }
  {                is added to the anagram, with letter counts and    }
  {                letter set constructed                             }
  {-------------------------------------------------------------------}

  var
    I,J : integer;               { used as subscripts for OldTarget.Letters,
                                   NewTarget.Letters, and NewLetters }
    LetterUsed : char;           { a letter used in the new anagram word }
    NewLetters : WordArray;      { used to construct the list of letters
                                   remaining }

  begin  { BuildNewTarget }

    { copy all the letters from OldTarget to NewLetters }

    for J := 1 to MaxWordLength do
      NewLetters[J] := OldTarget.Letters[J];

    { erase from NewLetters each letter that appears in TrialWord }

    for I := 1 to TrialWord.Length do
      begin
        LetterUsed := Lowercase(TrialWord.Letters[I]);
        if LetterUsed in ['a'..'z'] then
          begin

            { LetterUsed is a letter from TrialWord -- find a matching
              letter in NewLetters and erase it }

            J := 1;
            while NewLetters[J] <> LetterUsed do
              J := J + 1;
            NewLetters[J] := ' '
          end  { if }
      end;  { for }

    { copy all of the non-blank characters from NewLetters into
      NewTarget.Letters; set NewTarget.Length accordingly }

    I := 0;
    for J := 1 to MaxWordLength do
      if NewLetters[J] <> ' ' then
        begin
          I := I + 1;
          NewTarget.Letters[I] := NewLetters[J]
        end;
    NewTarget.Length := I;

    { fill unused characters in NewTarget.Letters with blanks }

    for J := I + 1 to MaxWordLength do
      NewTarget.Letters[J] := ' ';

    { count letters in NewTarget and build set of letters contained }

    CountLetters(NewTarget)

  end;  { BuildNewTarget }

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

  begin  { FindMultiGrams }

    { examine each word in the table, from StartIndex to the end of the table }

    for I := StartIndex to TableSize do
      begin

        { check anagram relationship between table word and old target }

        AnagramResult := AreAnagrams(OldTarget,WordTable[I]);
        if (AnagramResult = PartialAnagram)
          and (WordsInPhrase(OldPartialPhrase) < MaxAnagramWords - 1) then
            begin

              { the table word can be part of the anagram, and there is
                still room for another word after it.  add the new word
                to the anagram, and build the list of letters remaining }

              AddToPhrase(OldPartialPhrase,WordTable[I].Letters,
                          WordTable[I].Length,NewPartialPhrase);
              BuildNewTarget(OldTarget,WordTable[I],NewTarget);

              { call FindMultiGrams to search for another word, beginning
                the search with the current word in the table }

              NewStartIndex := I;
              FindMultiGrams(NewPartialPhrase,NewTarget,WordTable,
                             TableSize,NewStartIndex,AnagramCount)
            end  { if PartialAnagram and ... }

        else if AnagramResult = FullAnagram then
          begin

            { the table word completes a multi-word anagram -- add it
              to the phrase and display the anagram }

            AddToPhrase(OldPartialPhrase,WordTable[I].Letters,
                        WordTable[I].Length,NewPartialPhrase);
            WriteAnagram(NewPartialPhrase,AnagramCount)

          end  { if FullAnagram }

      end  { for }

  end;  { FindMultiGrams }

{*********************************************************************}

procedure AskUser(var WantsToContinue : boolean);

  {-------------------------------------------------------------------}
  {  AskUser is called after all anagrams for a given target word     }
  {  are displayed.  It asks the user if he or she wants to enter     }
  {  another target word, and sets the parameter WantsToContinue      }
  {  according to the answer received.                                }
  {                                                                   }
  {  If the first character entered by the user is 'y' or 'Y',        }
  {  WantsToContinue is set to true; if the first character is        }
  {  'n' or 'N', WantsToContinue is set to false.  The prompting is   }
  {  repeated if necessary until one of these two results is obtained.}
  {                                                                   }
  {  Input parameters: None                                           }
  {                                                                   }
  {  Output parameters:                                               }
  {    WantsToContinue - true if user wants to enter another          }
  {                      target word                                  }
  {-------------------------------------------------------------------}

  var
    InputCharacter : char;     { the first character entered by the user }

  begin  { AskUser }

    { prompt and read from terminal until one of the characters
      'y','Y','n','N' is obtained }

    repeat
      writeln;
      write('Another target word? (y/n) : ');
      readln(InputCharacter)
    until InputCharacter in ['y','Y','n','N'];

    { set WantsToContinue according to the input received }

    WantsToContinue := (InputCharacter in ['y','Y'])

  end;  { AskUser }

{*********************************************************************}

procedure CleanUp;

  {-------------------------------------------------------------------}
  {  CleanUp is called once at the end of the program.  It closes the }
  {  file WORDLIST, if this is required in the Pascal version being   }
  {  used.                                                            }
  {-------------------------------------------------------------------}

  begin  { CleanUp }

    { ## Non-standard statements -- Turbo Pascal version }

    close(WORDLIST)

    { End Turbo Pascal version }

  end;  { CleanUp }

{*********************************************************************}

begin  { main program }

  Initialize;

  { read target words and display anagrams as long as the user wants
    to continue }

  repeat  { until WantsToContinue is false }

    { get a target word from the user }

    GetTargetWord(TargetWord);

    { read the dictionary, loading into WordTable each dictionary word
      that is a full or partial anagram of the target word.  stop when
      the end of the dictionary is reached, or when the table is full }

    reset(WORDLIST);
    writeln('Scanning dictionary...');
    TableSize := 0;
    while (not eof(WORDLIST)) and (TableSize < MaxTableSize) do
      begin
        GetDictionaryWord(DictionaryWord);
        AnagramResult := AreAnagrams(TargetWord,DictionaryWord);
        if AnagramResult <> NotAnagram then
          LoadTable(DictionaryWord,WordTable,TableSize)
      end;  { while }

    { if the table became full before reaching the end of WORDLIST, display
      a warning message }

    if not eof(WORDLIST) then
      begin
        writeln('Dictionary table is full -- skipping words after ');
        for I := 1 to DictionaryWord.Length do
          write(DictionaryWord.Letters[I]);
        writeln('.')
      end;  { if table is full }

    { initialize the multi-word anagram to blanks }

    InitPhrase(AnagramPhrase);

    { find and display multi-word anagrams }

    AnagramCount := 0;
    writeln('Searching for anagrams...');
    FindMultiGrams(AnagramPhrase,TargetWord,WordTable,
                   TableSize,1,AnagramCount);

    { if no anagrams were found, display a message }

    if AnagramCount = 0 then
      writeln('No anagrams found.');

    { does the user want to enter another target word ? }

    AskUser(WantsToContinue)

  until (WantsToContinue = false);
  CleanUp

end.
