{---------------------------------------------------------------------}
{  The following procedures and functions implement operations on the }
{  abstract data type PhraseType.  This data type can be used to      }
{  represent a phrase, which is a sequence of words separated by      }
{  spaces.  The maximum number of characters in a word is given by    }
{  the constant MaxWordLength.  The maximum number of characters in a }
{  phrase (including the spaces between words) is given by the        }
{  constant MaxPhraseLength.  The following procedures and functions  }
{  are provided:                                                      }
{                                                                     }
{     InitPhrase - initializes a phrase so that it contains no words  }
{     AddToPhrase - adds a new word at the end of a phrase            }
{     DisplayPhrase - displays a phrase on the terminal               }
{     WordsInPhrase - returns the current number of words in a phrase }
{                                                                     }
{  In order to use this abstract data type, the main program must     }
{  contain the following constant and type definitions.  Items        }
{  enclosed in <<...>> are to be supplied by the programmer.          }
{                                                                     }
{    const                                                            }
{      MaxWordLength = << integer between 1 and 80 >>;                }
{      MaxPhraseLength = << integer between 1 and 80 >>;              }
{                                                                     }
{    type                                                             }
{                                                                     }
{     WordArray = array[1..MaxWordLength] of char;                    }
{                                                                     }
{     PhraseType = record                                             }
{                    CharCount : 0..MaxPhraseLength;                  }
{                    WordCount : integer;                             }
{                    Contents : array[1..MaxPhraseLength] of char     }
{                end;                                                 }
{                                                                     }
{  In this implementation of PhraseType, a phrase is represented by   }
{  a record that contains the following fields:                       }
{                                                                     }
{      CharCount - the number of characters currently in the phrase   }
{      WordCount - the number of words currently in the phrase        }
{      Contents - an array containing the characters in the phrase    }
{                                                                     }
{  The maximum number of characters allowed in a phrase is 80.        }
{                                                                     }
{  A word to be added to a phrase is represented as an (unpacked)     }
{  array of characters, with length given by the constant             }
{  MaxWordLength.                                                     }
{---------------------------------------------------------------------}


{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

procedure InitPhrase(var Phrase : PhraseType);

  {-------------------------------------------------------------------}
  {  InitPhrase is called to initialize a phrase.  It sets the        }
  {  contents of the phrase to all blanks, the length of the phrase   }
  {  to zero, and the number of words in the phrase to zero.          }
  {                                                                   }
  {  Input parameters: None                                           }
  {                                                                   }
  {  Output parameters:                                               }
  {    Phrase - the initialized phrase                                }
  {-------------------------------------------------------------------}

  var
    I : integer;       { used as subscript for Contents }

  begin  { InitPhrase }

    for I := 1 to MaxPhraseLength do
      Phrase.Contents[I] := ' ';
    Phrase.CharCount := 0;
    Phrase.WordCount := 0

  end;  { InitPhrase }

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

procedure AddToPhrase(OldPhrase : PhraseType; NewWord : WordArray;
                    WordLength : integer; var NewPhrase : PhraseType);

  {-------------------------------------------------------------------}
  {  AddToPhrase adds a word to an existing phrase, creating a new    }
  {  phrase. If the new word will not fit in the existing phrase,     }
  {  an error message is displayed.                                   }
  {                                                                   }
  {  Input parameters:                                                }
  {    OldPhrase - the existing phrase to which a word is to be added }
  {    NewWord - the new word to add to the phrase                    }
  {    WordLength - the number of characters in NewWord               }
  {                                                                   }
  {  Output parameters:                                               }
  {    NewPhrase - the phrase with the new word added                 }
  {-------------------------------------------------------------------}

  var
    I : integer;                { used as subscript for Contents }
    StartingPoint : integer;    { starting point in Contents for new word }

  begin  { AddToPhrase }

    { check to be sure that the new word will fit in the phrase }

    if (OldPhrase.CharCount + WordLength) > MaxPhraseLength then
      writeln('*** AddToPhrase: Word is too long to be added to phrase')
    else
      begin  { add the word to the phrase }

        { copy the contents of the old phrase to the new phrase }

        for I := 1 to MaxPhraseLength do
          NewPhrase.Contents[I] := OldPhrase.Contents[I];

        { start the new word after the end of the previous word(s) }

        StartingPoint := OldPhrase.CharCount + 1;

        { add the new word, setting the phrase length and the number of words }

        for I := 1 to WordLength do
          NewPhrase.Contents[StartingPoint + I] := NewWord[I];
        NewPhrase.CharCount := OldPhrase.CharCount + WordLength + 1;
        NewPhrase.WordCount := OldPhrase.WordCount + 1;

      end  { adding word to the phrase }

  end;  { AddToPhrase }

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

procedure DisplayPhrase(Phrase : PhraseType);

  {-------------------------------------------------------------------}
  {  DisplayPhrase displays a phrase on the terminal.  The characters }
  {  in the phrase are displayed, followed by a carriage return.      }
  {                                                                   }
  {  Input parameters:                                                }
  {    Phrase - the phrase to be displayed                            }
  {                                                                   }
  {  Output parameters: None                                          }
  {-------------------------------------------------------------------}

  var
    I : integer;      { used as subscript for Phrase.Contents }

  begin  { DisplayPhrase }

    for I := 1 to Phrase.CharCount do
      write(Phrase.Contents[I]);
    writeln

  end;  { DisplayPhrase }

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

function WordsInPhrase(Phrase : PhraseType) : integer;

  {-------------------------------------------------------------------}
  {  WordsInPhrase returns the number of words currently in a phrase. }
  {                                                                   }
  {  Input parameters:                                                }
  {    Phrase - the phrase to be processed                            }
  {                                                                   }
  {  Output parameters: None                                          }
  {                                                                   }
  {  Function value: the number of words in Phrase                    }
  {-------------------------------------------------------------------}

  begin  { WordsInPhrase }

    WordsInPhrase := Phrase.WordCount

  end;  { WordsInPhrase }
