program MakeDict(input,output,WORDSORT,DICTFILE);

{---------------------------------------------------------------------}
{  This program creates the dictionary file DICTFILE that is needed   }
{  by the program Anagram4.  It reads a list of dictionary words,     }
{  ordered by length (longest first) from the file WORDSORT.  For     }
{  each such word, it constructs and writes a record to DICTFILE.     }
{  This record contains the word and its length, a count of the       }
{  number of occurrences of each letter in the word, and a set of the }
{  letters contained in the word.                                     }
{                                                                     }
{  The words read from WORDSORT are checked to be sure that they are  }
{  in the proper order.  If a word is out of sequence (i.e., longer   }
{  than the previous word), an error message is displayed and the     }
{  program terminates.                                                }
{                                                                     }
{  This program contains procedures CountLetters and                  }
{  GetDictionaryWord that are identical to the corresponding          }
{  procedures in program Anagram3.                                    }
{                                                                     }
{  Input:  Dictionary words, sorted by length, are read from WORDSORT }
{                                                                     }
{  Output: Dictionary records are written to DICTFILE                 }
{                                                                     }
{  This program contains non-standard Pascal statements to support    }
{  Turbo Pascal and Berkeley Pascal versions.  Non-standard Pascal    }
{  statements appear in the following locations:                      }
{    1. in the compiler directives to include "cases.i"               }
{    2. in procedures Initialize and CleanUp                          }
{  These non-standard statements are preceded by a comment that       }
{  begins with '## N'.                                                }
{---------------------------------------------------------------------}

const
  MaxWordLength = 20;        { maximum length of dictionary word }

type
  WordArray = array[1..MaxWordLength] of char;
  CountOfLetters = array['a'..'z'] of integer;
  SetOfLetters = set of 'a'..'z';

  WordInfo = record                         { dictionary word }
               Letters : WordArray;           { the word itself }
               Length : integer;              { length of the word }
               LetterCounts : CountOfLetters; { count of letters 'a' to 'z' }
               LetterSet : SetOfLetters       { letters contained in the word }
             end;  { WordInfo }
var
  WORDSORT : text;                { file containing words sorted by length }
  DICTFILE : file of WordInfo;    { dictionary file to be constructed }
  DictionaryWord : WordInfo;      { word for the dictionary }
  PreviousLength : integer;       { length of the previous dictionary word }
  SequenceIsCorrect : boolean;    { the dictionary words are in the proper
                                    order (longest first) }
  I : integer;                    { used as subscript for letters in word }

{*********************************************************************}

{---------------------------------------------------------------------}
{  The following compiler directives include the file "cases.i",      }
{  which contains the following functions:                            }
{    Lowercase - convert letters to lower case                        }
{    Uppercase - convert letters to upper case                        }
{                                                                     }
{  Note: Uppercase is not used in this program.                       }
{---------------------------------------------------------------------}

{$I cases.i} { for Turbo Pascal; treated as comment in standard Pascal }

{ ## Non-standard statements -- Berkeley Pascal version }
{
#include "cases.i"
}
{ End Berkeley Pascal version }

{*********************************************************************}

procedure Initialize;

  {-------------------------------------------------------------------}
  {  Initialize is called once at the beginning of the program.  It   }
  {  opens the files WORDSORT and DICTFILE.                           }
  {-------------------------------------------------------------------}

  begin  { Initialize }

    { ## Non-standard statements -- Turbo Pascal version }

    assign(WORDSORT,'WORDSORT');
    assign(DICTFILE,'DICTFILE');

    { End Turbo Pascal version }

    reset(WORDSORT);
    rewrite(DICTFILE);

  end;  { Initialize }

{*********************************************************************}

procedure CountLetters(var InputWord : WordInfo);

  {-------------------------------------------------------------------}
  {  CountLetters is called to count the number of times each         }
  {  letter of the alphabet occurs in a target word or a dictionary   }
  {  word.  The word whose letters are to be counted is passed in     }
  {  InputWord.Letters.  The counts of letters are stored by the      }
  {  procedure in InputWord.LetterCounts. This procedure also builds  }
  {  the set InputWord.LetterSet, which indicates the letters that    }
  {  occur at least once in the input word.                           }
  {                                                                   }
  {  Upper case letters are converted to lower case before being      }
  {  counted.                                                         }
  {                                                                   }
  {  Input parameters:                                                }
  {    InputWord - contains the word whose characters are to be       }
  {                counted in the field Letters.  The length of       }
  {                the word is contained in the field Length.         }
  {                                                                   }
  {  Output parameters:                                               }
  {    InputWord - with the counts of alphabetic characters stored    }
  {                in the array LetterCounts, and the set of letters  }
  {                stored in LetterSet                                }
  {-------------------------------------------------------------------}

  var
    Letter : char;         { used as subscript for InputWord.LetterCounts }
    I : integer;           { used as subscript for InputWord.Letters }
    Character : char;      { character from InputWord.Letters }

  begin  { CountLetters }

    { initialize all letter counts to zero, and the set of letters to empty }

    for Letter := 'a' to 'z' do
      InputWord.LetterCounts[Letter] := 0;
    InputWord.LetterSet := [];

    { examine each character in the input word, converting it to lower case
      if alphabetic.  if the converted character is 'a' through 'z', add one
      to the corresponding count and enter the letter into LetterSet }

    for I := 1 to InputWord.Length do
      begin
        Character := Lowercase(InputWord.Letters[I]);
        if Character in ['a'..'z'] then
          begin
            InputWord.LetterCounts[Character] :=
                 InputWord.LetterCounts[Character] + 1;
            InputWord.LetterSet := InputWord.LetterSet + [Character]
          end  { if }
      end  { for }

  end;  { CountLetters }

{*********************************************************************}

procedure  GetDictionaryWord(var DictionaryWord : WordInfo);

  {-------------------------------------------------------------------}
  {  GetDictionaryWord is called to read a word from the dictionary.  }
  {  It reads characters from WORDSORT until end-of-line is detected  }
  {  or the maximum length word has been read.  If a word being read  }
  {  from the dictionary is longer than the maximum allowable word    }
  {  length, the excess characters are discarded with no error        }
  {  indication.                                                      }
  {                                                                   }
  {  Non-alphabetic and upper case letters in the dictionary word are }
  {  retained without change.  This allows the word to be displayed   }
  {  in its proper form when it appears as part of an anagram.        }
  {                                                                   }
  {  Input parameters: None                                           }
  {                                                                   }
  {  Output parameters:                                               }
  {    DictionaryWord - the dictionary word read                      }
  {-------------------------------------------------------------------}

  var
    CharacterCount : integer;            { count of characters read so far }

  begin  { GetDictionaryWord }

    { read characters of the dictionary word until end-of-line is reached,
      or until the maximum length word is read }

    CharacterCount := 0;
    while (not eoln(WORDSORT)) and (CharacterCount < MaxWordLength) do
      begin
        CharacterCount := CharacterCount + 1;
        read(WORDSORT,DictionaryWord.Letters[CharacterCount])
      end;  { while }

    { skip any unread characters and the end-of-line marker }

    readln(WORDSORT);
    DictionaryWord.Length := CharacterCount;

    { pad the dictionary word with blanks to its maximum length }

    for CharacterCount := DictionaryWord.Length + 1 to MaxWordLength do
      DictionaryWord.Letters[CharacterCount] := ' ';

    { count letters in dictionary word, and build set of letters in it }

    CountLetters(DictionaryWord)

  end;  { GetDictionaryWord }

{*********************************************************************}

procedure CleanUp;

  {-------------------------------------------------------------------}
  {  CleanUp is called once at the end of the program.  It closes the }
  {  files WORDSORT and DICTFILE, if this is required in the Pascal   }
  {  version being used, and waits for the user to enter a carriage   }
  {  return before exiting the program.                               }
  {-------------------------------------------------------------------}

  begin  { CleanUp }

    { ## Non-standard statements -- Turbo Pascal version }

    close(WORDSORT);
    close(DICTFILE);

    { End Turbo Pascal version }

    write('Press return to exit program');
    readln

  end;  { CleanUp }

{*********************************************************************}

begin  { main program }

  Initialize;
  PreviousLength := MaxWordLength;
  SequenceIsCorrect := true;
  writeln('Building dictionary file...');

  { read dictionary words from WORDSORT and write dictionary records to
    DICTFILE, checking for proper sequence.  stop when the end of WORDSORT
    is reached, or when a sequence error is found }

  while (SequenceIsCorrect) and (not eof(WORDSORT)) do
    begin
      GetDictionaryWord(DictionaryWord);
      if DictionaryWord.Length > PreviousLength then
        SequenceIsCorrect := false
      else
        begin
          PreviousLength := DictionaryWord.Length;
          write(DICTFILE,DictionaryWord)
        end  { else }
    end;  { while }

  { display a "normal termination" message or an error message, according
    to whether a sequence error was found }

  if SequenceIsCorrect then
    writeln('Dictionary file complete.')
  else
    begin
      write('*** WORDSORT out of sequence at word ');
      for I := 1 to DictionaryWord.Length do
        write(DictionaryWord.Letters[I]);
      writeln(' -- terminating run.')
    end;  { else }
  CleanUp

end.
