program Life(input,output);

{---------------------------------------------------------------------}
{  This program plays the simulation game called "Life", which was    }
{  invented by the British mathematician John Horton Conway.  The     }
{  game simulates the growth and development of a community of living }
{  organisms.  Each organism occupies a single cell in a rectangular  }
{  grid (similar to a checkerboard).  Each such cell has eight        }
{  neighboring cells -- two horizontally, two vertically, and four    }
{  diagonally.  Births and deaths of organisms from one generation    }
{  to the next are governed by the following simple rules:            }
{                                                                     }
{     1. Every organism with two or three neighboring organisms       }
{        survives into the next generation.                           }
{     2. Every organism with more than three neighbors dies from      }
{        overcrowding.  Every organism with fewer than two neighbors  }
{        dies from isolation.                                         }
{     3. A new organism is born in every empty cell that has exactly  }
{        three neighboring organisms.                                 }
{                                                                     }
{  Further details and references about the game can be found in two  }
{  "Mathematical Games" columns by Martin Gardner (Scientific         }
{  American, October 1970 and February 1971), as well as in many      }
{  other sources.                                                     }
{                                                                     }
{  This program plays Life on a grid that has 18 rows and 79 columns. }
{  The left and right edges of the grid are considered adjacent --    }
{  for example, the cell in row 10, column 1 is adjacent to the cell  }
{  in row 10, column 79.  Similarly, the top and bottom edges of the  }
{  grid are considered adjacent.  This ensures that every cell in the }
{  grid has exactly eight neighboring cells.                          }
{                                                                     }
{  The grid is represented in the program as a two-dimensional array  }
{  of characters.  Cells that contain an organism are represented     }
{  with the character '*'; cells with no organism contain a blank.    }
{                                                                     }
{  At the beginning of the simulation, organisms are placed into      }
{  cells in a random pattern.  The probability that any particular    }
{  cell contains an organism is given by the constant InitialFraction.}
{  The program displays the initial configuration of cells, and then  }
{  runs the simulation for a number of generations specified by the   }
{  user, displaying each generation.  This process continues until    }
{  the user terminates it by requesting no more generations.          }
{                                                                     }
{  Input:  The number of generations of the simulation to run is      }
{          entered by the user.                                       }
{                                                                     }
{  Output: Each generation is displayed on the terminal.              }
{                                                                     }
{  This program contains non-standard Pascal statements to support    }
{  Turbo Pascal and Berkeley Pascal versions.  Non-standard Pascal    }
{  statements appear in the following locations:                      }
{    1. in the 'uses' clause immediately following this comment block }
{    2. in the compiler directives to include "screen.i"              }
{    3. in the random number generation function RandomNumber and     }
{       the random number initialization procedure InitRandom.        }
{    4. in the screen display procedure Display
{  These non-standard statements are preceded by a comment that       }
{  begins with '## N'.                                                }
{---------------------------------------------------------------------}

  { ## Non-standard statements -- Turbo Pascal version }

  uses Crt;

  { End Turbo Pascal version }

const
  CellRows = 18;                  { number of rows in array of cells }
  CellCols = 79;                  { number of columns in array of cells }
  InitialFraction = 0.10;         { initial fraction of cells occupied }
  StartLine = 3;                  { first line on terminal screen to be
                                    used in displaying the array of cells }

type
  CellArray = array[1..CellRows,1..CellCols] of char;
  RowSubscript = 1..CellRows;
  ColSubscript = 1..CellCols;

var
  Cells : CellArray;              { current generation of organisms }
  OldCells : CellArray;           { previous generation of organisms }
  Row : RowSubscript;             { Row and Col are used as subscripts }
  Col : ColSubscript;             {    for Cells and OldCells          }
  NumberOfCycles : integer;       { number of generations user wants to run }
  Cycle : integer;                { current generation number }

{*********************************************************************}

{---------------------------------------------------------------------}
{  The following compiler directives include the file "screen.i",     }
{  which contains the following procedures:                           }
{    ClearScreen - clear terminal screen and home cursor              }
{    MoveCursor - move cursor to a specified location                 }
{    ClearLine - clear the line on which the cursor is positioned     }
{---------------------------------------------------------------------}

{$I screen.i} { for Turbo Pascal; treated as comment in standard Pascal }

{ ## Non-standard statements -- Berkeley Pascal version }
{
#include "screen.i"
}
{ End Berkeley Pascal version }

{*********************************************************************}

function RandomNumber : real;

  {-------------------------------------------------------------------}
  {  RandomNumber returns a random real number between 0 and 1.  The  }
  {  random numbers generated are uniformly distributed in this       }
  {  range.  Before calling RandomNumber, the random-number generator }
  {  should be initialized by a call to procedure InitRandom.         }
  {                                                                   }
  {  RandomNumber calls a non-standard Pascal library routine to      }
  {  generate the random numbers required.                            }
  {                                                                   }
  {  Function value: a random real number between 0 and 1             }
  {-------------------------------------------------------------------}

  begin  { RandomNumber }

    { ## Non-standard statements -- Turbo Pascal version }

    RandomNumber := Random

    { End Turbo Pascal version }

    { ## Non-standard statements -- Berkeley Pascal version }
    {
    RandomNumber := random(1.0)
    }
    { End Berkeley Pascal version }

  end;  { RandomNumber }

{*********************************************************************}

procedure InitRandom(SeedValue : integer);

  {-------------------------------------------------------------------}
  {  InitRandom initializes the random-number generator.  The         }
  {  sequence of random numbers generated is determined by the        }
  {  parameter SeedValue, if this value is not zero.  If SeedValue    }
  {  is zero, the random-number generator is initialized with a       }
  {  value derived from the system clock.  This causes a different    }
  {  sequence of random numbers to be generated each time the program }
  {  is executed.  The procedure calls a non-standard Pascal library  }
  {  routine to perform this operation.                               }
  {                                                                   }
  {  Input parameters:                                                }
  {    SeedValue - a value to be used in initializing the random      }
  {           number generator, or zero                               }
  {                                                                   }
  {  Output parameters: none                                          }
  {                                                                   }
  {  Side effect:  The random number generator is initialized,        }
  {                depending upon the value of SeedValue              }
  {-------------------------------------------------------------------}

  var
    Dummy : integer;        { a dummy variable required by the syntax of
                              Berkeley Pascal -- value is never used }

  begin  { InitRandom }

    { ## Non-standard statements -- Turbo Pascal version }

    if SeedValue = 0 then
      Randomize
    else
      RandSeed := SeedValue

    { End Turbo Pascal version }

    { ## Non-standard statements -- Berkeley Pascal version }
    {
    if SeedValue = 0 then
      Dummy := seed(wallclock)
    else
      Dummy := seed(SeedValue)
    }
    { End Berkeley Pascal version }

  end;  { InitRandom }

{*********************************************************************}

procedure Initialize(var Cells : CellArray; var OldCells : CellArray);

  {-------------------------------------------------------------------}
  {  Initialize is called once at the beginning of the program.  It   }
  {  clears the terminal screen and displays a message that gives     }
  {  the value of the constant InitialFraction (the probability that  }
  {  a cell contains an organism at the start of the simulation).     }
  {                                                                   }
  {  For each cell in Cells, a random number between 0 and 1 is       }
  {  obtained.  If the random number is less than InitialFraction,    }
  {  the cell is set to '*' (indicating an organism).  Otherwise,     }
  {  the cell is set to blank.  All of the cells in OldCells (the     }
  {  previous generation of cells) are set to blank.                  }
  {                                                                   }
  {  Input parameters: None                                           }
  {                                                                   }
  {  Output parameters:                                               }
  {    Cells - initialized to a random pattern of organisms           }
  {    OldCells - set to all blanks                                   }
  {-------------------------------------------------------------------}

  var
    Row : RowSubscript;      { Row and Col are used as subscripts }
    Col : ColSubscript;      {    for Cells and OldCells          }

  begin  { Initialize }

    { clear screen and display header information }

    ClearScreen;
    writeln('Life simulation with initial fraction of cells occupied = ',
             InitialFraction:4:2);

    { initialize all entries in OldCells to blank }

    for Row := 1 to CellRows do
      for Col := 1 to CellCols do
          OldCells[Row,Col] := ' ';

    { initialize the random number generator }

    InitRandom(0);

    { initialize each entry in Cells to '*' if the random number selected
      is less than InitialFraction; otherwise initialize the entry to blank }

    for Row := 1 to CellRows do
      for Col := 1 to CellCols do
        begin

          { note: RandomNumber is a function that generates a different
            random value each time this statement is executed }

          if RandomNumber < InitialFraction then
            Cells[Row,Col] := '*'
          else
            Cells[Row,Col] := ' '
        end  { for }

  end;  { Initialize }

{*********************************************************************}

procedure Display(var Cells : CellArray; var OldCells : CellArray;
                  StartLine : integer);

  {-------------------------------------------------------------------}
  {  Display displays the contents of the array Cells on the          }
  {  terminal, one row per line of the screen.  The first row of      }
  {  cells begins on the line specified by the parameter StartLine.   }
  {                                                                   }
  {  After displaying the contents of Cells, this version of the      }
  {  procedure delays further execution of the program by a specified }
  {  number of milliseconds.  This delay is introduced to adjust the  }
  {  pace of the screen display on fast machines.  The number of      }
  {  milliseconds of delay is specified by the constant DelayCount,   }
  {  which can be changed to provide satisfactory operation on your   }
  {  machine.  The delay is accomplished by calling the nonstandard   }
  {  Turbo Pascal procedure Delay.                                    }
  {                                                                   }
  {  It is assumed that the contents of the array OldCells have       }
  {  already been displayed, and that this display remains on the     }
  {  terminal screen.  Display attempts to avoid re-displaying        }
  {  cells that have not changed from the previous version.  It       }
  {  does this by scanning each row of Cells, comparing the cell      }
  {  contents to those stored in OldCells.                            }
  {                                                                   }
  {  If the contents of a cell have changed, an internal procedure    }
  {  DisplaySegment is called to display a segment of the row that    }
  {  begins with the changed cell.  The scan then continues, skipping }
  {  over the segment that was displayed.                             }
  {                                                                   }
  {  Before Display is called for the first time, the terminal screen }
  {  should be cleared to blanks, and OldCells should be initialized  }
  {  to blanks.  This insures that the first version of the array     }
  {  will be correctly displayed.                                     }
  {                                                                   }
  {  Input parameters:                                                }
  {    Cells - the array whose contents are to be displayed           }
  {    OldCells - the previous version of the array, or blanks if     }
  {               this is the first call to Display                   }
  {      Note: Cells and OldCells are var parameters for efficiency   }
  {            in parameter passing.                                  }
  {    StartLine - the line on the screen where the display of Cells  }
  {               is to begin                                         }
  {                                                                   }
  {  Output parameters: None                                          }
  {-------------------------------------------------------------------}

  const
    DelayCount = 250;             { number of milliseconds to delay }

  var
    Row : RowSubscript;           { Row and Col are used as subscripts }
    Col : ColSubscript;           {    for Cells and OldCells          }
    Length : integer;             { the length of the segment displayed }
    EndOfRow : boolean;           { the end of a row has been reached }

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

  procedure DisplaySegment(var Cells: CellArray; var OldCells : CellArray;
                           Row : RowSubscript; Col : ColSubscript;
                           var Length : integer; StartLine : integer);

  {-------------------------------------------------------------------}
  {  DisplaySegment is called each time Display finds a cell that     }
  {  has changed in value from the previous version.  It scans the    }
  {  row containing the changed cell to find the segment to be        }
  {  displayed.  This scan begins with the changed cell and           }
  {  continues until the end of the row is reached, or until six      }
  {  consecutive unchanged cells are found in the row.                }
  {                                                                   }
  {  (This rule for constructing the segment to display takes into    }
  {  account a typical amount of overhead involved in moving          }
  {  the cursor to a new location.  On most terminals, it is more     }
  {  efficient to display a small number of unchanged cells than to   }
  {  reposition the cursor to skip over them.)                        }
  {                                                                   }
  {  The segment that is found is displayed on the terminal, after    }
  {  moving the cursor to the proper starting location.  The length   }
  {  of the segment displayed is returned to the calling procedure.   }
  {                                                                   }
  {  Input parameters:                                                }
  {    Cells - the array whose contents are being displayed           }
  {    OldCells - the previous version of the array, or blanks if     }
  {               this is the first call to DisplaySegment            }
  {      Note: Cells and OldCells are var parameters for efficiency   }
  {            in parameter passing.                                  }
  {    Row,Col - the row and column in which a changed cell occurs    }
  {    StartLine - the line on the screen where the display of Cells  }
  {               begins                                              }
  {                                                                   }
  {  Output parameters:                                               }
  {    Length - the length of the segment displayed                   }
  {-------------------------------------------------------------------}

    var
      CurrentCol : ColSubscript;     { used as subscript for scanning the row }
      LastChangedCol : ColSubscript; { last changed cell found in the row }
      UnchangedCells : integer;      { number of consecutive unchanged cells }
      WriteCol : ColSubscript;       { used as subscript to display segment }
      EndOfSegment : boolean;        { the end of the segment to display
                                       has been found }

    begin  { DisplaySegment }

      { start scanning with the changed column passed as parameter }

      LastChangedCol := Col;
      CurrentCol := Col;
      UnchangedCells := 0;
      EndOfSegment := false;

      { scan cells until the end of the row is reached, or until six
        consecutive unchanged cells are found }

      repeat  { until EndOfSegment }
        if CurrentCol = CellCols then
          EndOfSegment := true
        else
          begin

            { not at end of row -- check whether next cell is changed }

            CurrentCol := CurrentCol + 1;
            if Cells[Row,CurrentCol] = OldCells[Row,CurrentCol] then
              begin

                { count unchanged cell; stop scanning if six unchanged
                  cells have been found }

                UnchangedCells := UnchangedCells + 1;
                if UnchangedCells = 6 then
                  EndOfSegment := true
              end  { if cell is unchanged }
            else
              begin

                { remember where the changed cell was found and reset the
                  count of consecutive unchanged cells to zero }

                LastChangedCol := CurrentCol;
                UnchangedCells := 0
              end  { else -- cell is changed }
          end  { else -- not at end of row }
      until EndOfSegment;

      { position cursor and display segment, ending with last changed cell
        that was found }

      MoveCursor(StartLine + Row - 1,Col);
      for WriteCol := Col to LastChangedCol do
        write(Cells[Row,WriteCol]);
      writeln;

      { set length of segment that was displayed }

      Length := LastChangedCol - Col + 1

    end;  { DisplaySegment }

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

  begin  { Display }

    { scan each row of the cell array, looking for a cell that has
      changed from the previous generation }

    for Row := 1 to CellRows do
      begin
        Col := 1;
        EndOfRow := false;
        repeat  { until EndOfRow }
          if Cells[Row,Col] = OldCells[Row,Col] then
            begin

              { cell is unchanged -- go on to next cell, unless at end of row }

              if Col = CellCols then
                EndOfRow := true
              else
                Col := Col + 1

            end  { if cell is unchanged }
          else
            begin

              { a changed cell was found -- display a segment starting with
                that cell and skip over the segment displayed }

              DisplaySegment(Cells,OldCells,Row,Col,Length,StartLine);
              if (Col + Length) > CellCols then
                EndOfRow := true
              else
                Col := Col + Length
            end  { else -- cell is changed }
        until EndOfRow
      end;  { for }

    { introduce a delay to adjust the pace of the display for fast machines }

    { ## Non-standard statements -- Turbo Pascal version }

    Delay(DelayCount)

    { End Turbo Pascal version }

  end;  { Display }

{*********************************************************************}

procedure GetCount(var NumberOfCycles : integer);

  {-------------------------------------------------------------------}
  {  GetCount is called to get instructions from the user about the   }
  {  number of generations to run the simulation.  It moves the       }
  {  cursor to a position below the displayed cells, and prompts the  }
  {  user to enter the number of generations desired.  It then reads  }
  {  an integer from the terminal and returns this integer as the     }
  {  parameter NumberOfCycles.                                        }
  {                                                                   }
  {  Input parameters: None                                           }
  {                                                                   }
  {  Output parameters:                                               }
  {    NumberOfCycles - the number of generations to run              }
  {-------------------------------------------------------------------}

  begin  { GetCount }

    MoveCursor(22,1);
    ClearLine;
    write('Enter number of generations to run, or 0 to stop: ');
    readln(NumberOfCycles)

  end;  { GetCount }

{*********************************************************************}

procedure NextGeneration(var Cells : CellArray; var OldCells : CellArray);

  {-------------------------------------------------------------------}
  {  NextGeneration applies the rules for birth, death, and survival  }
  {  of organisms to create the next generation (Cells) from the      }
  {  previous generation (OldCells).                                  }
  {                                                                   }
  {  NextGeneration scans the cells of the previous generation,       }
  {  counting the number of organisms that are neighbors of each      }
  {  cell.  (An internal function NeighborCount is used for this      }
  {  purpose.)   Empty cells with exactly three neighbors have an     }
  {  organism born in the next generation; otherwise they remain      }
  {  empty.  Organisms with fewer than two or more than three         }
  {  neighbors die (the cell becomes empty in the next generation);   }
  {  otherwise they survive.                                          }
  {                                                                   }
  {  Input parameters:                                                }
  {    OldCells - the previous generation of organisms                }
  {        Note: OldCells is a var parameter for efficiency in        }
  {              parameter passing.                                   }
  {                                                                   }
  {  Output parameters:                                               }
  {    Cells - the next generation of organisms                       }
  {-------------------------------------------------------------------}

  var
    Row : RowSubscript;      { Row and Col are used as subscripts }
    Col : ColSubscript;      {    for Cells and OldCells          }
    Neighbors : integer;     { the number of neighboring organisms for a cell }

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

function NeighborCount(var OldCells : CellArray; Row : RowSubscript;
                       Col : ColSubscript) : integer;

  {-------------------------------------------------------------------}
  {  NeighborCount counts the number of organisms that are neighbors  }
  {  of a specified cell at location [Row,Col] in the previous        }
  {  generation.  It examines all cells that lie in an area with row  }
  {  number between (Row - 1) and (Row + 1), and column number        }
  {  between (Col - 1) and (Col + 1).  The organisms that exist in    }
  {  those cells (except for the original cell at [Row,Col]) are      }
  {  counted to find the number of neighbors of the cell.             }
  {                                                                   }
  {  During this process, the left and right edges of the grid are    }
  {  considered to be adjacent.  Thus, if (Col + 1) is greater then   }
  {  CellCols (i.e., beyond the right edge of the grid), the cell     }
  {  actually examined is in column 1 (i.e., at the left edge).  If   }
  {  (Col - 1) is less than 1 (i.e., beyond the left edge of the      }
  {  grid), the cell actually examined is at CellCols  (i.e., at the  }
  {  right edge).  Likewise, the top and bottom rows are considered   }
  {  to be adjacent, with similar adjustments made if (Row + 1) is    }
  {  greater than CellRows or (Row - 1) is less than 1.               }
  {                                                                   }
  {  Input parameters:                                                }
  {    OldCells - the previous generation of organisms                }
  {        Note: OldCells is a var parameter for efficiency in        }
  {              parameter passing.                                   }
  {    Row,Col - the row and column of the cell whose neighbors are   }
  {              to be counted                                        }
  {                                                                   }
  {  Output parameters: None                                          }
  {                                                                   }
  {  Function value: the number of neighbors of the cell at [Row,Col] }
  {-------------------------------------------------------------------}

  var
    TryRow : integer;              { the row and column of a cell to be       }
    TryCol : integer;              {    examined, before adjusting for the    }
                                   {    edges of the grid                     }

    NeighborRow : RowSubscript;    { the row and column of a cell to be       }
    NeighborCol : ColSubscript;    {    examined, after adjusting for the     }
                                   {    edges of the grid                     }

    Neighbors : integer;           { number of neighbors counted }

  begin  { NeighborCount }

    Neighbors := 0;

    { examine cells in the neighborhood of the cell specified }

    for TryRow := (Row - 1) to (Row + 1) do
      for TryCol := (Col - 1) to (Col + 1) do
        begin

          { adjust row and column numbers for the edges of the grid }

          if TryRow < 1 then
            NeighborRow := CellRows
          else if TryRow > CellRows then
            NeighborRow := 1
          else
            NeighborRow := TryRow;

          if TryCol < 1 then
            NeighborCol := CellCols
          else if TryCol > CellCols then
            NeighborCol := 1
          else
            NeighborCol := TryCol;

          { count organisms, except for the one in the original cell }

          if (NeighborRow <> Row) or (NeighborCol <> Col) then
            if OldCells[NeighborRow,NeighborCol] = '*' then
              Neighbors := Neighbors +  1
        end;  { for }

    { set function value }

    NeighborCount := Neighbors

  end;  { NeighborCount }

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

  begin  { NextGeneration }

    { examine all cells, counting neighbors of each }

    for Row := 1 to CellRows do
      for Col := 1 to CellCols do
        begin
          Neighbors := NeighborCount(OldCells,Row,Col);
          if OldCells[Row,Col] = ' ' then
            begin

              { cell was empty in the previous generation -- it remains empty
                unless it had exactly three neighbors }

              if Neighbors = 3 then
                Cells[Row,Col] := '*'
              else
                Cells[Row,Col] := ' '
            end  { if cell was empty }
          else
            begin

              { cell had an organism in the previous generation -- it survives
                unless it had fewer than two, or more than three, neighbors }

              if (Neighbors < 2) or (Neighbors > 3) then
                Cells[Row,Col] := ' '
              else
                Cells[Row,Col] := '*'
            end  { else -- cell was not empty }
        end  { for }

  end;  { NextGeneration }

{*********************************************************************}

procedure CleanUp;

  {-------------------------------------------------------------------}
  {  CleanUp is called once at the end of the program.  It waits for  }
  {  the user to enter a carriage return before exiting the program.  }
  {-------------------------------------------------------------------}

  begin  { CleanUp }

    write('Press return to exit');
    readln

  end;  { CleanUp }

{*********************************************************************}

begin  { main program }

  { initialize and display the first generation }

  Initialize(Cells,OldCells);
  Display(Cells,OldCells,StartLine);

  { run the simulation for a number of generations entered by the user,
    continuing until the user requests zero generations }

  GetCount(NumberOfCycles);
  while NumberOfCycles > 0 do
    begin
      for Cycle := 1 to NumberOfCycles do
        begin

          { copy current generation to previous generation }

          for Row := 1 to CellRows do
            for Col := 1 to CellCols do
              OldCells[Row,Col] := Cells[Row,Col];

          { generate and display the next generation }

          NextGeneration(Cells,OldCells);
          Display(Cells,OldCells,StartLine)
        end;  { for }

      { does the user want to continue the simulation? }

      GetCount(NumberOfCycles)
    end;  { while }
  CleanUp

end.
