procedure Display(var Cells : CellArray; var OldCells : CellArray;
                  StartLine : integer);

  {-------------------------------------------------------------------}
  {  Display displays the contents of the array Cells on the          }
  {  terminal, one row per line of the screen.  The first row of      }
  {  cells begins on the line specified by the parameter StartLine.   }
  {                                                                   }
  {  After displaying the contents of Cells, this version of the      }
  {  procedure delays further execution of the program by a specified }
  {  number of milliseconds.  This delay is introduced to adjust the  }
  {  pace of the screen display on fast machines.  The number of      }
  {  milliseconds of delay is specified by the constant DelayCount,   }
  {  which can be changed to provide satisfactory operation on your   }
  {  machine.  The delay is accomplished by calling the nonstandard   }
  {  Turbo Pascal procedure Delay.                                    }
  {                                                                   }
  {  It is assumed that the contents of the array OldCells have       }
  {  already been displayed, and that this display remains on the     }
  {  terminal screen.  Display attempts to avoid re-displaying        }
  {  cells that have not changed from the previous version.  It       }
  {  does this by scanning each row of Cells, comparing the cell      }
  {  contents to those stored in OldCells.                            }
  {                                                                   }
  {  If the contents of a cell have changed, an internal procedure    }
  {  DisplaySegment is called to display a segment of the row that    }
  {  begins with the changed cell.  The scan then continues, skipping }
  {  over the segment that was displayed.                             }
  {                                                                   }
  {  Before Display is called for the first time, the terminal screen }
  {  should be cleared to blanks, and OldCells should be initialized  }
  {  to blanks.  This insures that the first version of the array     }
  {  will be correctly displayed.                                     }
  {                                                                   }
  {  Input parameters:                                                }
  {    Cells - the array whose contents are to be displayed           }
  {    OldCells - the previous version of the array, or blanks if     }
  {               this is the first call to Display                   }
  {      Note: Cells and OldCells are var parameters for efficiency   }
  {            in parameter passing.                                  }
  {    StartLine - the line on the screen where the display of Cells  }
  {               is to begin                                         }
  {                                                                   }
  {  Output parameters: None                                          }
  {                                                                   }
  {  In order to use this procedure, the main program must contain    }
  {  the following constant and type definitions.  Items enclosed     }
  {  in <<...>> are to be supplied by the programmer.                 }
  {                                                                   }
  {    const                                                          }
  {      CellRows = <<integer between 1 and 20>>;                     }
  {      CellCols = <<integer between 1 and 79>>;                     }
  {                                                                   }
  {    type                                                           }
  {      CellArray = array[1..CellRows,1..CellCols] of char;          }
  {      RowSubscript = 1..CellRows;                                  }
  {      ColSubscript = 1..CellCols;                                  }
  {                                                                   }
  {  In addition, the main program must include the terminal output   }
  {  routines from the file "screen.i".                               }
  {-------------------------------------------------------------------}

  const
    DelayCount = 250;             { number of milliseconds to delay }

  var
    Row : RowSubscript;           { Row and Col are used as subscripts }
    Col : ColSubscript;           {    for Cells and OldCells          }
    Length : integer;             { the length of the segment displayed }
    EndOfRow : boolean;           { the end of a row has been reached }

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

  procedure DisplaySegment(var Cells: CellArray; var OldCells : CellArray;
                           Row : RowSubscript; Col : ColSubscript;
                           var Length : integer; StartLine : integer);

  {-------------------------------------------------------------------}
  {  DisplaySegment is called each time Display finds a cell that     }
  {  has changed in value from the previous version.  It scans the    }
  {  row containing the changed cell to find the segment to be        }
  {  displayed.  This scan begins with the changed cell and           }
  {  continues until the end of the row is reached, or until six      }
  {  consecutive unchanged cells are found in the row.                }
  {                                                                   }
  {  (This rule for constructing the segment to display takes into    }
  {  account a typical amount of overhead involved in moving the      }
  {  the cursor to a new location.  On most terminals, it is more     }
  {  efficient to display a small number of unchanged cells than to   }
  {  reposition the cursor to skip over them.)                        }
  {                                                                   }
  {  The segment that is found is displayed on the terminal, after    }
  {  moving the cursor to the proper starting location.  The length   }
  {  of the segment displayed is returned to the calling procedure.   }
  {                                                                   }
  {  Input parameters:                                                }
  {    Cells - the array whose contents are being displayed           }
  {    OldCells - the previous version of the array, or blanks if     }
  {               this is the first call to DisplaySegment            }
  {      Note: Cells and OldCells are var parameters for efficiency   }
  {            in parameter passing.                                  }
  {    Row,Col - the row and column in which a changed cell occurs    }
  {    StartLine - the line on the screen where the display of Cells  }
  {               begins                                              }
  {                                                                   }
  {  Output parameters:                                               }
  {    Length - the length of the segment displayed                   }
  {-------------------------------------------------------------------}

    var
      CurrentCol : ColSubscript;     { used as subscript for scanning the row }
      LastChangedCol : ColSubscript; { last changed cell found in the row }
      UnchangedCells : integer;      { number of consecutive unchanged cells }
      WriteCol : ColSubscript;       { used as subscript to display segment }
      EndOfSegment : boolean;        { the end of the segment to display
                                       has been found }

    begin  { DisplaySegment }

      { start scanning with the changed column passed as parameter }

      LastChangedCol := Col;
      CurrentCol := Col;
      UnchangedCells := 0;
      EndOfSegment := false;

      { scan cells until the end of the row is reached, or until six
        consecutive unchanged cells are found }

      repeat  { until EndOfSegment }
        if CurrentCol = CellCols then
          EndOfSegment := true
        else
          begin

            { not at end of row -- check whether next cell is changed }

            CurrentCol := CurrentCol + 1;
            if Cells[Row,CurrentCol] = OldCells[Row,CurrentCol] then
              begin

                { count unchanged cell; stop scanning if six unchanged
                  cells have been found }

                UnchangedCells := UnchangedCells + 1;
                if UnchangedCells = 6 then
                  EndOfSegment := true
              end  { if cell is unchanged }
            else
              begin

                { remember where the changed cell was found and reset the
                  count of consecutive unchanged cells to zero }

                LastChangedCol := CurrentCol;
                UnchangedCells := 0
              end  { else -- cell is changed }
          end  { else -- not at end of row }
      until EndOfSegment;

      { position cursor and display segment, ending with last changed cell
        that was found }

      MoveCursor(StartLine + Row - 1,Col);
      for WriteCol := Col to LastChangedCol do
        write(Cells[Row,WriteCol]);
      writeln;

      { set length of segment that was displayed }

      Length := LastChangedCol - Col + 1

    end;  { DisplaySegment }

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

  begin  { Display }

    { scan each row of the cell array, looking for a cell that has
      changed from the previous generation }

    for Row := 1 to CellRows do
      begin
        Col := 1;
        EndOfRow := false;
        repeat  { until EndOfRow }
          if Cells[Row,Col] = OldCells[Row,Col] then
            begin

              { cell is unchanged -- go on to next cell, unless at end of row }

              if Col = CellCols then
                EndOfRow := true
              else
                Col := Col + 1

            end  { if cell is unchanged }
          else
            begin

              { a changed cell was found -- display a segment starting with
                that cell and skip over the segment displayed }

              DisplaySegment(Cells,OldCells,Row,Col,Length,StartLine);
              if (Col + Length) > CellCols then
                EndOfRow := true
              else
                Col := Col + Length
            end  { else -- cell is changed }
        until EndOfRow
      end;  { for }

    { introduce a delay to adjust the pace of the display for fast machines }

    { ## Non-standard statements -- Turbo Pascal version }

    Delay(DelayCount)

    { End Turbo Pascal version }

  end;  { Display }
