{---------------------------------------------------------------------}
{  The following procedures and function implement operations on      }
{  the abstract data type ListOfRecords.  This data type can be used  }
{  to store and retrieve a collection of records. The fields          }
{  contained in these records can be defined by the programmer.       }
{                                                                     }
{  Procedures:                                                        }
{    InitList - initialize a new (empty) list                         }
{    GetFirst - get the first record (if any) in a list               }
{    GetNext - get the next record (if any) in a list                 }
{    UpdateCurrent - update the last record read from a list          }
{    DeleteCurrent - delete the last record read from a list          }
{    AddNew - add a new record to a list                              }
{                                                                     }
{  Function:                                                          }
{    NumberIn - returns the number of records in a list               }
{                                                                     }
{  The records in a list can be retrieved, one at a time, by using    }
{  the operations GetFirst and GetNext.  GetFirst begins a new scan   }
{  of the records in the list and retrieves the first record.  Each   }
{  record after the first is retrieved with GetNext.  When no more    }
{  records remain to be retrieved, the parameter EndOfList is set to  }
{  true.                                                              }
{                                                                     }
{  This sequence of operations is guaranteed to retrieve every record }
{  in the list before terminating.  However, the order in which the   }
{  records are retrieved is not defined, and may vary from one scan   }
{  of the list to the next.                                           }
{                                                                     }
{  The record most recently retrieved is designated as the current    }
{  record of the list.  This record can be modified by using the      }
{  operation UpdateCurrent, or deleted from the list by using the     }
{  operation DeleteCurrent.  It is not possible to modify or delete   }
{  records other than the current record.                             }
{                                                                     }
{  New records can be added to the list at any time with the          }
{  operation AddNew.  If a new record is added during a scan of the   }
{  list, it will be retrieved later during that same scan.            }
{                                                                     }
{  In order to use these routines, the main program must contain the  }
{  following constant and type definitions.  Items enclosed in        }
{  <<...>> are to be supplied by the programmer.                      }
{                                                                     }
{    const                                                            }
{      MaximumListSize = << maximum size of a list >>                 }
{                                                                     }
{    type                                                             }
{      ListEntry = record                                             }
{                    << definitions of fields in a record >>          }
{                  end;                                               }
{                                                                     }
{      ListOfRecords = record                                         }
{                MaximumSize : 1..MaximumListSize;                    }
{                CurrentSize : integer;                               }
{                Entries : array[1..MaximumListSize] of ListEntry     }
{                CurrentEntry : integer;                              }
{                NextEntry : integer;                                 }
{              end;                                                   }
{                                                                     }
{  In this implementation of ListOfRecords, a list is represented by  }
{  a record that contains the following fields:                       }
{                                                                     }
{    MaximumSize - the maximum number of records in the list          }
{    CurrentSize - the number of records currently in the list        }
{    Entries - an array containing the records in the list            }
{    CurrentEntry - the subscript of the current record in Entries    }
{    NextEntry - the subscript of the next record in Entries          }
{                                                                     }
{  CurrentEntry is 0 if the current entry has been deleted.           }
{  NextEntry is 0 if the current record is the last one in the list.  }
{  Both CurrentEntry and NextEntry are 0 if no records have yet been  }
{  retrieved.                                                         }
{                                                                     }
{  The records in the list always occupy contiguous elements of the   }
{  array Entries.  The first record occupies the element with         }
{  subscript 1, and the last record occupies the element with         }
{  subscript CurrentSize.  When a record is deleted from the list,    }
{  the remaining records are rearranged as needed to maintain this    }
{  arrangement.                                                       }
{---------------------------------------------------------------------}


procedure InitList (var List : ListOfRecords; ListLimit : integer);

  {-------------------------------------------------------------------}
  {  InitList initializes a new (empty) list, setting its maximum     }
  {  size to the value specified in ListLimit.  If the list size      }
  {  specified is too large, an error message is displayed and the    }
  {  maximum allowable list size is used.                             }
  {                                                                   }
  {  Input parameters:                                                }
  {    List - the list to be initialized                              }
  {    ListLimit - the maximum size of the list                       }
  {                                                                   }
  {  Output parameters:                                               }
  {    List - the initialized list                                    }
  {-------------------------------------------------------------------}

  begin  { InitList }

    if ListLimit > MaximumListSize then
      begin
        writeln('*** InitList: specified size is greater than maximum');
        List.MaximumSize := MaximumListSize
      end  { if }
    else
      List.MaximumSize := ListLimit;
    List.CurrentSize := 0;
    List.CurrentEntry := 0;
    List.NextEntry := 0

  end;  { InitList }

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

procedure GetFirst(var List : ListOfRecords; var Entry : ListEntry;
                    var EndOfList : boolean);

  {-------------------------------------------------------------------}
  {  GetFirst gets the first entry (if any) in List, returning it in  }
  {  the parameter Entry.  The entry retrieved is made the current    }
  {  entry for the list.  The following entry (if any) is made the    }
  {  next entry.  If the list contains no entries, EndOfList is set   }
  {  to true; otherwise, EndOfList is set to false.                   }
  {                                                                   }
  {  Input parameters:                                                }
  {    List - the list whose first entry is to be retrieved           }
  {                                                                   }
  {  Output parameters:                                               }
  {    List - current entry set to the entry retrieved, and next      }
  {           entry set to the following entry (if any)               }
  {    Entry - the first entry in List; this parameter remains        }
  {            unchanged if the list is empty                         }
  {    EndOfList - true if List is empty; false otherwise             }
  {-------------------------------------------------------------------}

  begin  { GetFirst }

    if List.CurrentSize = 0 then
      EndOfList := true
    else
      begin
        List.CurrentEntry := 1;
        if List.CurrentEntry = List.CurrentSize then
          List.NextEntry := 0
        else
          List.NextEntry := List.CurrentEntry + 1;
        Entry := List.Entries[List.CurrentEntry];
        EndOfList := false
      end  { else }

  end;  { GetFirst }

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

procedure GetNext(var List : ListOfRecords; var Entry : ListEntry;
                    var EndOfList : boolean);

  {-------------------------------------------------------------------}
  {  GetNext gets the next entry in List (if any), returning it in    }
  {  the parameter Entry.  If there is no next entry in the list,     }
  {  EndOfList is set to true; otherwise, EndOfList is set to false.  }
  {                                                                   }
  {  The entry retrieved (if any) is made the current entry for the   }
  {  list. The following entry (if any) is made the next entry.       }
  {                                                                   }
  {  Input parameters:                                                }
  {    List - the list whose next entry is to be retrieved            }
  {                                                                   }
  {  Output parameters:                                               }
  {    List - current entry set to the entry retrieved, and next      }
  {           entry set to the following entry (if any)               }
  {    Entry - the next entry in List; this parameter remains         }
  {            unchanged if there is no next entry to retrieve        }
  {    EndOfList - true if there was no next entry to retrieve;       }
  {                false otherwise                                    }
  {-------------------------------------------------------------------}

  begin  { GetNext }

    if List.NextEntry = 0 then
      EndOfList := true
    else
      begin
        List.CurrentEntry := List.NextEntry;
        if List.CurrentEntry = List.CurrentSize then
          List.NextEntry := 0
        else
          List.NextEntry := List.CurrentEntry + 1;
        Entry := List.Entries[List.CurrentEntry];
        EndOfList := false
      end

  end;  { GetNext }

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

procedure AddNew(var List : ListOfRecords; NewEntry : ListEntry);

  {-------------------------------------------------------------------}
  {  AddNew adds the entry NewEntry to List.  If the list is already  }
  {  at its maximum size, an error message is displayed and the       }
  {  new entry is not added.  The current list entry is not changed   }
  {  by this operation. If the next entry in the list was previously  }
  {  undefined, the new entry becomes the next entry.                 }
  {                                                                   }
  {  Input parameters:                                                }
  {    List - the list to which NewEntry is to be added               }
  {    NewEntry - the entry to add                                    }
  {                                                                   }
  {  Output parameters:                                               }
  {    List - with NewEntry added at the end of the list              }
  {-------------------------------------------------------------------}

  begin  { AddNew }

    if List.CurrentSize = List.MaximumSize then
      writeln('*** AddNew: list overflow')
    else
      begin
        List.CurrentSize := List.CurrentSize + 1;
        List.Entries[List.CurrentSize] := NewEntry;
        if List.NextEntry = 0 then
          List.NextEntry := List.CurrentSize
      end  { else }

  end;  { AddNew }

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

procedure DeleteCurrent(var List : ListOfRecords);

  {-------------------------------------------------------------------}
  {  DeleteCurrent deletes the current entry from List.  If there is  }
  {  no current entry defined, an error message is displayed. The     }
  {  current entry for the list becomes undefined after this          }
  {  operation.                                                       }
  {                                                                   }
  {  If the entry being deleted is not the last one in the list, the  }
  {  last entry in the list is shifted to replace the entry that was  }
  {  deleted.  (This is done to avoid leaving unused entries within   }
  {  the list.)  This entry becomes the new next entry in the list.   }
  {                                                                   }
  {  Input parameters:                                                }
  {    List - the list whose current entry is to be deleted           }
  {                                                                   }
  {  Output parameters:                                               }
  {    List - with the entry deleted                                  }
  {-------------------------------------------------------------------}

  begin  { DeleteCurrent }

    if List.CurrentEntry = 0 then
      writeln('*** DeleteCurrent: no current entry to delete')
    else
      begin
        if List.CurrentEntry < List.CurrentSize then
          begin
            List.Entries[List.CurrentEntry] := List.Entries[List.CurrentSize];
            List.NextEntry := List.CurrentEntry
          end; { if }
        List.CurrentSize := List.CurrentSize - 1;
        List.CurrentEntry := 0
      end  { else }

  end;  { DeleteCurrent }

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

procedure UpdateCurrent(var List : ListOfRecords; Entry : ListEntry);

  {-------------------------------------------------------------------}
  {  UpdateCurrent updates the current entry in List, replacing it    }
  {  with Entry.  If there is no current entry defined, an error      }
  {  message is displayed.  After this operation, the current entry   }
  {  remains the one that was updated (with the new contents).        }
  {                                                                   }
  {  Input parameters:                                                }
  {    List - the list whose current entry is to be updated           }
  {    Entry - the new contents for the current entry                 }
  {                                                                   }
  {  Output parameters:                                               }
  {    List - with the current entry updated                          }
  {-------------------------------------------------------------------}

  begin  { UpdateCurrent }

    if List.CurrentEntry = 0 then
      writeln('*** UpdateCurrent: no current entry to update')
    else
      List.Entries[List.CurrentEntry] := Entry

  end;  { UpdateCurrent }

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

function NumberIn(var List : ListOfRecords) : integer;

  {-------------------------------------------------------------------}
  {  NumberIn returns a function value equal to the current number of }
  {  entries in List.                                                 }
  {                                                                   }
  {  Input parameters:                                                }
  {    List - the list whose number of entries is to be returned      }
  {      Note: List is a var parameter for efficiency in parameter    }
  {            passing.                                               }
  {                                                                   }
  {  Output parameters: None                                          }
  {                                                                   }
  {  Function value: the current number of entries in List            }
  {-------------------------------------------------------------------}

  begin  { NumberIn }

    NumberIn := List.CurrentSize

  end;  { NumberIn }
