program Wator(input,output);

{---------------------------------------------------------------------}
{  This program simulates the ecological relationship of the two      }
{  dominant species on the unusual planet Wa-Tor.  This planet is     }
{  shaped like a torus (a doughnut), and is entirely covered by       }
{  water -- hence its name.  The two dominant species on Wa-Tor are   }
{  called sharks and fish, after the earthly creatures they most      }
{  closely resemble.  The sharks eat the fish, and the fish live on   }
{  microscopic organisms uniformly distributed throughout the water.  }
{                                                                     }
{  This program simulates the ocean of Wa-Tor on a rectangular grid   }
{  that has 18 rows and 79 columns.  Each cell in the grid can hold   }
{  one fish or one shark (or may be empty).  The left and right edges }
{  of the grid are considered adjacent --  for example, a shark at    }
{  right-hand side of the grid that swims to the right will reappear  }
{  at the left-hand side.  Similarly, the top and bottom edges of the }
{  grid are considered adjacent.  These adjacencies make the grid     }
{  equivalent to the torus shape of the planet.                       }
{                                                                     }
{  Time passes in discrete jumps called chronons.  During each        }
{  chronon, each fish swims to an adjacent empty cell (selected at    }
{  random).  If all adjacent cells are occupied, the fish does not    }
{  move.  After the fish move, each shark swims to an adjacent cell   }
{  occupied by a fish and eats the fish.  If there are no adjacent    }
{  cells containing fish, the shark moves at random in the same way   }
{  a fish does.  When each shark or fish reaches a certain age, it    }
{  produces a single offspring.  (Both species reproduce by cloning   }
{  themselves.)  If a shark does not find food within a certain       }
{  length of time, it starves.                                        }
{                                                                     }
{  At the beginning of the simulation, the program distributes sharks }
{  and fish in the ocean in a random pattern.  It then simulates the  }
{  swimming, breeding, and hunting behaviors described above for a    }
{  specified number of chronons.  After each move, the distribution   }
{  of sharks and fish is displayed on the terminal.  In this display, }
{  sharks are indicated by "o" and fish by ".".                       }
{                                                                     }
{  The procedure Display, which is contained in the file "display.i', }
{  is used to display the contents of the ocean.                      }
{                                                                     }
{  The program keeps track of the sharks and fish by maintaining a    }
{  list of each type of creature.  This list records the position of  }
{  the shark or fish, its age, and (for sharks) the length of time    }
{  since it has eaten.  An abstract data type ListOfRecords is used   }
{  to store these lists.  The procedures and function that implement  }
{  ListOfRecords are contained in the file "reclist.i".               }
{                                                                     }
{  The simulation is governed by the following parameters, which may  }
{  be set by the user at the beginning of the run:                    }
{                                                                     }
{    FishInitial  - the initial number of fish in the ocean           }
{    FishBreed    - the age (in chronons) at which a fish reproduces  }
{    FishLimit    - the maximum number of fish (limit imposed by the  }
{                   amount of fish food available)                    }
{    SharkInitial - the initial number of sharks in the ocean         }
{    SharkBreed   - the age (in chronons) at which a shark reproduces }
{    SharkStarve  - the length of time (in chronons) a shark can      }
{                   survive without eating                            }
{                                                                     }
{  Default values are used for any of these parameters that the user  }
{  does not specify.                                                  }
{                                                                     }
{  Further details and discussions about Wa-Tor can be found in a     }
{  "Computer Recreations" column by A. K. Dewdney (Scientific         }
{  American, December 1984.)  The simulation performed by this        }
{  program is similar to the one discussed there.                     }
{                                                                     }
{  Input:  The parameters for the simulation are entered by the user  }
{          from the terminal.                                         }
{                                                                     }
{  Output: The distribution of sharks and fish, and the total number  }
{          of each, are displayed on the terminal after each move.    }
{                                                                     }
{  This program contains non-standard Pascal statements to support    }
{  Turbo Pascal and Berkeley Pascal versions.  Non-standard Pascal    }
{  statements appear in the following locations:                      }
{    1. in the 'uses' clause immediately following this comment block }
{    2. in the compiler directives to include "screen.i", "random.i", }
{       "display.i", and "reclist.i"                                  }
{    3. in the function RandomInteger                                 }
{  These non-standard statements are preceded by a comment that       }
{  begins with '## N'.                                                }
{---------------------------------------------------------------------}

  { ## Non-standard statements -- Turbo Pascal version }

  uses Crt;

  { End Turbo Pascal version }

const

  { default values for the simulation parameters }

  FishInitialDefault = 200;
  FishBreedDefault = 3;
  FishLimitDefault = 400;
  SharkInitialDefault = 20;
  SharkBreedDefault = 3;
  SharkStarveDefault = 3;

  { end default values }

  CellRows = 18;               { number of rows in the ocean array }
  CellCols = 79;               { number of columns in the ocean array }

  FishSymbol = '.';            { character used to display a fish }
  SharkSymbol = 'o';           { character used to display a shark }
  PauseAfter = 50;             { number of chronons to run before pausing }
  MaximumListSize = 1422;      { maximum size of a list of sharks or fish --
                                 should be equal to CellRows * CellCols }

type

  CellArray = array[1..CellRows,1..CellCols] of char;
  RowSubscript = 1..CellRows;
  ColSubscript = 1..CellCols;

  ListEntry = record                   { entry in list of sharks or fish }
                Row : RowSubscript;      { row and column position }
                Col : ColSubscript;
                Age : integer;           { age of the shark or fish }
                TimeSinceFed : integer   { time since shark has eaten }
              end;  { ListEntry }

  ListOfRecords = record
                    MaximumSize : 1..MaximumListSize;
                    CurrentSize : integer;
                    Entries : array[1..MaximumListSize] of ListEntry;
                    CurrentEntry : integer;
                    NextEntry : integer;
                  end;  { ListOfRecords }

var

  {-----------------------------------------------------------------}
  {  The following variables represent parameter values used in the }
  {  simulation.  These variables are assigned values at the start  }
  {  of the simulation by the procedure GetParameters.  Thereafter, }
  {  they remain constant during the execution of the program.      }
  {                                                                 }
  {  These variables are used freely by the other procedures of the }
  {  program, as though they were true constants.  They are not     }
  {  passed as parameters, except to the procedure GetParameters.   }
  {-----------------------------------------------------------------}

  FishInitial : integer;       { initial number of fish in the ocean }
  FishBreed : integer;         { age at which a fish reproduces }
  FishLimit : integer;         { maximum number of fish }
  SharkInitial : integer;      { initial number of sharks in the ocean }
  SharkBreed : integer;        { age at which a shark reproduces }
  SharkStarve : integer;       { length of time a shark can go without eating }

  {-----------------------------------------------------------------}
  { End of simulation parameters                                    }
  {-----------------------------------------------------------------}

  Ocean : CellArray;           { current contents of ocean }
  OldOcean : CellArray;        { previous contents of ocean }
  FishList : ListOfRecords;    { list of fish at end of last breeding cycle }
  SharkList : ListOfRecords;   { list of sharks at end of last breeding cycle }
  CycleNumber : integer;       { number of chronons simulated so far }
  WantsToContinue : boolean;   { user wants to continue the simulation }

{*********************************************************************}

{---------------------------------------------------------------------}
{  The following compiler directives include the file "screen.i",     }
{  which contains the following procedures:                           }
{    ClearScreen - clear terminal screen and home cursor              }
{    MoveCursor - move cursor to a specified location                 }
{    ClearLine - clear the line on which the cursor is positioned     }
{---------------------------------------------------------------------}

{$I screen.i} { for Turbo Pascal; treated as comment in standard Pascal }

{ ## Non-standard statements -- Berkeley Pascal version }
{
#include "screen.i"
}
{ End Berkeley Pascal version }

{*********************************************************************}

{---------------------------------------------------------------------}
{  The following compiler directives include the file "display.i",    }
{  which contains the following procedure:                            }
{    Display - display two-dimensional array on terminal              }
{---------------------------------------------------------------------}

{$I display.i} { for Turbo Pascal; treated as comment in standard Pascal }

{ ## Non-standard statements -- Berkeley Pascal version }
{
#include "display.i"
}
{ End Berkeley Pascal version }

{*********************************************************************}

{---------------------------------------------------------------------}
{  The following compiler directives include the file "random.i",     }
{  which contains the following procedure and function:               }
{    InitRandom - initialize the random number generator              }
{    RandomNumber - generate a random number between 0 and 1          }
{---------------------------------------------------------------------}

{$I random.i} { for Turbo Pascal; treated as comment in standard Pascal }

{ ## Non-standard statements -- Berkeley Pascal version }
{
#include "random.i"
}
{ End Berkeley Pascal version }

{*********************************************************************}

{---------------------------------------------------------------------}
{  The following compiler directives include the file "reclist.i",    }
{  which contains the following procedures and function.              }
{                                                                     }
{  Procedures:                                                        }
{    InitList - initialize a new (empty) list                         }
{    GetFirst - get the first record (if any) in a list               }
{    GetNext - get the next record (if any) in a list                 }
{    UpdateCurrent - update the last record read from a list          }
{    DeleteCurrent - delete the last record read from a list          }
{    AddNew - add a new record to a list                              }
{                                                                     }
{  Function:                                                          }
{    NumberIn - returns the number of records in a list               }
{---------------------------------------------------------------------}

{$I reclist.i} { for Turbo Pascal; treated as comment in standard Pascal }

{ ## Non-standard statements -- Berkeley Pascal version }
{
#include "reclist.i"
}
{ End Berkeley Pascal version }

{*********************************************************************}

procedure GetParameters (var FishInitial : integer; var FishBreed : integer;
                var FishLimit : integer; var SharkInitial : integer;
                var SharkBreed : integer; var SharkStarve : integer);

  {-------------------------------------------------------------------}
  {  GetParameters is called once at the beginning of the simulation. }
  {  Its purpose is to get from the user the parameter values that    }
  {  are to be used in the simulation.  For each parameter, a         }
  {  message is displayed that gives a description of the parameter,  }
  {  the allowable range of values it can take, and its default       }
  {  value.  The user may enter a carriage return to accept the       }
  {  default value.  If the user enters a parameter value to use      }
  {  instead of the default, the value entered is checked against the }
  {  maximum and minimum allowable values.  If the value entered is   }
  {  outside of the allowed range, an error message is displayed;     }
  {  the prompting is repeated until a valid parameter value is       }
  {  obtained.                                                        }
  {                                                                   }
  {  After all of the parameter values are obtained, GetParameters    }
  {  clears the terminal screen and displays a header line that shows }
  {  the parameter values being used for the current simulation.      }
  {                                                                   }
  {  An internal procedure GetParam is used to perform the prompting, }
  {  reading of parameter values, and range checking.                 }
  {                                                                   }
  {  Input parameters: None                                           }
  {                                                                   }
  {  Output parameters:                                               }
  {    FishInitial - the initial number of fish                       }
  {    FishBreed - the age at which a fish breeds                     }
  {    FishLimit - the maximum number of fish                         }
  {    SharkInitial - the initial number of sharks                    }
  {    SharkBreed - the age at which a shark breeds                   }
  {    SharkStarve - the length of time a shark can go without eating }
  {-------------------------------------------------------------------}

  type
    PromptString = packed array[1..30] of char;

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

  procedure GetParam (Prompt : PromptString; Minimum : integer;
                      Maximum : integer; Default : integer;
                      var ParameterValue : integer);

  {-------------------------------------------------------------------}
  {  GetParam prompts the user with a parameter description and the   }
  {  allowed range of values and default value for that parameter.    }
  {  It then reads input from the terminal.  If the user enters a     }
  {  carriage return, GetParam sets the parameter value to its        }
  {  default.  If the user enters a value, GetParam checks the value  }
  {  entered against the allowed range.  The prompting and reading    }
  {  are repeated, if necessary, until a valid parameter value is     }
  {  obtained.                                                        }
  {                                                                   }
  {  Input parameters:                                                }
  {    Prompt - the parameter description to be displayed as a prompt }
  {    Minimum - the minimum value for the parameter                  }
  {    Maximum - the maximum value for the parameter                  }
  {    Default - the default value for the parameter                  }
  {                                                                   }
  {  Output parameters:                                               }
  {    ParameterValue - the value obtained from the user, or the      }
  {                     default if no value was entered               }
  {-------------------------------------------------------------------}

    var
      ValidParameter : boolean;    { a valid parameter value was obtained }
      Bell : char;                 { character that causes the terminal
                                     bell to sound }

    begin

      { if ASCII codes are being used, the character with ordinal value 7
        will ring the terminal bell }

      Bell := chr(7);

      { display the prompt string and read input from the terminal, repeating
        until a valid parameter value is obtained }

      repeat
        write(Prompt,'(',Minimum:3,' - ',Maximum:3,')  [',Default:3,']  :');
        if eoln(input) then
          begin

            { user entered a carriage return -- accept the default value }

            readln;
            ParameterValue := Default;
            ValidParameter := true
          end  { if default value accepted }
        else
          begin

            { user entered a value -- read it and check it for validity }

            readln(ParameterValue);
            if (ParameterValue < Minimum) or (ParameterValue > Maximum) then
              begin
                ValidParameter := false;
                write(Bell)
              end  { if value entered was out of range }
            else
              ValidParameter := true
          end  { else -- user entered a parameter value }
      until ValidParameter

    end;  { GetParam }

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

  begin  { GetParameters }

    { clear the terminal screen and display instructions }

    ClearScreen;
    writeln('               WaTor Parameter Selection');
    writeln;
    writeln('Enter desired values for each parameter (range of possible');
    writeln('values is given in parentheses), or press return to accept the');
    writeln('default value given in [...].');
    writeln;

    { call GetParam to get a value for each parameter }

    GetParam('FI: Initial number of fish    ',
             1,999,FishInitialDefault,FishInitial);
    GetParam('FB: Fish breeding age         ',
             1,100,FishBreedDefault,FishBreed);
    GetParam('FL: Limit on number of fish   ',
             FishInitial,999,FishLimitDefault,FishLimit);
    GetParam('SI: Initial number of sharks  ',
             1,999,SharkInitialDefault,SharkInitial);
    GetParam('SB: Shark breeding age        ',
             1,100,SharkBreedDefault,SharkBreed);
    GetParam('SS: Shark starvation time     ',
             1,100,SharkStarveDefault,SharkStarve);

    { display the header message showing parameter values being used }

    ClearScreen;
    writeln('FI = ',FishInitial:3,'     FB = ',FishBreed:3,
            '     FL = ',FishLimit:3,'     SI = ',SharkInitial:3,
            '     SB = ',SharkBreed:3,'     SS = ',SharkStarve:3);

  end;  { GetParameters }

{*********************************************************************}

procedure Initialize(var Ocean : CellArray; var OldOcean : CellArray;
                var FishList : ListOfRecords; var SharkList : ListOfRecords);

  {-------------------------------------------------------------------}
  {  Initialize is called after the parameter values are obtained.    }
  {  It places the initial number of sharks and fish (given by the    }
  {  parameters FishInitial and SharkInitial) at random into Ocean    }
  {  and sets OldOcean to blanks.  Information about the sharks and   }
  {  fish generated is stored in the lists SharkList and FishList.    }
  {                                                                   }
  {  Input parameters: None                                           }
  {                                                                   }
  {  Output parameters:                                               }
  {    Ocean - initialized with specified numbers of sharks and fish  }
  {    OldOcean - set to blanks                                       }
  {    FishList - list containing information about fish generated    }
  {    SharkList - list containing information about sharks generated }
  {-------------------------------------------------------------------}

  var
    Row : RowSubscript;       { Row and Col are used as subscripts }
    Col : ColSubscript;       {    for Ocean and OldOcean          }
    NewRow : RowSubscript;    { NewRow and NewCol indicate the position }
    NewCol : ColSubscript;    {    to generate a new shark or fish      }
    Count : integer;          { number of sharks or fish generated so far }
    Fish : ListEntry;         { fish being generated }
    Shark : ListEntry;        { shark being generated }

  begin  { Initialize }

    { initialize the random number generator }

    InitRandom(0);

    { set all cells in Ocean and OldOcean to blank }

    for Row := 1 to CellRows do
      for Col := 1 to CellCols do
        begin
          Ocean[Row,Col] := ' ';
          OldOcean[Row,Col] := ' '
        end;  { for }

    { initialize fish list and generate initial number of fish }

    InitList(FishList,FishLimit);
    for Count := 1 to FishInitial do
      begin

        { select cells of Ocean at random until a blank cell is found  --
          note that RandomNumber is a function that generates a different
          random number each time it is called }

        repeat
          NewRow := trunc(RandomNumber * CellRows) + 1;
          NewCol := trunc(RandomNumber * CellCols) + 1
        until Ocean[NewRow,NewCol] = ' ';

        { place fish in the ocean and enter it in FishList, setting its age
          at random in the allowed range }

        Ocean[NewRow,NewCol] := FishSymbol;
        Fish.Row := NewRow;
        Fish.Col := NewCol;
        Fish.Age := trunc(RandomNumber * FishBreed);
        AddNew(FishList,Fish)
      end;  { for }

    { initialize shark list and generate initial number of sharks }

    InitList(SharkList,MaximumListSize);
    for Count := 1 to SharkInitial do
      begin

        { select cells of Ocean at random until a blank cell is found  --
          note that RandomNumber is a function that generates a different
          random number each time it is called }

        repeat
          NewRow := trunc(RandomNumber * CellRows) + 1;
          NewCol := trunc(RandomNumber * CellCols) + 1
        until Ocean[NewRow,NewCol] = ' ';

        { place shark in the ocean and enter it in SharkList, setting its age
          and time since eating at random in the allowed ranges }

        Ocean[NewRow,NewCol] := SharkSymbol;
        Shark.Row := NewRow;
        Shark.Col := NewCol;
        Shark.Age := trunc(RandomNumber * SharkBreed);
        Shark.TimeSinceFed := trunc(RandomNumber * SharkStarve);
        AddNew(SharkList,Shark)
      end  { for }

  end;  { Initialize }

{*********************************************************************}

procedure FindMove (var Ocean : CellArray; OldRow : RowSubscript;
               OldCol : ColSubscript; Target : char; var MoveFound : boolean;
               var NewRow : RowSubscript; var NewCol : ColSubscript);

  {-------------------------------------------------------------------}
  {  FindMove finds a cell in Ocean that is adjacent to the cell      }
  {  at [OldRow,OldCol] and contains the character specified in       }
  {  Target.  If such a cell exists, the parameter MoveFound is set   }
  {  to true and the location of the cell is returned in the          }
  {  parameters NewRow and NewCol. If no such cell exists, MoveFound  }
  {  is set to false and NewRow and NewCol are unchanged.  If more    }
  {  than one such cell exists, one of them is selected at random.    }
  {                                                                   }
  {  In examining cells adjacent to [OldRow,OldCol], the left and     }
  {  right edges of the ocean are considered to be adjacent, and the  }
  {  top and bottom edges are considered to be adjacent.              }
  {                                                                   }
  {  FindMove is used to find moves for sharks and fish, selecting a  }
  {  move at random if more than one is available.  For example, it   }
  {  is called with Target = ' ' to find an empty cell for a fish to  }
  {  swim to; it is called with Target = FishSymbol to find an        }
  {  adjacent fish for a shark to eat.                                }
  {                                                                   }
  {  Input parameters:                                                }
  {    Ocean - the current generation of sharks and fish              }
  {        Note: Ocean is a var parameter for efficiency in           }
  {              parameter passing                                    }
  {    OldRow,OldCol - the row and column of the cell from which a    }
  {                    move is to be found                            }
  {    Target - the character to be sought in an adjacent cell        }
  {                                                                   }
  {  Output parameters:                                               }
  {    MoveFound - true if a move is found, and false otherwise       }
  {    NewRow,NewCol - the row and column of the cell selected for    }
  {                    the move.  If no move is found, NewRow and     }
  {                    NewCol are unchanged.                          }
  {-------------------------------------------------------------------}

  type
    Move = record                     { entry in array of possible moves }
             MoveRow : RowSubscript;    { MoveRow and MoveCol are the row and }
             MoveCol : ColSubscript     {    column of a possible move        }
           end;

  var
    PossibleMoves : array[1..8] of Move; { possible moves that are found }
    NumberOfMoves : integer;             { number of possible moves found }

    TryRow : integer;                    { the row and column of a cell to be }
    TryCol : integer;                    {    examined, before adjusting for  }
                                         {    the edges of the ocean          }

    Row : RowSubscript;                  { the row and column of a cell to be }
    Col : ColSubscript;                  {    examined, after adjusting for   }
                                         {    the edges of the ocean          }

    I : integer;                         { subscript for PossibleMoves }

  begin  { FindMove }

    NumberOfMoves := 0;

    { examine cells in the neighborhood of the cell specified }

    for TryRow := OldRow - 1 to OldRow + 1 do
      for TryCol := OldCol - 1 to OldCol + 1 do
        begin

          { adjust row and column numbers for the edges of the ocean }

          if TryRow < 1 then Row := CellRows
            else if TryRow > CellRows then Row := 1
            else Row := TryRow;
          if TryCol < 1 then Col := CellCols
            else if TryCol > CellCols then Col := 1
            else Col := TryCol;

          { if the cell being examined is not the original cell, and contains
            the target being sought, enter it in the array of possible moves }

          if (Row <> OldRow) or (Col <> OldCol) then
            if Ocean[Row,Col] = Target then
              begin
                NumberOfMoves := NumberOfMoves + 1;
                PossibleMoves[NumberOfMoves].MoveRow := Row;
                PossibleMoves[NumberOfMoves].MoveCol := Col
              end  { if }
        end;

    { if the array contains no possible moves, set MoveFound to false.
      otherwise, set MoveFound to true and select a move at random from the
      array.  note that RandomNumber is a function that generates a
      different number each time it is called }

    if NumberOfMoves = 0 then
      MoveFound := false
    else
      begin
        MoveFound := true;
        I := trunc(RandomNumber * NumberOfMoves) + 1;
        NewRow := PossibleMoves[I].MoveRow;
        NewCol := PossibleMoves[I].MoveCol
      end  { else }

  end;  { FindMove }

{*********************************************************************}

procedure MoveFish(var Ocean : CellArray; var FishList : ListOfRecords);

  {-------------------------------------------------------------------}
  {  MoveFish simulates the swimming and breeding of fish during one  }
  {  chronon of the simulation.                                       }
  {                                                                   }
  {  For each fish, FindMove is called to search for an empty cell    }
  {  that it can swim to.  If such a cell is found, the fish is moved }
  {  to that location.  If the fish being moved is of breeding age,   }
  {  and the number of fish is not at the maximum allowed, the fish   }
  {  produces an offspring by cloning.  The offspring is added to the }
  {  end of the list of fish, and temporarily remains at the original }
  {  location of the parent.  (It will be examined for a possible     }
  {  swimming move later during the scanning of the list.)  The ages  }
  {  of both the parent and the offspring are set to 0.               }
  {                                                                   }
  {  If there is no adjacent empty cell, the fish does not swim or    }
  {  breed.                                                           }
  {                                                                   }
  {  Input parameters:                                                }
  {    Ocean - the current generation of sharks and fish              }
  {    FishList - the list of fish currently in existence             }
  {                                                                   }
  {  Output parameters:                                               }
  {    Ocean - the current generation of sharks and fish, reflecting  }
  {            the swimming and breeding of fish                      }
  {    FishList - the list of fish after swimming and breeding        }
  {-------------------------------------------------------------------}

  var
    Fish : ListEntry;             { the fish currently being examined }
    MoveFound : boolean;          { a move was found for the fish }
    NewRow : RowSubscript;        { the row and column of the cell (if any) }
    NewCol : ColSubscript;        {    to which the fish will move          }
    NewFish : ListEntry;          { a new fish being generated by breeding }
    EndOfFish : boolean;          { all fish have been examined }

  begin  { MoveFish }

    { scan the list of fish, generating swimming and breeding moves }

    GetFirst(FishList,Fish,EndOfFish);
    while not EndOfFish do
      begin

        { add one chronon to the age of the fish }

        Fish.Age := Fish.Age + 1;

        { find a swimming move for the fish, if one exists }

        FindMove(Ocean,Fish.Row,Fish.Col,' ',MoveFound,NewRow,NewCol);
        if MoveFound then
          begin

            { a swimming move was found.  check to see if the fish can breed }

            if (Fish.Age >= FishBreed)
                      and (NumberIn(FishList) < FishLimit) then
              begin

                { the fish can breed.  generate the offspring and add it to the
                  list, setting the ages of both fish to 0.  the offspring
                  remains at the original location of the parent }

                NewFish := Fish;
                NewFish.Age := 0;
                AddNew(FishList,NewFish);
                Fish.Age := 0
              end  { if }
            else

              { the fish cannot breed; its original location becomes empty }

              Ocean[Fish.Row,Fish.Col] := ' ';

            { move the fish to its new location }

            Ocean[NewRow,NewCol] := FishSymbol;
            Fish.Row := NewRow;
            Fish.Col := NewCol
          end;  { if swimming move was found }

        { update current fish and get the next fish in the list }

        UpdateCurrent(FishList,Fish);
        GetNext(FishList,Fish,EndOfFish)
      end  { while }

  end;  { MoveFish }

{*********************************************************************}

procedure MoveSharks(var Ocean : CellArray; var SharkList : ListOfRecords;
                     var FishList : ListOfRecords);

  {-------------------------------------------------------------------}
  {  MoveSharks simulates the hunting, swimming, and breeding of      }
  {  sharks during one chronon of the simulation.  It first scans the }
  {  list of sharks, removing the entries for sharks that have        }
  {  starved during the previous chronon.  It then scans the list a   }
  {  second time, generating the hunting, swimming, and breeding      }
  {  moves for the sharks that remain.  After all of the sharks have  }
  {  moved, the fish that were eaten are removed from the list of     }
  {  fish.                                                            }
  {                                                                   }
  {  For each shark, FindMove is called to search for an adjacent     }
  {  cell that contains a fish.  If such a cell is found, the shark   }
  {  swims to that location and eats the fish.  If no adjacent fish   }
  {  is found, FindMove is called again to search for an empty cell   }
  {  that the shark can swim to.  If the shark being moved is of      }
  {  breeding age, it produces an offspring by cloning.  The          }
  {  offspring is added to the end of the list of sharks, and         }
  {  temporarily remains at the original location of the parent.      }
  {  (It will be examined for possible hunting and swimming moves     }
  {  later during the scanning of the list.)  The ages  of both the   }
  {  parent and the offspring are set to 0.                           }
  {                                                                   }
  {  If there is no adjacent empty cell, the shark does not swim or   }
  {  breed.                                                           }
  {                                                                   }
  {  Input parameters:                                                }
  {    Ocean - the current generation of sharks and fish              }
  {    SharkList - the list of sharks currently in existence          }
  {    FishList - the list of fish currently in existence             }
  {                                                                   }
  {  Output parameters:                                               }
  {    Ocean - the current generation of sharks and fish, reflecting  }
  {            the hunting, swimming, and breeding of sharks          }
  {    SharkList - the list of sharks after swimming and breeding     }
  {    FishList - the list of fish, after removing those eaten by     }
  {               sharks                                              }
  {-------------------------------------------------------------------}

  var
    Shark : ListEntry;            { the shark currently being examined }
    MoveFound : boolean;          { a move was found for the shark }
    NewRow : RowSubscript;        { the row and column of the cell (if any) }
    NewCol : ColSubscript;        {    to which the shark will move         }
    NewShark : ListEntry;         { a new shark being generated by breeding }
    EndOfSharks : boolean;        { all sharks have been examined }
    Fish : ListEntry;             { the fish currently being examined }
    EndOfFish : boolean;          { all fish have been examined }

  begin  { MoveSharks }

    { scan the list of sharks, removing those that have starved }

    GetFirst(SharkList,Shark,EndOfSharks);
    while not EndOfSharks do
      begin
        if Shark.TimeSinceFed > SharkStarve then
          begin
            Ocean[Shark.Row,Shark.Col] := ' ';
            DeleteCurrent(SharkList)
          end;  { if shark has starved }
        GetNext(SharkList,Shark,EndOfSharks)
      end;  { while }

    { scan the list again, generating hunting, swimming, and breeding moves }

    GetFirst(SharkList,Shark,EndOfSharks);
    while not EndOfSharks do
      begin

        { add one chronon to the age of the shark and the time since fed }

        Shark.Age := Shark.Age + 1;
        Shark.TimeSinceFed := Shark.TimeSinceFed + 1;

        { find a fish for the shark to eat, if one exists }

        FindMove(Ocean,Shark.Row,Shark.Col,FishSymbol,MoveFound,NewRow,NewCol);

        { if a fish is found, set time since fed to 0; otherwise, find a
          swimming move for the shark }

        if MoveFound then
          Shark.TimeSinceFed := 0
        else
          FindMove(Ocean,Shark.Row,Shark.Col,' ',MoveFound,NewRow,NewCol);
        if MoveFound then
          begin

            { a hunting or swimming move was found.  check to see if the
              shark can breed }

            if (Shark.Age >= SharkBreed) then
              begin

                { the shark can breed.  generate the offspring and add it to
                  the list, setting the ages of both sharks to 0.  the
                  offspring remains at the original location of the parent }

                NewShark := Shark;
                NewShark.Age := 0;
                AddNew(SharkList,NewShark);
                Shark.Age := 0
              end  { if }
            else

              { the shark cannot breed; its original location becomes empty }

              Ocean[Shark.Row,Shark.Col] := ' ';

            { move the shark to its new location }

            Ocean[NewRow,NewCol] := SharkSymbol;
            Shark.Row := NewRow;
            Shark.Col := NewCol
          end;  { if move was found }

        { update current shark and get next shark in the list }

        UpdateCurrent(SharkList,Shark);
        GetNext(SharkList,Shark,EndOfSharks)
      end;  { while }

    { scan the list of fish, removing those eaten by sharks }

    GetFirst(FishList,Fish,EndOfFish);
    while not EndOfFish do
      begin

        { if the position of the fish is now occupied by a shark, remove it
          from the list }

        if Ocean[Fish.Row,Fish.Col] = SharkSymbol then
          DeleteCurrent(FishList);
        GetNext(FishList,Fish,EndOfFish)
      end { while }

  end;  { MoveSharks }

{*********************************************************************}

procedure DisplayOcean(var Ocean : CellArray; var OldOcean : CellArray;
               var FishList : ListOfRecords; var SharkList : ListOfRecords);

  {-------------------------------------------------------------------}
  {  DisplayOcean is called to display the current ocean on the       }
  {  terminal.  It calls the procedure Display (from the file         }
  {  "display.i") to display the contents of Ocean, beginning on      }
  {  line 3 of the screen.  It then displays the total numbers of     }
  {  fish and sharks on a line below the ocean.                       }
  {                                                                   }
  {  The procedure Display attempts to avoid re-displaying parts of   }
  {  the ocean that have not changed from the previous version        }
  {  (OldOcean).  On the first call to DisplayOcean, it is assumed    }
  {  that the terminal screen has been cleared to blanks and that     }
  {  OldOcean contains all blanks.                                    }
  {                                                                   }
  {  After Ocean is displayed, its contents are copied to OldOcean.   }
  {                                                                   }
  {  Input parameters:                                                }
  {    Ocean - the current ocean to be displayed                      }
  {    OldOcean - the previous version of the ocean, or blanks if     }
  {               this is the first version displayed                 }
  {    FishList - the list of fish currently in existence             }
  {    SharkList - the list of sharks currently in existence          }
  {                                                                   }
  {      Note: All parameters are var parameters for efficiency in    }
  {            parameter passing.                                     }
  {                                                                   }
  {  Output parameters:                                               }
  {    OldOcean - set to the contents of Ocean                        }
  {-------------------------------------------------------------------}

  begin  { DisplayOcean }

    { display the current ocean and totals of fish and sharks }

    Display(Ocean,OldOcean,3);
    MoveCursor(22,1);
    writeln('Fish = ',NumberIn(FishList):4,
            '            Sharks = ',NumberIn(SharkList):4);

    { copy ocean displayed to previous ocean }

    OldOcean := Ocean

  end;  { DisplayOcean }

{*********************************************************************}

procedure AskUser(var WantsToContinue : boolean);

  {-------------------------------------------------------------------}
  {  AskUser is called after the simulation has run for a number      }
  {  of chronons given by the constant PauseAfter.  It asks the       }
  {  user if he or she wants to continue the simulation, and sets the }
  {  parameter WantsToContinue according to the answer received.      }
  {                                                                   }
  {  If the first character entered by the user is 'y' or 'Y',        }
  {  WantsToContinue is set to true; if the first character is        }
  {  'n' or 'N', WantsToContinue is set to false.  The prompting is   }
  {  repeated if necessary until one of these two results is obtained.}
  {                                                                   }
  {  Input parameters: None                                           }
  {                                                                   }
  {  Output parameters:                                               }
  {    WantsToContinue - true if the user wants to continue the       }
  {                      simulation                                   }
  {-------------------------------------------------------------------}

  var
    InputCharacter : char;     { the first character entered by the user }

  begin  { AskUser }

    { position the cursor on a line below the ocean display.  prompt and
      read from the terminal until one of the characters 'y','Y','n','N'
      is obtained }

    repeat
      MoveCursor(23,1);
      write('Continue? (y/n) : ');
      readln(InputCharacter)
    until InputCharacter in ['y','Y','n','N'];

    { set WantsToContinue according to the input received }

    WantsToContinue := (InputCharacter in ['y','Y'])

  end;  { AskUser }

{*********************************************************************}

procedure CleanUp;

  {-------------------------------------------------------------------}
  {  CleanUp is called once at the end of the program.  It moves the  }
  {  cursor to a line below the ocean display, and waits for the user }
  {  to enter a carriage return before exiting the program.           }
  {-------------------------------------------------------------------}

  begin  { CleanUp }

    MoveCursor(23,1);
    ClearLine;
    write('Press return to exit ');
    readln

  end;  { CleanUp }

{*********************************************************************}

begin  { main program }

  { get the simulation parameters from the user and initialize the ocean }

  GetParameters(FishInitial,FishBreed,FishLimit,SharkInitial,SharkBreed,
                SharkStarve);
  Initialize(Ocean,OldOcean,FishList,SharkList);

  { run the simulation until the number of sharks or the number of fish
    falls to zero, or until the user does not want to continue }

  CycleNumber := 0;
  WantsToContinue := true;
  repeat  { until the user wants to stop, or all the sharks or fish are gone }

    { move the fish and display the result }

    MoveFish(Ocean,FishList);
    DisplayOcean(Ocean,OldOcean,FishList,SharkList);

    { move the sharks and display the result }

    MoveSharks(Ocean,SharkList,FishList);
    DisplayOcean(Ocean,OldOcean,FishList,SharkList);

    { ask the user for instructions after each PauseAfter chronons }

    CycleNumber := CycleNumber + 1;
    if (CycleNumber mod PauseAfter) = 0 then
        AskUser(WantsToContinue)

  until (WantsToContinue = false) or (NumberIn(FishList) = 0)
         or (NumberIn(SharkList) = 0);

  CleanUp

end.
