program Loan2(input,output);

{--------------------------------------------------------------------}
{  This program simulates the financing of a purchase over a         }
{  one-year period, with the payments being drawn from a savings     }
{  account.  Interest on savings is paid monthly.  The program       }
{  displays the amount of interest earned, payment made, new savings }
{  balance, interest paid, and new loan balance for each month.  It  }
{  also calculates and displays the total amount of interest earned  }
{  on savings for the year, and the total amount of interest paid    }
{  (finance charges) for the year.                                   }
{                                                                    }
{  Input :   none                                                    }
{            (The amount being financed, monthly payment, starting   }
{             balance in savings, and interest rate on savings are   }
{             given by constants in the program.)                    }
{                                                                    }
{  Output :  Listing of interest earned, payment made, new savings   }
{            balance, interest paid, and new loan balance for each   }
{            month from 1 (first month of loan) to 12 (last month    }
{            of loan).  Listing of total interest earned and total   }
{            interest paid over the period of the loan.              }
{                                                                    }
{  Modification history:                                             }
{            This program performs the same input, output, and       }
{            computations as the original version (program Loan1).   }
{            However, it uses parameters to pass information to and  }
{            from procedures, whereas the original program used      }
{            global variables for this purpose.                      }
{--------------------------------------------------------------------}

const
  AmountFinanced = 3000;       { amount financed }
  PaymentAmount = 263.05;      { amount of monthly payment }
  StartingBalance = 3000;      { starting balance in savings account }
  SavingsRate = 0.065;         { annual interest rate on savings }
  LoanRate = 0.095;            { annual interest rate on loan }

var
  Month : integer;             { current month number, starting at 1 }
  SavingsBalance : real;       { current savings balance }
  LoanBalance : real;          { current loan balance }
  InterestEarned : real;       { interest earned for current month }
  InterestPaid : real;         { interest paid for current month   }
  TotalInterestEarned : real;  { total interest earned }
  TotalInterestPaid : real;    { total interest paid }

{********************************************************************}

procedure Initialize(var SavingsBalance : real; var LoanBalance : real;
               var TotalInterestEarned : real; var TotalInterestPaid : real);

  {------------------------------------------------------------------}
  {  Initialize is called once at the beginning of the program.      }
  {  It sets the savings balance and loan balance to their starting  }
  {  values, and initializes the totals to zero.                     }
  {                                                                  }
  {  Input parameters: None                                          }
  {                                                                  }
  {  Output parameters:                                              }
  {    SavingsBalance - starting savings balance from constant       }
  {                     StartingBalance                              }
  {    LoanBalance - starting loan balance from constant             }
  {                  AmountFinanced                                  }
  {    TotalInterestEarned - zero                                    }
  {    TotalInterestPaid - zero                                      }
  {------------------------------------------------------------------}

  begin  { Initialize }

    { set starting balances }

    SavingsBalance := StartingBalance;
    LoanBalance := AmountFinanced;

    { initialize totals to zero }

    TotalInterestEarned := 0;
    TotalInterestPaid := 0

  end;   { Initialize }

{********************************************************************}

procedure WriteHeadings;

  {------------------------------------------------------------------}
  {  WriteHeadings is called once at the beginning of the program.   }
  {  It writes the headings for the listing of monthly results.      }
  {------------------------------------------------------------------}

  begin  { WriteHeadings }

    writeln;
    writeln('         INTEREST              SAVINGS    INTEREST    LOAN  ');
    writeln('MONTH     EARNED    PAYMENT    BALANCE      PAID     BALANCE');
    writeln

  end;   { WriteHeadings }

{********************************************************************}

procedure ComputeNewBalance(var SavingsBalance : real;
               var InterestEarned : real; var TotalInterestEarned : real;
               var LoanBalance : real; var InterestPaid : real;
               var TotalInterestPaid : real);

  {------------------------------------------------------------------}
  {  ComputeNewBalance is called once for each month of the loan.    }
  {  It computes the amount of interest earned and the amount of     }
  {  interest paid for the current month.  The new savings balance   }
  {  and new loan balance are computed, reflecting the interest      }
  {  earned, interest paid, and loan payment made.  The interest     }
  {  earned and interest paid are added to the yearly totals.        }
  {                                                                  }
  {  Input parameters:                                               }
  {    SavingsBalance - savings balance at start of the current month}
  {    Loan Balance - loan balance at start of the current month     }
  {    TotalInterestEarned - total interest earned on savings prior  }
  {                          to the current month                    }
  {    TotalInterestPaid - total interest paid on loan prior to the  }
  {                        current month                             }
  {                                                                  }
  {  Output parameters:                                              }
  {    SavingsBalance - savings balance at end of the current month  }
  {    InterestEarned - amount of interest earned on savings for the }
  {                     current month                                }
  {    TotalInterestEarned - total interest earned on savings        }
  {                          including the current month             }
  {    LoanBalance - loan balance at end of the current month        }
  {    InterestPaid - amount of interest paid on loan for the        }
  {                   current month                                  }
  {    TotalInterestPaid - total interest paid on the loan including }
  {                        the current month                         }
  {------------------------------------------------------------------}

  begin  { ComputeNewBalance }

    {  compute interest earned and interest paid for one month --
        monthly interest is 1/12 of annual interest }

    InterestEarned := SavingsBalance * (SavingsRate / 12);
    InterestPaid := LoanBalance * (LoanRate / 12);

    {  compute new savings and loan balances }

    SavingsBalance := SavingsBalance + InterestEarned - PaymentAmount;
    LoanBalance := LoanBalance + InterestPaid - PaymentAmount;

    {  add interest earned and interest paid to totals }

    TotalInterestEarned := TotalInterestEarned + InterestEarned;
    TotalInterestPaid := TotalInterestPaid + InterestPaid

  end;   { ComputeNewBalance }

{********************************************************************}

procedure WriteBalance(Month : integer; SavingsBalance : real;
             InterestEarned : real; LoanBalance : real; InterestPaid : real);

  {------------------------------------------------------------------}
  {  WriteBalance is called once for each month of the loan.  It     }
  {  displays the month number, interest earned, payment made,       }
  {  new savings balance, interest paid, and new loan balance.       }
  {                                                                  }
  {  Input parameters:                                               }
  {    Month - current month number                                  }
  {    SavingsBalance - savings balance at end of the current month  }
  {    InterestEarned - interest earned on savings for current month }
  {    LoanBalance - loan balance at end of the current month        }
  {    InterestPaid - interest paid on loan for the current month    }
  {                                                                  }
  {  Output parameters : None                                        }
  {------------------------------------------------------------------}

  begin  { WriteBalance }

    writeln(Month:5, InterestEarned:11:2, PaymentAmount:11:2,
           SavingsBalance:11:2, InterestPaid:11:2, LoanBalance:11:2)

  end;   { WriteBalance }

{********************************************************************}

procedure WriteTotals(TotalInterestEarned : real; TotalInterestPaid : real);

  {------------------------------------------------------------------}
  {  WriteTotals is called once at the end of the program.  It       }
  {  displays the total amount of interest earned and the total      }
  {  amount of interest paid for the term of the loan.               }
  {                                                                  }
  {  Input parameters:                                               }
  {    TotalInterestEarned - total amount of interest earned on      }
  {                          savings                                 }
  {    TotalInterestPaid - total amount of interest paid on the loan }
  {                                                                  }
  {  Output parameters: None                                         }
  {------------------------------------------------------------------}

  begin  { WriteTotals }

    writeln;
    writeln('TOTAL INTEREST PAID   = ', TotalInterestPaid:8:2);
    writeln('TOTAL INTEREST EARNED = ', TotalInterestEarned:8:2);
    writeln

  end;   { WriteTotals }

{*********************************************************************}

procedure CleanUp;

  {-------------------------------------------------------------------}
  {  CleanUp is called once at the end of the program.  It waits for  }
  {  the user to enter a carriage return before exiting the program.  }
  {-------------------------------------------------------------------}

  begin  { CleanUp }

    write('Press return to exit program');
    readln

  end;  { CleanUp }

{********************************************************************}

begin  { main program }

  { initialize variables and write headings }

  Initialize(SavingsBalance,LoanBalance,TotalInterestEarned,TotalInterestPaid);
  WriteHeadings;

  { compute and write balances for each month of the loan }

  for Month := 1 to 12 do
    begin
      ComputeNewBalance(SavingsBalance,InterestEarned,TotalInterestEarned,
                        LoanBalance,InterestPaid,TotalInterestPaid);
      WriteBalance(Month,SavingsBalance,InterestEarned,LoanBalance,InterestPaid)
    end;

  { write totals and exit program }

  WriteTotals(TotalInterestEarned,TotalInterestPaid);
  CleanUp

end.
