program Index1(input,output,IDATA);

{---------------------------------------------------------------------}
{  This program searches a file IDATA for occurrences of an index     }
{  word specified by the user, and displays a list of pages on which  }
{  the index word was found.  If no occurrences of the index word are }
{  found, a message indicating this is displayed.  Pages in IDATA     }
{  are identified by an entry  "@ n"  at the beginning of a page,     }
{  where  "n"  is the page number.                                    }
{                                                                     }
{  The index word may contain any of the characters 'a' through 'z'   }
{  and apostrophe.  The maximum length for the index word is given by }
{  the constant MaximumWordLength.  An error message is displayed if  }
{  the user enters an index word that does not conform to these       }
{  requirements.  The case of letters is not considered when          }
{  searching for the index word.                                      }
{                                                                     }
{  Input:  The index word to be searched for is entered by the user   }
{          from the terminal.  The text to be searched, with page     }
{          numbers identified, is read from file IDATA.               }
{                                                                     }
{  Output: A list of page numbers on which the specified index word   }
{          occurs is displayed on the terminal.                       }
{                                                                     }
{  This program contains non-standard Pascal statements to support    }
{  Turbo Pascal and Berkeley Pascal versions.  Non-standard Pascal    }
{  statements appear in the following locations:                      }
{    1. in the compiler directives to include "cases.i"               }
{    2. in procedures Initialize and CleanUp                          }
{  These non-standard statements are preceded by a comment that       }
{  begins with '## N'.                                                }
{---------------------------------------------------------------------}

const
  Quote = '''';                        { single quote (apostrophe) }
  PageFlag = '@';                      { marks the page number in IDATA }
  MaximumWordLength = 20;              { longest word to be considered }
  BlankWord = '                    ';  { blanks to maximum word length }

type
  AlphaString = packed array [1..MaximumWordLength] of char;
  CharacterSet = set of char;

var
  LegalCharacters : CharacterSet; { characters legal in index word }

     { Note: LegalCharacters is a constant during the execution of this
       program.  It is set by the procedure Initialize to the value
       ['a'..'z',Quote].  As usual with constants, LegalCharacters is used
       by other procedures without being passed as a parameter. }

  IDATA : text;                   { file containing text to be searched }
  TextWord : AlphaString;         { word from IDATA }
  PageNumber : integer;           { current page number in IDATA }
  IndexWord : AlphaString;        { index word to be searched for }
  PageLastFound : integer;        { page on which IndexWord was last found }
  EndOfText : boolean;            { end of IDATA has been reached }

{*********************************************************************}

{---------------------------------------------------------------------}
{  The following compiler directives include the file "cases.i",      }
{  which contains the following functions:                            }
{    Lowercase - convert letters to lower case                        }
{    Uppercase - convert letters to upper case                        }
{                                                                     }
{  Note: Uppercase is not used in this program.                       }
{---------------------------------------------------------------------}

{$I cases.i} { for Turbo Pascal; treated as comment in standard Pascal }

{ ## Non-standard statements -- Berkeley Pascal version }
{
#include "cases.i"
}
{ End Berkeley Pascal version }

{*********************************************************************}

procedure Initialize(var LegalCharacters : CharacterSet);

  {-------------------------------------------------------------------}
  {  Initialize is called once at the beginning of the program.       }
  {  It opens the file IDATA and initializes the set of characters    }
  {  considered legal as part of a word.                              }
  {                                                                   }
  {  Input Parameters: None                                           }
  {                                                                   }
  {  Output Parameters:                                               }
  {    LegalCharacters - the set of characters legal as part of a     }
  {                      word, initialized to ['a'..'z',Quote]        }
  {-------------------------------------------------------------------}

  begin  { Initialize }

    { ## Non-standard statements -- Turbo Pascal version }

    assign(IDATA,'IDATA');

    { End Turbo Pascal version }

    reset(IDATA);
    LegalCharacters := ['a'..'z',Quote]

  end;  { Initialize }

{*********************************************************************}

procedure GetIndexWord(var IndexWord : AlphaString);

  {-------------------------------------------------------------------}
  {  GetIndexWord prompts the user to enter the index word to be      }
  {  searched for.  It then reads the word entered, converting        }
  {  alphabetic characters to lower case and checking that the word   }
  {  meets the following conditions:                                  }
  {      (a) at least 1 character long                                }
  {      (b) no more than MaximumWordLength characters long           }
  {      (c) contains no characters other than 'a' through 'z' and    }
  {          apostrophe.  The set LegalCharacters is used in making   }
  {          this comparison.                                         }
  {                                                                   }
  {  An error message is displayed if the index word entered does not }
  {  meet these conditions.  The prompting and reading are repeated,  }
  {  if necessary, until a valid index word is obtained.              }
  {                                                                   }
  {  Input parameters: None                                           }
  {                                                                   }
  {  Output parameters:                                               }
  {    IndexWord - the valid index word obtained                      }
  {-------------------------------------------------------------------}

  var
    LengthCounter : integer;     { number of characters read }
    CurrentCharacter : char;     { current character }
    InputIsCorrect : boolean;    { input is correct so far }

  begin  { GetIndexWord }

    { prompt for an index word, and read input, until a valid word
      is obtained }

    repeat  { until InputIsCorrect }
      writeln;
      write('Enter index word (1 to ',MaximumWordLength:2,' letters): ');

      { initialize IndexWord to all blanks and length read to zero }

      LengthCounter := 0;
      IndexWord := BlankWord;

      { no errors found so far }

      InputIsCorrect := true;

      { read characters from terminal until end-of-line is reached or
        an error is detected }

      while (InputIsCorrect) and (not eoln(input)) do
        begin
          if LengthCounter < MaximumWordLength then
            begin

              { read next character of word and check it for validity }

              read(CurrentCharacter);
              CurrentCharacter := Lowercase(CurrentCharacter);
              if CurrentCharacter in LegalCharacters then
                begin
                  LengthCounter := LengthCounter + 1;
                  IndexWord[LengthCounter] := CurrentCharacter
                end  { if CurrentCharacter is legal }
              else
                begin
                  InputIsCorrect := false;
                  writeln('*** Index word may not contain ',Quote,
                           CurrentCharacter,Quote,'.')
                end  { if CurrentCharacter is not legal }
            end  { if length < maximum }

          else  { word entered is too long }
            begin
              InputIsCorrect := false;
              writeln('*** Index word is too long.')
            end  { word is too long }
        end;  { while }

      { be sure that at least one character was read }

      if LengthCounter < 1 then
        begin
          InputIsCorrect := false;
          writeln('*** Index word must contain at least one character.')
        end;

      { skip any unread characters and end-of-line marker }

      readln
    until InputIsCorrect

  end;  { GetIndexWord }

{*********************************************************************}

procedure GetTextWord(var TextWord : AlphaString; var EndOfText : boolean;
                      var PageNumber : integer);

  {-------------------------------------------------------------------}
  {  GetTextWord is called to read the next word from IDATA.  It      }
  {  returns the word read and the page number on which the word      }
  {  appears.  If no word is found (because the end of IDATA has      }
  {  been reached), EndOfText is set to true.                         }
  {                                                                   }
  {  A word read by this procedure consists of a string of characters }
  {  all of which are 'a' through 'z' or apostrophe.  The constant    }
  {  set LegalCharacters is used to make this check.  The maximum     }
  {  length of a word is given by the constant MaximumWordLength.     }
  {  Words longer than this length are truncated to the maximum       }
  {  length with no error message.  Alphabetic characters in TextWord }
  {  are converted to lower case.                                     }
  {                                                                   }
  {  An internal procedure FindNextWord is used to locate the         }
  {  beginning of the word to be read (if one is found), and to       }
  {  indicate the page number on which the word appears.              }
  {                                                                   }
  {  Input parameters:                                                }
  {    PageNumber - the page number on which the previous text word   }
  {                 was found, or 0 if this is the first call to      }
  {                 GetTextWord                                       }
  {                                                                   }
  {  Output parameters:                                               }
  {    TextWord - the word read from IDATA (unless end-of-file        }
  {               was detected), with alphabetic characters converted }
  {               to lower case                                       }
  {    EndOfText - set to true if the end of IDATA was found          }
  {                (and no word was read); set to false otherwise     }
  {    PageNumber - the number of the page on which TextWord was      }
  {                 found (unless end-of-file was detected)           }
  {-------------------------------------------------------------------}

  var
    LengthCounter : integer;    { number of characters in the word so far }
    CurrentCharacter : char;    { current character }

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

  procedure FindNextWord(var CurrentCharacter : char; var EndOfText : boolean;
                         var PageNumber : integer);

  {-------------------------------------------------------------------}
  {  FindNextWord locates the beginning of the next word to be read   }
  {  from IDATA, unless end-of-file is detected first.  It scans      }
  {  IDATA until it finds a character that begins a legal text        }
  {  word (one that is in the constant set LegalCharacters).  This    }
  {  character is converted to lower case (if it is alphabetic) and   }
  {  returned in the parameter CurrentCharacter.  If  the end of      }
  {  IDATA is reached before such a character is found, EndOfText is  }
  {  set to true; otherwise, EndOfText is set to false.               }
  {                                                                   }
  {  If a page number flag ( "@ n" , where "n" is the page number) is }
  {  found while scanning for the next text word, the value of "n" is }
  {  stored in the parameter PageNumber.                              }
  {                                                                   }
  {  Input parameters:                                                }
  {    PageNumber - the current page number in IDATA                  }
  {                                                                   }
  {  Output parameters:                                               }
  {    CurrentCharacter - the character that begins the next text     }
  {                word (unless end-of-file was detected), converted  }
  {                to lower case if alphabetic                        }
  {    EndOfText - set to true if the end of IDATA was reached        }
  {                without finding a word; set to false otherwise     }
  {    PageNumber - the current page number in IDATA; equal to the    }
  {                 value of the input parameter unless a page number }
  {                 flag was found while reading IDATA                }
  {                                                                   }
  {  Side effect:  the file IDATA is positioned after the first       }
  {                character of the next word (unless end-of-file     }
  {                is detected)                                       }
  {-------------------------------------------------------------------}

    var
      Scanning : boolean;     { searching for beginning of next word }

    begin  { FindNextWord }

      { read IDATA until the beginning of a word is found, or until
        end-of-file is detected }

      Scanning := true;
      while (Scanning) and (not eof(IDATA)) do
        begin
          if eoln(IDATA) then

            { if end-of-line is reached, go to next line }

            readln(IDATA)
          else
            begin
              read(IDATA,CurrentCharacter);
              if CurrentCharacter = PageFlag then

                { "@" was found -- read new page number }

                read(IDATA,PageNumber)
              else
                begin

                  { stop scanning if CurrentCharacter begins a word }

                  CurrentCharacter := Lowercase(CurrentCharacter);
                  if CurrentCharacter in LegalCharacters then
                    Scanning := false
                end  { if not PageFlag }
            end  { if not eoln }
        end;  { while }

      { set EndOfText according to whether end-of-file was detected }

      EndOfText := eof(IDATA)

    end;  { FindNextWord }

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

  begin  { GetTextWord }

    { find the beginning of the next word (if any) and the page number
      on which it occurs }

    FindNextWord(CurrentCharacter,EndOfText,PageNumber);

    { if end-of-file was not reached, read the next word }

    if not EndOfText then
      begin

        { initialize text word to all blanks }

        LengthCounter := 0;
        TextWord := BlankWord;

        { read characters until one is found that is not in LegalCharacters.
          store the characters read in TextWord, with letters converted to
          lower case.  discard characters in excess of maximum word length }

        while CurrentCharacter in LegalCharacters do
          begin
            if LengthCounter < MaximumWordLength then
              begin
                LengthCounter := LengthCounter + 1;
                TextWord[LengthCounter] := CurrentCharacter
              end;  { if }
            read(IDATA,CurrentCharacter);
            CurrentCharacter := Lowercase(CurrentCharacter)
          end  { while }
      end  { if not EndOfText }

  end;  { GetTextWord }

{*********************************************************************}

procedure CleanUp;

  {-------------------------------------------------------------------}
  {  CleanUp is called once at the end of the program.  It closes     }
  {  the file IDATA, if this is required in the Pascal version        }
  {  being used, and waits for the user to enter a carriage return    }
  {  before exiting the program.                                      }
  {-------------------------------------------------------------------}

  begin  { CleanUp }

    { ## Non-standard statements -- Turbo Pascal version }

    close(IDATA);

    { End Turbo Pascal version }

    write('Press return to exit program');
    readln

  end;  { CleanUp }

{*********************************************************************}

begin  { main program }

  Initialize(LegalCharacters);

  { get index word from user }

  GetIndexWord(IndexWord);

  { no page numbers or words found yet }

  PageNumber := 0;
  PageLastFound := -1;

  { get first word from IDATA.  if end-of-file is detected, write an
    error message. }

  GetTextWord(TextWord,EndOfText,PageNumber);
  if EndOfText then
    writeln('*** No text data supplied.')
  else
    writeln(IndexWord,' appears on pages:');

  { read words from IDATA, comparing each to the index word, until
    end-of-file is reached }

  while not EndOfText do
    begin
      if (TextWord = IndexWord) and (PageNumber <> PageLastFound) then
        begin

          { display each page number on which the index word was found,
            unless that number has already been displayed }

          PageLastFound := PageNumber;
          writeln(PageNumber:4)
        end;  { if word matches index term }
      GetTextWord(TextWord,EndOfText,PageNumber)
    end;  { while }

  { if the index word was not found on any page, inform the user of this }

  if PageLastFound = -1 then
    writeln('*** Index word not found.');
  CleanUp

end.
